/**
 ** seglocCmdTaskCodels.c
 **
 ** Codels appeles par la tache seglocCmdTask
 **
 ** Auteur: Sara Fleury
 ** Date:   15/05/99
 **
 **/

#ifdef VXWORKS
#include <vxWorks.h>
#include <time.h>
#else
#include <portLib.h>
#endif

#ifdef CLASSIFICATION
#include "xes.h"
#endif

#ifdef __linux__
#include <sys/time.h>
#include <unistd.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <h2mathLib.h>

#include "server/seglocHeader.h"
#include "server/seglocPosterWriteLibProto.h"
#include "seglocConst.h"

#include <lasdef.h>
#include <lasegm_p.h>
#include <las_struct_p.h>
#include <dbliste.h>
#include <image_mgr_p.h>

#include <odom_funct_p.h>

#include <odometer_p.h>
#include <odometer.h>  /* Glob_pos ... */
#include <err_pos_p.h>

#include <lasers_robots.h>
#include <user.h>  
#include <map_mgr_p.h>
#include <match_p.h>
#include <map_builder_p.h>

/* #include <encoders.h> */

/* libsickAccess */
#include <sick_robots_p.h>

#include "seglocSegsToPosters.h"
#include "seglocSickToSegkit.h"
#include "seglocReadLaserData.h"

# include <server/pomPosterLib.h>

/* Get from the sick module */
#include "server/sickPosterLib.h"
#include "sickConst.h"
#include "sick_save.h"

#ifdef CLASSIFICATION
#include "seglocSaveStat.h"
#endif
#include "libutils/seglocUtils.h"
#include "libmap/seglocMapUtils.h"

int seglogTestDebug=1;
int seglocVerbose = 5;
/* #define VERBOSE */


/*
typedef struct SEGLOC_ENV {
  const char * const envName;
  const char * const path;
  const char * const mapFileName;
  const char * const replaySickData;
} SEGLOC_ENV;

*/
/* static const SEGLOC_ENV seglocEnvs[] = SEGLOC_DEFAULT_ENVS;
#define SEGLOC_NB_ENVS (sizeof(seglocEnvs)/sizeof(SEGLOC_ENV)) */


static int RobId = -1;
  

#define NB_MAX_POSITIONS 500

/* #define LOG_DATA */
#ifdef LOG_DATA
static FILE *filelog;
#endif

/* poster to read current robot pos */
static BOOL posterIsPOM_POS = TRUE;
static POSTER_ID posPosterId;

static long positionsArray[NB_MAX_POSITIONS*4+1];
static long transformsArray[NB_MAX_POSITIONS*4+1];
static int  imageSegmentNum=-1;
static SICK_MEASURES_STR sickMeasures;
static SICK_CONFIG_STR sickConfig;
static laserConfiguration sick_config;
static double sick_max_range;
static BOOL seglocNoRobotMotion = FALSE;
static BOOL seglocNoMotionJustStarted = FALSE;

/* ------------------------- PROTOTYPES ------------------------- */
static BOOL seglocSegmentation(void);

static BOOL seglocStartLocalisation(int *bilan, SEGLOC_LOC_MODE locMode);
static void seglocEndLocalisation(int *bilan);
static BOOL seglocMainLocalisation(void);
static void seglocRecordSegkitTransf();

static void seglocShowResults(int verboseLevel, u_long msec);
static BOOL seglocGetCurrentOdoPos (SEGLOC_POS_VAR_SPEED *pos, int *date);
static void seglocCorrectPos(SEGLOC_POS_VAR_SPEED *prev, SEGLOC_POS_VAR *new);
static BOOL seglocTestJump(SEGLOC_POS_VAR *old, 
			   SEGLOC_POS *new);
#ifdef DEBUG
static BOOL seglocPrintJump(SEGLOC_POS_VAR *old, 
			    SEGLOC_POS *new);
#endif
static void recordSickSeglocPos();

static void seglocLocQuality();
static void seglocChangeSegkitPos(SEGLOC_POS_VAR *initPos);
static BOOL seglocUpdateCompass(void);
static void seglocRecordCompas(void);
static BOOL seglocInitLocalisation(SEGLOC_LOC_MODE locMode, 
				   SEGLOC_POS_VAR *initPos,
				   BOOL *usingCompas);

/*******************************************************************/

/* static int                     seglocPomDate; */   /* pom ticks */

static POM_POS		pomPos;		/* pom pos */
static POM_EULER_V	sensorPos;
static int sickPomDate, sickPomDate_at_loc; /* are both usefull ? */
static int initFlag=0;

/*
 * To replace GetRobotEncoders.
 * Read current estimated (with odo or pom) robot position.
 */

int
GetRobotEncoders(double *enc, int RobId)
{
  SEGLOC_POS_VAR_SPEED pomPos;
  int date;

  if (!seglocGetCurrentOdoPos(&pomPos, &date)) {
    return -1;
  }
  enc[0] = pomPos.pos.x;
  enc[1] = pomPos.pos.y;
  enc[2] = pomPos.pos.theta;
  enc[3] = enc[2];
  printf("segloc: current maintobase (%.2f,%.2f,%.2f)\n",enc[0],enc[1],enc[2]);    
  return 0;
}

/*------------------------------------------------------------------------
 * Init
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_segloc_INIT_FAILED
 */

/* SeglocInitExec  -  codel EXEC de Init
   Retourne:  EXEC END ETHER FAIL ZOMBIE */


ACTIVITY_EVENT SeglocInitExec(SEGLOC_INIT* init, int *bilan)
{

  double rob_pos[4]= {1.0,0.0,0.0,0.0}; 
  SICK_MEASURES_HEADER_STR header;
  char fileName[SEGLOC_MAX_STRLEN];
  /*int numEnv=-1;*/

#if 0
  /* record env */
  for (i=0; i<SEGLOC_NB_ENVS; i++) {
    if (!strcmp(init->initParams.envName, seglocEnvs[i].envName)) {
      numEnv = i;
      break;
    }
  }
  if (numEnv == -1) {
    *bilan = S_segloc_BAD_PARAMS;
    return ETHER;
  }

  /* Default map */
  SDI_F->fileParams.baseName[0] = '\0';
  SDI_F->fileParams.fileNum = -1;
  strcat(SDI_F->fileParams.baseName, seglocEnvs[numEnv].path);
  strcat(SDI_F->fileParams.baseName, seglocEnvs[numEnv].mapFileName);
  strcpy(SDI_F->offlineParams.baseName, seglocEnvs[numEnv].replaySickData);

  /**
   ** Load default map
   **/
  seglocLoadMapExec(&SDI_F->fileParams, bilan);
#endif


  /**
   ** Look for the position poster
   **/
  if (init->posPosterName[0] != '\0') {
    /* look for a poster of type POM_POS or POM_ME_POS */
    if (seglocPOM_POSPosterFind(init->posPosterName, 
				&posPosterId) != OK) {
      
      /* poster found but length not seem correct : try POM_ME_POS */
      if (posPosterId) {
	if (seglocPOM_ME_POSPosterFind(init->posPosterName, 
				       &posPosterId) == OK) {
	  posterIsPOM_POS = FALSE;
	}
	
	/* no correct poster */
	else {
	  if (posPosterId) printf ("segloc: poster %s of bad type\n", 
				   init->posPosterName);
	  else printf ("segloc: can't find poster %s\n", 
		       init->posPosterName);
	  posPosterId = 0;
	} /* no poster found */
      } /* POM_ME_POS */
    } /* POM_POS */
    if (posPosterId)
      printf ("segloc: robot pos from %s poster\n", 
	      init->posPosterName);
    else
      printf ("segloc: fail to find pos poste. Use sick data.\n");
  }

  /* no poster given */
  else 
    printf ("segloc: no pos poster given. Will use sick data only\n");

  /* real mode : get the robot current encoders (locoRobotPositionPosterRead) */
  if (!SDI_F->offlineParams.offlineOn) {
    if (GetRobotEncoders(rob_pos,RobId)<0) {
      *bilan = S_segloc_POSTER_NOT_FOUND;
      return ETHER;
    }
  } 
  /* offline mode */
  else {
    sprintf(fileName, "%s.%03d", 
	    SDI_F->offlineParams.baseName, SDI_F->offlineParams.fileNum);    
    sickReadFile(fileName,NULL,&header,NULL);

    rob_pos[0] = header.sickPomPos.mainToBase.euler.x;
    rob_pos[1] = header.sickPomPos.mainToBase.euler.y;
    rob_pos[2] = header.sickPomPos.mainToBase.euler.yaw;
    rob_pos[3] = rob_pos[2];
  }

  /* initializing the global position and the transform */
  init_glob_pos(rob_pos);

  /* Init SDI */
  SDI_F->seglocTransf.pos.x = 0.0;
  SDI_F->seglocTransf.pos.y = 0.0;
  SDI_F->seglocTransf.pos.theta = 0.0;

  SDI_F->sickLocoPos.pos.x = 0.0;
  SDI_F->sickLocoPos.pos.y = 0.0;
  SDI_F->sickLocoPos.pos.theta = 0.0;

  SDI_F->sickSeglocPos.pos.x = 0.0;
  SDI_F->sickSeglocPos.pos.y = 0.0;
  SDI_F->sickSeglocPos.pos.theta = 0.0;

  SDI_F->currentLocoPos.pos.x = 0.0;
  SDI_F->currentLocoPos.pos.y = 0.0;
  SDI_F->currentLocoPos.pos.theta = 0.0;

  SDI_F->currentSeglocPos.pos.x = 0.0;
  SDI_F->currentSeglocPos.pos.y = 0.0;
  SDI_F->currentSeglocPos.pos.theta = 0.0;

  if (SDI_F->offlineParams.offlineOn) {
    SDI_F->state.segmentationOn = SEGLOC_FALSE;
  }

  /* Init sick */
  /* Getting a laser image from file */
  if (SDI_F->offlineParams.offlineOn) {
    sprintf(fileName, "%s.%03d", 
	    SDI_F->offlineParams.baseName, SDI_F->offlineParams.fileNum);    
    seglocReadLaserData(TRUE, fileName, &sickMeasures);
  }
  /* Getting a laser image from poster */
  else {
    seglocReadLaserData(FALSE, SDI_F->initParams.laserDataPosterName, 
			&sickMeasures);
  }

  /* WARNING : To be done after InitSick() */
  RealVariance = SDI_F->locParams[SEGLOC_LOC_LOOP].segkitSimulErr ? 0 : 1;
  SetAccuracyMove(SDI_F->locParams[SEGLOC_LOC_LOOP].simulErrLin,
		  SDI_F->locParams[SEGLOC_LOC_LOOP].simulErrAng,
		  SDI_F->locParams[SEGLOC_LOC_LOOP].simulErrLinAng);


  initFlag=1;
  
  /* Load the map */
  if (init->mapFile.baseName[0] != '\0')
    seglocLoadMapExec(&init->mapFile, bilan); 

  /* No map given: start to build one */
  else {

    /* would be nice to be able to start automatically a new map, but no so easy */
    printf ("segloc: no map given, for now it is not allow... sorry\n");
    *bilan = S_segloc_FILE_NOT_FOUND;
#if 0
    /* add segments */
    if (!SDI_F->offlineParams.offlineOn) {
      printf ("segloc: no map given, start a new one\n");

      /* segmentation */
      SDI_F->state.segmentationOn = SEGLOC_TRUE;

      seglocSegAndLoc(bilan);

      /* add first segments */
      seglocAddSegsInMapExec (bilan);
      if (*bilan != OK)
	printf ("segloc: seglocAddSegsInMapExec failed\n");

      /* Ca semble important .... (cf RetrieveMap) */
      UpdateLocalMap();

      seglocSegAndLoc(bilan);
      
      /* localizes once the robot at (0,0,0) */
      if (!seglocStartLocalisation(bilan, SEGLOC_LOC_LOCAL))
	return ETHER;

      /* one loop for the loc */
      seglocSegAndLoc(bilan);

      seglocEndLocalisation(bilan);
    }
#endif
  }

  return ETHER;
}


int seglocPomExports() 
{
  POM_ME_POS		posPoster;	/* our position */

    sensorPos.euler.yaw = SDI_F->sickSeglocPos.pos.theta ;
    sensorPos.euler.pitch = 0.0 ;
    sensorPos.euler.roll = 0.0 ;
    sensorPos.euler.x = SDI_F->sickSeglocPos.pos.x;
    sensorPos.euler.y = SDI_F->sickSeglocPos.pos.y;
    sensorPos.euler.z = 0.0;

    memset(&sensorPos.var, 0, sizeof(sensorPos.var));
    sensorPos.var.voo=SDI_F->sickSeglocPos.var.vtt;
    sensorPos.var.vox=SDI_F->sickSeglocPos.var.vxt;
    sensorPos.var.vxx=SDI_F->sickSeglocPos.var.vxx;
    
    sensorPos.var.voy=SDI_F->sickSeglocPos.var.vyt;
    sensorPos.var.vxy=SDI_F->sickSeglocPos.var.vxy;
    sensorPos.var.vyy=SDI_F->sickSeglocPos.var.vyy;
    
/* compute our position and exports it into the POM poster*/
    posPoster.confidence = 0.8; /* xxx */
    posPoster.kind = POM_ME_ABSOLUTE;
    posPoster.date1 = sickPomDate_at_loc;
    posPoster.main = sensorPos;
    if (seglocMEPosPosPosterWrite (SEGLOC_MEPOS_POSTER_ID, &posPoster) != OK) {
      printf("seglocPomExports: could not update pos poster\n");
    }
  return 0;
}



/*------------------------------------------------------------------------
 *
 * seglocCmdInit  --  Codel d'initialisation (SDI/f, ...)
 *
 * Description: 
 * 
 * Retourne:    OK ou ERROR
 */
STATUS seglocCmdInit(int *bilan)
{
  int i;
  /*int numEnv;*/

#ifdef LOG_DATA
  if (!(filelog = fopen("/home/sara/seglocLog", "w"))) {
    printf ("Cannot open /home/sara/seglocLog\n");
  }
  else {
    printf ("/home/sara/seglocLog open\n");
  }
#endif


#ifdef CLASSIFICATION
  char hostname[SEGLOC_MAX_STRLEN];
  char *envname;

  envname = getenv("XES_HOST");
  if (envname != NULL)
    strcpy(hostname, envname);
  else    
    strcpy(hostname, "localhost");
  printf ("xes_init(%s)\n", hostname);
  if (xes_init(hostname) == -1) {
    perror("xes");
    return ERROR;
  }
  setbuf(stdout, NULL);
  setbuf(stdin, NULL);
  xes_set_title("segloc exec io");
#endif


#if 0
  /* default env */
  strcpy(SDI_F->seglocInit.envName, SEGLOC_DEFAULT_ENV);
  for (i=0; i<SEGLOC_NB_ENVS; i++) {
    if (!strcmp(SDI_F->seglocInit.envName, seglocEnvs[i].envName)) {
      numEnv = i;
      break;
    }
  }
  if (numEnv != -1) {
    SDI_F->fileParams.baseName[0] = '\0';
    SDI_F->fileParams.fileNum = -1;
    strcat(SDI_F->fileParams.baseName, seglocEnvs[numEnv].path);
    strcat(SDI_F->fileParams.baseName, seglocEnvs[numEnv].mapFileName);
    strcpy(SDI_F->offlineParams.baseName, seglocEnvs[numEnv].replaySickData);
  }
#endif

  /* Init SDI */
  SDI_F->state.mapLoaded = SEGLOC_FALSE;
#ifdef CLASSIFICATION
  SDI_F->state.segmentationOn = SEGLOC_FALSE;
#else
  SDI_F->state.segmentationOn = SEGLOC_TRUE;
#endif
  SDI_F->state.localisationOn = SEGLOC_FALSE;

  SDI_F->offlineParams.offlineOn = SEGLOC_OFFLINE_MODE;
  SDI_F->offlineParams.setSegmentationOff = SEGLOC_TRUE;
  SDI_F->offlineParams.fileNum = SEGLOC_OFFLINE_FILENUM;
/*   SDI_F->offlineParams.readPrevClassif = SEGLOC_FALSE; */

  /* Parameters for segmentation */
  SDI_F->segParams.leastSquareThres = SEGLOC_DEFAULT_SEG_LEAST_SQUARE_THRES;
  SDI_F->segParams.nbMinSegPts = SEGLOC_DEFAULT_SEG_NB_MIN_PTS;
  SDI_F->segParams.distMaxSegPts = SEGLOC_DEFAULT_SEG_DIST_MAX_PTS;

  /* Parameters for localisation */

  /* Common values */
  for (i=0; i<SEGLOC_NB_LOC_MODES; i++) {
    SDI_F->locParams[i].locMode = i;
    SDI_F->locParams[i].nbMaxFailures    = SEGLOC_DEFAULT_LOC_NB_MAX_FAILURES;
    SDI_F->locParams[i].variancesFilter  = SEGLOC_TRUE;
    SDI_F->locParams[i].visibilityFilter = SEGLOC_FALSE;
    SDI_F->locParams[i].useCompass       = SEGLOC_FALSE;
    SDI_F->locParams[i].updatePosMode    = SEGLOC_NO_UPDATE;
    SDI_F->locParams[i].angleUncertainty = DEG_TO_RAD(SEGLOC_DEFAULT_LOC_ANGLE_UNCERTAINTY_DEG);
    SDI_F->locParams[i].posDistMahaThres = SEGLOC_DEFAULT_LOC_POS_DIST_MAHA_THRES;
    SDI_F->locParams[i].mapDistMahaThres = SEGLOC_DEFAULT_LOC_MAP_DIST_MAHA_THRES;
    SDI_F->locParams[i].factUncertaintyTest = SEGLOC_DEFAULT_FACT_UNCERTAINTY_TEST;

    SDI_F->locParams[i].corridorModeAllowed = SEGLOC_DEFAULT_CORRIDOR_MODE_ALLOWED;
    SDI_F->locParams[i].segkitSimulErr = SEGLOC_DEFAULT_ERR_SIMUL;
    SDI_F->locParams[i].simulErrLin = SEGLOC_DEFAULT_ERR_SIMUL_LIN;
    SDI_F->locParams[i].simulErrAng = SEGLOC_DEFAULT_ERR_SIMUL_ANG;
    SDI_F->locParams[i].simulErrLinAng = SEGLOC_DEFAULT_ERR_SIMUL_LIN_ANG;
  }
    
  /* Specific values */
/*   SDI_F->locParams[SEGLOC_LOC_LOCAL].visibilityFilter = SEGLOC_TRUE; */
#ifndef CLASSIFICATION
  SDI_F->locParams[SEGLOC_LOC_LOCAL].useCompass      = SEGLOC_FALSE;
  SDI_F->locParams[SEGLOC_LOC_LOCAL].variancesFilter = SEGLOC_TRUE;
  SDI_F->locParams[SEGLOC_FORCE_LOC_LOCAL].posDistMahaThres = SEGLOC_DEFAULT_LOC_POS_DIST_MAHA_THRES_FORCE;
  SDI_F->locParams[SEGLOC_FORCE_LOC_LOCAL].mapDistMahaThres = SEGLOC_DEFAULT_LOC_MAP_DIST_MAHA_THRES_FORCE;

#else
  SDI_F->locParams[SEGLOC_LOC_LOCAL].variancesFilter = SEGLOC_FALSE;
  SDI_F->locParams[SEGLOC_LOC_ONCE].variancesFilter  = SEGLOC_FALSE;
  SDI_F->locParams[SEGLOC_LOC_LOOP].variancesFilter  = SEGLOC_FALSE;
#endif
/*   SDI_F->locParams[SEGLOC_LOC_LOOP].updatePosMode    = SEGLOC_UPDATE_POS; */

  /* Init data compass */
  SDI_F->compass.compassEnv.environmentNorthDeg = XR4000_COMPASS_ENVIRONMENT_NORTH_DEG;
  SDI_F->compass.compassEnv.environmentAccuracyDeg = XR4000_COMPASS_ENVIRONMENT_PRECISION_DEG;
  SDI_F->compass.compassRes.rawAngleDeg = 0.0;
  SDI_F->compass.compassRes.correctedAngleDeg = 0.0;
  SDI_F->compass.compassRes.angleUncDeg = 0.0;
  SDI_F->compass.compassRes.deltaAngleAverageDeg = 0.0;
  SDI_F->compass.compassRes.deltaAngleSquareAverageRad = 0.0;
  SDI_F->compass.compassRes.deltaAngle3sigmaDeg = 0.0;
  SDI_F->compass.compassRes.deltaAngleMinDeg = 0.0;
  SDI_F->compass.compassRes.deltaAngleMaxDeg = 0.0;
  SDI_F->compass.compassRes.nbData = 0;
  
  /** 
   ** STATS
   **/
  SDI_F->stat.nbLocOK = 0;
  SDI_F->stat.nbLocFailed = 0;
  SDI_F->stat.nbLocCrashed = 0;
  SDI_F->stat.nbPartialLoc = 0;

  return OK;
}


/*------------------------------------------------------------------------
 *
 * seglocCmdEnd  --  Codel de fin (SDI/f, ...)
 *
 * Description: 
 * 
 * Retourne:    OK ou ERROR
 */
STATUS seglocCmdEnd()
{

#ifdef LOG_DATA
  fclose(filelog);
#endif
  return OK;
}

/*------------------------------------------------------------------------
 * LoadMap
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_segloc_FILE_NOT_FOUND
 *              S_segloc_FAILED
 */

/* seglocLoadMapExec  -  codel EXEC de LoadMap
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT seglocLoadMapExec (SEGLOC_FILE_PARAMS *mapFile, int *bilan)
{
  FILE *fd;
  char fileName[SEGLOC_MAX_STRLEN+8];

  if (SDI_F->initParams.mapFile.baseName[0] == '\0') {
    printf ("seglocLoadMapExec: no file given\n");
    return ETHER;
  }

  /* Nom fichier */
  if (mapFile->num> 0)
    sprintf(fileName, "%s.%03d", 
	    mapFile->baseName, mapFile->num);
  else
    strcpy(fileName, mapFile->baseName);

  /* Test acces fichier */
  fd = fopen(fileName, "r");
  if (fd == NULL) {
    printf ("seglocLoadMapExec: cannot open file \"%s\"\n", fileName);
    *bilan = S_segloc_FILE_NOT_FOUND;
    return ETHER;
  }
  fclose(fd);

  /* Lecture de la carte */
  if(! RetrieveMap(fileName)) {
    *bilan = S_segloc_FAILED;
    return ETHER;
  }

  /* Enregistrement de la carte dans le poster */
  if(!seglocSgl2MapPoster()) {
    printf ("seglocLoadMapExec: record map poster error\n");
    *bilan = S_segloc_FAILED;
    return ETHER;
  }
  SDI_F->state.mapLoaded = SEGLOC_TRUE;
  return ETHER;
}


/* seglocSaveMapExec  -  codel EXEC de SaveMap
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT seglocSaveMapExec (SEGLOC_FILE_PARAMS *mapFile, int *bilan)
{
  FILE *fd;
  char fileName[SEGLOC_MAX_STRLEN+8];

  /* Nom fichier */
  if (mapFile->num> 0)
    sprintf(fileName, "%s.%03d", 
	    mapFile->baseName, mapFile->num);
  else
    strcpy(fileName, mapFile->baseName);

  /* Test acces fichier */
  fd = fopen(fileName, "w");
  if (fd == NULL) {
    printf ("seglocSaveMapExec: cannot open %s\n", fileName);
    *bilan = S_segloc_FILE_NOT_FOUND;
    return ETHER;
  }
  fclose(fd);

  if (SaveMapToFile(fileName) != 1) {
    printf ("seglocSaveMapExec: record map poster error\n");
    *bilan = S_segloc_FAILED;
    return ETHER;
  }
  mapFile->num;
  return ETHER;
}

/* seglocClipMapExec  -  codel EXEC de ClipMap
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT seglocClipMapExec (SEGLOC_CLIP_MAP_PARAMS *clipMapParams, 
				  int *bilan)
{
  igla *map;

  map = GetCurrentMapAddr();
  if (map == NULL) {
    printf ("seglocClipMapExec: empty global map\n");
    return ETHER;
  }

  if (clipMapParams->clipMode == SEGLOC_REMOVE_OUT) {
    if (!seglocClipMap(map, clipMapParams->xmin, clipMapParams->xmax,
		       clipMapParams->ymin, clipMapParams->ymax)) {
      printf ("seglocClipMapExec: seglocClipMap in failed\n");
      *bilan = S_segloc_FAILED;
      return ETHER;
    }
  }
  else {
    if (!seglocRemoveRectangle(map, clipMapParams->xmin, clipMapParams->xmax,
			       clipMapParams->ymin, clipMapParams->ymax)) {
      printf ("seglocClipMapExec: seglocClipMap out failed\n");
      *bilan = S_segloc_FAILED;
      return ETHER;
    }
  }
  
  /* Ca semble important .... (cf RetrieveMap) */
  UpdateLocalMap();

  /* Enregistrement de la carte dans le poster */
  if(!seglocSgl2MapPoster()) {
    printf ("seglocLoadMapExec: record map poster error\n");
    *bilan = S_segloc_FAILED;
    return ETHER;
  }

  return ETHER;
}


/* seglocAddSegsInMapExec  -  codel EXEC de AddSegsInMap
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT seglocAddSegsInMapExec (int *bilan)
{

  /* Test if there is some segments to add */
  if (imageSegmentNum == -1) {
    *bilan = S_segloc_FAILED;
    printf ("seglocAddSegsInMapExec: no segment to add\n");
    return ETHER;
  }

  /* Add segments in map (WARNING: re-localisation) */
  DataIntegration(imageSegmentNum, &sick_config);
  
  /* Enregistrement de la carte dans le poster */
  if(!seglocSgl2MapPoster()) {
    printf ("seglocLoadMapExec: record map poster error\n");
    *bilan = S_segloc_FAILED;
    return ETHER;
  }
  SDI_F->state.mapLoaded = SEGLOC_TRUE;

  return ETHER;
}


/*------------------------------------------------------------------------
 * Localize
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_segloc_LOC_FAILED
 */

/* XXX REMARQUE: Avec cette approche la reponse ne vient qu'au cycle suivant
   de la tache d'exec */
static int locOK;
static int locBad;
static int locCorridor;

/* seglocLocalizeStart  -  codel START de Localize
   Retourne:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT seglocLocalizeStart (int *bilan)
{
  locOK = locBad = locCorridor  = 0;
  if (!seglocStartLocalisation(bilan, SEGLOC_LOC_LOOP))
    return ETHER;
  return EXEC;
}

/* seglocLocalizeExec  -  codel EXEC de Localize
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT seglocLocalizeExec (int *bilan)
{
  if (!SDI_F->state.localisationOn) {
    printf ("OK %d BAD %d CORRIDOR %d\n", locOK, locBad, locCorridor);
    *bilan = SDI_F->state.bilan;
    return ETHER;
  }
  SDI_F->state.locMode = SEGLOC_LOC_LOOP;
  if (SDI_F->locReport.nbSuccesses > 0) {
    locOK++;
  }
  else if (SDI_F->locReport.nbSuccesses == 0 && 
	   SDI_F->locReport.nbPossibilities > 0) {
    locCorridor++;
  }
  else if (!seglocNoRobotMotion) {
    locBad++;
  }

  return EXEC;
}

/* seglocLocalizeExec  -  codel EXEC de Localize
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT seglocLocalizeEnd (int *bilan)
{
  seglocEndLocalisation(bilan);
  printf ("OK %d BAD %d CORRIDOR %d\n", locOK, locBad, locCorridor);
  return ETHER;
}


/* seglocLocalizeEnd  -  codel INTER de Localize
   Voir plus haut (requete Localize) */
/*------------------------------------------------------------------------
 * LocalizeOnce
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_segloc_LOC_FAILED
 */
ACTIVITY_EVENT seglocLocalizeOnceStart (SEGLOC_POS *seglocPos, int *bilan)
{
  if (!seglocStartLocalisation(bilan, SEGLOC_LOC_ONCE))
    return ETHER;
  return END;
}

/* seglocLocalizeOnceExec  -  codel EXEC de LocalizeOnce
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT seglocLocalizeOnceEnd (SEGLOC_POS *seglocPos, int *bilan)
{
  seglocEndLocalisation(bilan);
  return ETHER;
}

/*------------------------------------------------------------------------
 * LocalLoc
 *
 * Description: 
 *
 * Bilans:      OK 
 */

/* seglocLocalLocExec  -  codel EXEC de LocalLoc
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT seglocLocalLocStart (SEGLOC_UNCERTAIN_POS *initPos, 
				    SEGLOC_POS *seglocPos,
				    int *bilan)
{
  if (!seglocStartLocalisation(bilan, SEGLOC_LOC_LOCAL))
    return ETHER;
  return END;
}

ACTIVITY_EVENT seglocLocalLocEnd (SEGLOC_UNCERTAIN_POS *initPos, 
				  SEGLOC_POS *currentSeglocPos,
				  int *bilan)
{
  seglocEndLocalisation(bilan);
  return ETHER;
}


/*------------------------------------------------------------------------
 * ReLoc
 *
 * Description: 
 *
 * Bilans:      OK 
 */

/* seglocReLocStart  -  codel START de ReLoc
   Retourne:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT seglocReLocStart (SEGLOC_POS *sickSeglocPos, int *bilan)
{
  int date;
  if (!seglocGetCurrentOdoPos(&SDI_F->currentLocoPos, &date)) {
    *bilan = S_segloc_POSTER_NOT_FOUND;
    return ETHER;
  }
  SDI_F->initPos.pos.x = SDI_F->currentSeglocPos.pos.x;
  SDI_F->initPos.pos.y = SDI_F->currentSeglocPos.pos.y;
  SDI_F->initPos.pos.theta = SDI_F->currentSeglocPos.pos.theta;

  SDI_F->initPos.unc.ux = SEGLOC_X_RELOC_THRESHOLD;
  SDI_F->initPos.unc.uy = SEGLOC_Y_RELOC_THRESHOLD;
  SDI_F->initPos.unc.utheta = SEGLOC_THETA_RELOC_THRESHOLD;
  

  /* XXXX Par la suite il faudrait intgrer la variance */
  /* SDI_F->initPos.var[0] = .... ??? */

  if (!seglocStartLocalisation(bilan, SEGLOC_LOC_LOCAL))
    return ETHER;
  return END;
}

/* seglocReLocEnd  -  codel END de ReLoc
   Retourne:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT seglocReLocEnd (SEGLOC_POS *sickSeglocPos, int *bilan)
{
  seglocEndLocalisation(bilan);
  return ETHER;
}



/*------------------------------------------------------------------------
 * ResetFrame
 *
 * Description: 
 *
 * Bilans:      OK 
 */

/* seglocResetFrameExec  -  codel EXEC de ResetFrame
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT seglocResetFrameExec (SEGLOC_START_END_POS *pos, int *bilan)
{
  double start[3];
  double end[3];

  start[0] = pos->start.x;
  start[1] = pos->start.y;
  start[2] = pos->start.theta;

  end[0] = pos->end.x;
  end[1] = pos->end.y;
  end[2] = pos->end.theta;

  ChangeMapCoordinates(start,end);

  /* Enregistrement de la carte dans le poster */
  if(!seglocSgl2MapPoster()) {
    printf ("seglocLoadMapExec: record map poster error\n");
    *bilan = S_segloc_FAILED;
    return ETHER;
  }

  return ETHER;
}


/*------------------------------------------------------------------------
 *
 * seglocSegAndLoc  --  codel de l'activite' permanente
 *
 * Description: 
 * 
 * Bilans:      OK 
 * 
 * Retourne:    OK ou ERROR
 */

STATUS seglocSegAndLoc(int *bilan)
{
  int date;

  if (!initFlag) return OK;
  
  /* Segmentation */
  if (SDI_F->state.segmentationOn || SDI_F->state.localisationOn) {

    /* Free previous image (sick points AND segments) */
    if (imageSegmentNum!=-1) {
      Gfree_image(imageSegmentNum);
      imageSegmentNum = -1;
    }

    /* new segmentation */
    if (!seglocSegmentation()) {
      *bilan = S_segloc_SEG_FAILED;
      SDI_F->state.bilan = S_segloc_SEG_FAILED;
      SDI_F->state.segmentationOn = SEGLOC_FALSE;
      SDI_F->state.localisationOn = SEGLOC_FALSE;
    }
  }

  /* Localisation */
  if (SDI_F->state.localisationOn) {
    if (!seglocMainLocalisation()) 
      SDI_F->state.localisationOn = SEGLOC_FALSE;
  }
  else {
    /* Il faut toujours mettre a jour loco */
    if (!seglocGetCurrentOdoPos(&SDI_F->currentLocoPos, &date)) {
      *bilan = S_segloc_POSTER_NOT_FOUND;
    }
    recordSickSeglocPos();
    seglocCorrectPos(&SDI_F->sickLocoPos, &SDI_F->sickSeglocPos); 
  }

  if (SDI_F->state.segmentationOn || SDI_F->state.localisationOn)
    if (SDI_F->offlineParams.offlineOn)
      (SDI_F->offlineParams.fileNum)++;
  


  return OK;
}


/* ---------------------------------------------------------------------- */
/*                         LOCAL FUNCTIONS                                */
/* ---------------------------------------------------------------------- */

/* ------------------------------------------------------------
 *
 * seglocGetCurrentPos - read current robot pos from poster 
 *                       POM_POS given as Init request parameter
 *
 */

static BOOL seglocGetCurrentOdoPos (SEGLOC_POS_VAR_SPEED *pos, int *date) 
{
  static POM_POS		pomPos;		/* current pos */
  static POM_ME_POS		pomMEPos;	/* current pos */

  /* offline: keep sick data pos */
  if (SDI_F->offlineParams.offlineOn) {
    *pos = SDI_F->sickLocoPos;
    *date = sickMeasures.header.sickPomPos.date;
    return TRUE;
  }

  /* no poster position : keep pos in sick poster */
  if (!posPosterId) {
    *pos = SDI_F->sickLocoPos;
    *date = sickMeasures.header.sickPomPos.date;
    return TRUE;
  }

  /* current pos */
  if (posterIsPOM_POS) {
    if (seglocPOM_POSPosterRead(posPosterId, &pomPos) == OK) {
      pos->pos.x = pomPos.mainToBase.euler.x;
      pos->pos.y = pomPos.mainToBase.euler.y;
      pos->pos.theta = pomPos.mainToBase.euler.yaw;
      
      pos->var.vxx = pomPos.mainToBase.var.vxx;
      pos->var.vyy = pomPos.mainToBase.var.vyy;
      pos->var.vtt = pomPos.mainToBase.var.voo;
      pos->var.vxy = pomPos.mainToBase.var.vxy;
      pos->var.vxt = pomPos.mainToBase.var.vox;
      pos->var.vyt = pomPos.mainToBase.var.voy;
      
      pos->speed.v = pomPos.VLocal.euler.x;
      pos->speed.w = pomPos.VLocal.euler.yaw;
      
      *date = pomPos.date;
    }
    else {
      return FALSE;
    }
  }
  
  else {
    if (seglocPOM_ME_POSPosterRead(posPosterId, &pomMEPos) == OK) {
      pos->pos.x = pomMEPos.main.euler.x;
      pos->pos.y = pomMEPos.main.euler.y;
      pos->pos.theta = pomMEPos.main.euler.yaw;
      
      pos->var.vxx = pomMEPos.main.var.vxx;
      pos->var.vyy = pomMEPos.main.var.vyy;
      pos->var.vtt = pomMEPos.main.var.voo;
      pos->var.vxy = pomMEPos.main.var.vxy;
      pos->var.vxt = pomMEPos.main.var.vox;
      pos->var.vyt = pomMEPos.main.var.voy;
      
      pos->speed.v = pomMEPos.VLocal.euler.x;
      pos->speed.w = pomMEPos.VLocal.euler.yaw;
      
      *date = pomMEPos.date1;
    }
    else {
      return FALSE;
    }
  }
  
#ifdef VERBOSE
  printf("Loco pos\n");
  printf("%.2f %.2f %.2f\n",
	SDI_F->currentLocoPos.pos.x ,
	SDI_F->currentLocoPos.pos.y,
	SDI_F->currentLocoPos.pos.theta );
#endif

  /* Compute new position */
  seglocCorrectPos(&SDI_F->currentLocoPos, &SDI_F->currentSeglocPos);

  return TRUE;
}

static void seglocCorrectPos(SEGLOC_POS_VAR_SPEED *prev, SEGLOC_POS_VAR *new)
{
  SEGLOC_POS_VAR old;
  old.pos = prev->pos;
  old.var = prev->var;
/*   printf ("seglocCorrectPos seglocCorrectPos\n"); */
  CalcAbsPos(&old, &SDI_F->seglocTransf, new);
  
  
}


/**
 **  seglocSegmentation  -  compute segments from sick data
 **
 **
 **/
static BOOL seglocSegmentation(void)
{
  char fileName[SEGLOC_MAX_STRLEN];
  static SEGLOC_POS prevPos={-1};
  BOOL status;

  sick_max_range =  SICK_SEG_RHO_MAX;
  /* Getting a laser image from file */
  if (SDI_F->offlineParams.offlineOn) {
    sprintf(fileName, "%s.%03d", 
	    SDI_F->offlineParams.baseName, SDI_F->offlineParams.fileNum);    
    status = seglocReadLaserData(TRUE, fileName, &sickMeasures);
  }
  /* Getting a laser image from poster */
  else {
    status = seglocReadLaserData(FALSE, SDI_F->initParams.laserDataPosterName, 
				 &sickMeasures);
  }
  if (!status) {
    printf ("seglocSegmentation: read laser data failed\n");
    return FALSE;
  }

  /* Set segkit image */
  imageSegmentNum = setSegkitLaserImage(&sickMeasures, &sick_config);

  /* Record position and variances on SDI */
 
  SDI_F->sickLocoPos.pos.x =
    sickMeasures.header.sickPomPos.mainToBase.euler.x;
  SDI_F->sickLocoPos.pos.y = 
    sickMeasures.header.sickPomPos.mainToBase.euler.y;
  SDI_F->sickLocoPos.pos.theta = 
    sickMeasures.header.sickPomPos.mainToBase.euler.yaw;
  SDI_F->sickLocoPos.speed.v = cos(SDI_F->sickLocoPos.pos.theta) * 
			       sickMeasures.header.sickPomPos.VLocal.euler.x
			       + sin(SDI_F->sickLocoPos.pos.theta) * 
			       sickMeasures.header.sickPomPos.VLocal.euler.y;
  SDI_F->sickLocoPos.speed.w = sickMeasures.header.sickPomPos.VLocal.euler.yaw;
#ifdef VERBOSE
  printf("Sick loco pos \n %.2f %.2f %.2f\n",
	 SDI_F->sickLocoPos.pos.x,
	 SDI_F->sickLocoPos.pos.y,
	 SDI_F->sickLocoPos.pos.theta);
#endif
  /* Test if motion */
  if (fabs(SDI_F->sickLocoPos.pos.x - prevPos.x) < 0.01 &&
      fabs(SDI_F->sickLocoPos.pos.y - prevPos.y) < 0.01 &&
      fabs(angleLimit(SDI_F->sickLocoPos.pos.theta 
		      - prevPos.theta)) < 0.008 /* && 
fabs(SDI_F->sickLocoPos.speed.v) < 0.03 && */
/*       fabs(SDI_F->sickLocoPos.speed.w) < 0.03 */) {
    if (seglocNoRobotMotion == FALSE) seglocNoMotionJustStarted = TRUE;
    else seglocNoMotionJustStarted = FALSE;
    seglocNoRobotMotion = TRUE;
  }
  else {
    seglocNoMotionJustStarted = FALSE;
    seglocNoRobotMotion = FALSE;
    prevPos = SDI_F->sickLocoPos.pos;
  }
      
  SDI_F->seglocTimes.sickTime = sickMeasures.header.rcv_time;
  sickPomDate = sickMeasures.header.sickPomPos.date;
  SDI_F->sickLocoPos.var.vxx = sickMeasures.header.sickPomPos.mainToBase.var.vxx;
  SDI_F->sickLocoPos.var.vyy = sickMeasures.header.sickPomPos.mainToBase.var.vyy;
  SDI_F->sickLocoPos.var.vtt = sickMeasures.header.sickPomPos.mainToBase.var.voo;
  SDI_F->sickLocoPos.var.vxy = sickMeasures.header.sickPomPos.mainToBase.var.vxy;
  SDI_F->sickLocoPos.var.vxt = sickMeasures.header.sickPomPos.mainToBase.var.vox;
  SDI_F->sickLocoPos.var.vyt = sickMeasures.header.sickPomPos.mainToBase.var.voy;

  SDI_F->sickLocoUncertainties.ux = SEGLOC_3SIGMA(SDI_F->sickLocoPos.var.vxx);
  SDI_F->sickLocoUncertainties.uy = SEGLOC_3SIGMA(SDI_F->sickLocoPos.var.vyy);
  SDI_F->sickLocoUncertainties.utheta = SEGLOC_3SIGMA(SDI_F->sickLocoPos.var.vtt);


  /* Set Segmentation params */
  if (!conf_sg_host(SDI_F->segParams.leastSquareThres,
		    SDI_F->segParams.nbMinSegPts,
		    SDI_F->segParams.distMaxSegPts,
		    sick_max_range)) {
    printf ("seglocSegmentation: bad parameters for conf_sg_host\n");
    return FALSE;
  }

  /* Segmentation */
  if (imageSegmentNum<0) {
    printf ("imageSegmentNum before segmentation: %d\n", imageSegmentNum);
    return FALSE;
  }
  if (segmentation(PtoIm(imageSegmentNum),1) == -1) {
    printf ("segloc : segmentation failed\n");
    return FALSE;
  }
  if (imageSegmentNum<0)
    printf ("imageSegmentNum after segmentation: %d\n", imageSegmentNum);

  /* Save in poster */
  if (!seglocSgl2SegPoster(imageSegmentNum)) {
    printf ("seglocSegmentation: pb in seglocSgl2SegPoster\n");
    printf ("(imageSegmentNum = %d)\n", imageSegmentNum);
    return FALSE;
  }
  return TRUE;
}


/**
 **  seglocShowResults  -  Display results
 **
 **
 **/
static void seglocShowResults(int verbose, u_long  msecDuration)
{
#ifdef LOG_DATA
  fprintf (filelog, "%d %5.2f %5.2f %5.2f %d %5.2f %5.2f %5.2f %5.2f %5.2f\n",
	   msecDuration, 
	   SDI_F->sickSeglocPos.pos.x,
	   SDI_F->sickSeglocPos.pos.y,
	   SDI_F->sickSeglocPos.pos.theta,
	   sickMeasures.header.sickPomPos.date,
	   sickMeasures.header.sickPomPos.mainToOrigin.euler.x,
	   sickMeasures.header.sickPomPos.mainToOrigin.euler.y,
	   sickMeasures.header.sickPomPos.mainToOrigin.euler.yaw,
	   SDI_F->sickLocoPos.speed.v,
	   SDI_F->sickLocoPos.speed.w);

  printf ("DTHETA = %d\n", 
	  (int)((SDI_F->sickSeglocPos.pos.theta-sickMeasures.header.sickPomPos.mainToOrigin.euler.yaw)*180.0/3.1415926));
#endif

  if (!verbose)
    return;

  if (verbose == 1) {
    printf("time: %lu\n", EXEC_TASK_ON_PERIOD(0));
    return;
  }

  printf("%s %4d suc (no %2d/%-2d %2d match)  %lu ms (p %lu ms)\n", 
	 SDI_F->locReport.nbSuccesses > 0 ? "OK: " 
	 : SDI_F->locReport.nbSuccesses == 0 && 
	 SDI_F->locReport.nbPossibilities == 1? "corridor: "
	 : "? ",
	 SDI_F->locReport.nbSuccesses,
	 SDI_F->segkitBest,
	 SDI_F->locReport.nbPossibilities, 
	 SDI_F->locReport.nbMatches,
	 msecDuration,
	 EXEC_TASK_ON_PERIOD(0));

#if 1
  if (SDI_F->segkitBest != -1) {
    printf ("conf %f", SDI_F->locReport.confidence);
#if 0
    printf(" (");
    for (i=0; i<SEGLOC_NB_CONF_ATTRIBUTES; i++)
      printf ("%5.2f ", SDI_F->confidenceAttributes[i]);
#endif
    printf (")\n");
  }
#endif
  if (verbose == 2)
    return;
  
  printf("   Pos: %.2f %.2f %.2f",
	 SDI_F->currentSeglocPos.pos.x,
	 SDI_F->currentSeglocPos.pos.y,
	 SDI_F->currentSeglocPos.pos.theta);

  if (verbose == 3) {
    printf ("\n\n");
    return;
  }

  printf("   Unc: %.3f %.3f %.3f\n",
	 SDI_F->sickSeglocUncertainties.ux,
	 SDI_F->sickSeglocUncertainties.uy,
	 SDI_F->sickSeglocUncertainties.utheta);

  if (verbose == 4)
    return;

   printf("   tr: %.3f %.3f %.3f\n",
	 SDI_F->seglocTransf.pos.x,
	 SDI_F->seglocTransf.pos.y,
	 SDI_F->seglocTransf.pos.theta); 

  if (verbose == 5)
    return;

  printf("   Current Loco Pos :     %.3f %.3f %.3f\n",
	 SDI_F->currentLocoPos.pos.x,
	 SDI_F->currentLocoPos.pos.y,
	 SDI_F->currentLocoPos.pos.theta);

  if (verbose == 6)
    return;

  printf("   Corrected Loco Pos : %.3f %.3f %.3f\n",
	 SDI_F->currentSeglocPos.pos.x,
	 SDI_F->currentSeglocPos.pos.y,
	 SDI_F->currentSeglocPos.pos.theta);
}

/* ---------------------------------------------------------------------- *
 *                         FUNCTIONS FOR LOCALISATION
 */

/**
 **  seglocStartLocalisation - 
 **/

static BOOL seglocStartLocalisation(int *bilan, SEGLOC_LOC_MODE locMode)
{
  /* Is there a map loaded ? */
  if (!SDI_F->state.mapLoaded) {
    seglocLoadMapExec(&SDI_F->initParams.mapFile, bilan);
    if (*bilan != OK)
      return FALSE;
  }

  /* Localisation params */
  SetRelocThres(SDI_F->locParams[locMode].posDistMahaThres);

  SDI_F->state.locMode = locMode;
  SDI_F->state.localisationOn = TRUE;
  return TRUE;
}

/**
 **  seglocEndLocalisation - 
 **/

static void seglocEndLocalisation(int *bilan)
{
  SDI_F->state.localisationOn = FALSE;
  *bilan = SDI_F->state.bilan;
}


/**
 **  seglocMainLocalisation -   THE MAIN PROCEDURE
 **
 **
 **/

#define EPSILON_V 0.001
#define EPSILON_W 0.01

static BOOL seglocMainLocalisation(void)
{
  BOOL visibilityFilter;
  BOOL variancesFilter;
  long mapDistMahaThres;
  int solutionFound=1;
  SEGLOC_POS solpos;
  BOOL matchedSegs;
  int date;
#ifdef CLASSIFICATION
  char fileName[256];
  int newbest;
#endif
  SEGLOC_POS_VAR initPos;
  SEGLOC_LOC_MODE locMode;
  static SEGLOC_LOC_MODE prevLocMode = SEGLOC_LOC_LOCAL;
  static int startWait = TRUE;
  SEGLOC_LOC_PARAMS *locParamsPtr;
  int compassUsed;
  u_long msec0, msec1;
#ifndef __linux__
  struct timespec tp; /*  mesure du temps */

  clock_gettime(CLOCK_REALTIME, &tp);
  msec0 = (tp.tv_nsec / 1000000) + (tp.tv_sec * 1000);
#else
  struct timeval tp;
  
  gettimeofday(&tp, NULL);
  msec0 = (tp.tv_usec / 1000) + (tp.tv_sec * 1000);
#endif


  /** ---------------------------- INIT ---------------------------- **/

  /* Select parameters according to localisation mod */
  locMode = SDI_F->state.locMode;

  /* LOOP + NO MOTION => Nothing to be done */
  if (locMode == SEGLOC_LOC_LOOP && prevLocMode == SEGLOC_LOC_LOOP 
      && seglocNoRobotMotion) {
    if (startWait) {
      printf ("segloc: no motion\n");
      startWait = FALSE;
    }
    return TRUE;
  }
  prevLocMode = locMode;
  startWait = TRUE;

  locParamsPtr = &SDI_F->locParams[locMode];
  variancesFilter = locParamsPtr->variancesFilter;

  /* Filtering data segments with length too small */
  /* Perhaps it will be useful to control the following params :
     (see FilterData in Map_builder.c)
        o DATA_ANG : min angular aperture of a segment for loc purpose
        o min_length : min seg length */
  FilterOutData(imageSegmentNum);

  /* Records robot position stored in new data image in rob_enc (and Vrob_enc)
     global variable. Then recomputes Glob_pos (and Vpos) using 
     Loc2Abs (and Vloc2abs).
     If segkit is used to simulate the probabilistic odometry, 
     this function must be called every period. 
     Otherwise it is useless in LOC_LOCAL mode. */
  set_encoders(PtoIm(imageSegmentNum)->enc, PtoIm(imageSegmentNum)->var);

  /* Record the estimated position at sick acquisition in global frame */
  /* Not a good idea because it is just an estimation and not the final 
     result that we should record in this frame */
  /*   recordSickSeglocPos(); */

  /* C'EST LA ! Distance max */
  segkitSetAspectDistMax(20);

  /* Init localisation parameters: Reads compass.
     Init Glob_pos and Vpos in LOC_LOCAL. Filters on/off ... */
  if (!seglocInitLocalisation(locMode, &initPos, &compassUsed))
    return FALSE;

  /** ------------------------ LOCALIZATION ---------------------------- **/

  /* Localisation */
  visibilityFilter = locParamsPtr->visibilityFilter;
  SDI_F->segkitBest = Local_localization(imageSegmentNum, 
					 initPos.pos.x, initPos.pos.y, 
					 transformsArray,
					 visibilityFilter, 
					 NB_MAX_POSITIONS,
					 &sick_config);

/*   printf("VPOS %f %f %f\n", Vpos[0], Vpos[2], Vpos[5]); */

  /** -------------------------- ANALYSES ----------------------------- **/

  /* Reports */
  SDI_F->locReport.nbPossibilities = GetNpossibilities();
  SDI_F->locReport.nbSuccesses = GetNSuccess();
  solutionFound = SDI_F->locReport.nbPossibilities > 0 ? 1 : 0;

  /* XXX TEST DEBUG */
  if (solutionFound && seglogTestDebug) {
    int best = SDI_F->segkitBest;
    double dx, dy, dth;

    if (best == -1) {
      printf ("WARNING: best == -1 !\n");
      solutionFound = 0;
      SDI_F->locReport.nbPossibilities = 0;
    }
    else {
      GetPositionArray(positionsArray, NB_MAX_POSITIONS);
      
      SDI_F->locReport.nbMatches = positionsArray[best*4+4];
      
      dx = positionsArray[best*4 + 1]/1000.0-Glob_pos[0];
      dy = positionsArray[best*4 + 2]/1000.0-Glob_pos[1];
      dth = positionsArray[best*4 + 3]/1000.0-Glob_pos[2];
      
      if (fabs(dx) > 0.001 || fabs(dy)  > 0.001 || fabs(dth) > 0.001) {
	printf ("WARNING: Glob_pos != positionsArray[best]\n");
	printf("          position: %.3f %.3f %.3f\n",
	       positionsArray[best*4 + 1]/1000.0,
	       positionsArray[best*4 + 2]/1000.0,
	       positionsArray[best*4 + 3]/1000.0);
	printf("          Diff: %f %f %f\n", dx, dy,dth);
      }
    }
  }

  /** ---------------------------- CORRIDOR ---------------------------- **/

  /* If failed retry using current variance of robot to match 
     segments one by one */
  /* Ne le faire que si on est SUR que le robot n'est pas trop perdu
     (car trouve toujours des segments pour se recaler) */
  if (locMode != SEGLOC_LOC_LOCAL && SDI_F->locReport.nbPossibilities == 0 
      && SDI_F->locParams[locMode].corridorModeAllowed) {


    /* ATTENTION : on force la position */
    if (locMode == SEGLOC_FORCE_LOC_LOCAL) {
      seglocChangeSegkitPos(&initPos);
    }

    SDI_F->segkitBest = -1;
    printf ("Corridor...");

    /* Cherche des segments pour se recaler */
    mapDistMahaThres = locParamsPtr->mapDistMahaThres;
    if (MapTrData(PtoIm(imageSegmentNum), Glob_pos, Vpos, mapDistMahaThres) 
	== NULL) {
      printf ("failed\n");
    }

    /* On a trouv */
    else {
      /* Test que ca ne fait pas de saut */
      solpos.x = Glob_pos[0];
      solpos.y = Glob_pos[1];
      solpos.theta = Glob_pos[2];
      if (!seglocTestJump(&initPos, &solpos)) {
	printf ("jumped\n");
	/* Solution rejete: on en revient  la position prcdente */
	seglocChangeSegkitPos(&initPos);
	SDI_F->state.bilan = S_segloc_BIG_JUMP;
	SDI_F->locReport.nbSuccesses = -1;
	solutionFound = 0;
      }
      
      /* Corridor  OK */
      else {

	/* Re-initializing Glob_pos, recomputes Loc2Abs */
	init_location_st(Glob_pos,Vpos);
      
	/* On a un truc utilisable */
	SDI_F->locReport.nbPossibilities = 1;
	solutionFound  = 1;
      }
    }
  }

  /** ---------------------------- TESTS ---------------------------- **/

  /* Test les sauts si n'est deja fait */
  if (solutionFound /* && !variancesFilter */) {
    solpos.x = Glob_pos[0];
    solpos.y = Glob_pos[1];
    solpos.theta = Glob_pos[2];
#ifdef DEBUG
    seglocPrintJump(&initPos, &solpos);
#endif
    if (!seglocTestJump(&initPos, &solpos)) {
      SDI_F->state.bilan = S_segloc_BIG_JUMP;
      solutionFound = 0;
      SDI_F->locReport.nbSuccesses = -1;
      /* Solution rejete: on en revient  la position prcdente */
      seglocChangeSegkitPos(&initPos);
    }
  }

  /** ----------------------- RECORDS RESULTS -------------------------- **/

  /* On remet toujours a jour currentPos ... */
  if (solutionFound) {
    SDI_F->state.bilan = OK;
    /* This is very important : records the date at which the sick acquisition occurs */
    sickPomDate_at_loc = sickPomDate;
  }

  if (solutionFound || locMode == SEGLOC_FORCE_LOC_LOCAL) {
    seglocRecordSegkitTransf();
  }
  if (!seglocGetCurrentOdoPos(&SDI_F->currentLocoPos, &date)) {
    SDI_F->state.bilan = S_segloc_POSTER_NOT_FOUND;
/*     return FALSE; */
  }
  if (solutionFound) {
    seglocPomExports();
  }
 
  /* Caracterisation de la qualit de la loc */
  seglocLocQuality();

  /* Bilan compas */
  if (solutionFound && locMode == SEGLOC_LOC_LOCAL && compassUsed) {
    seglocRecordCompas();
  }

  /* Show results */
#ifndef __linux__
  clock_gettime(CLOCK_REALTIME, &tp);
  msec1 = (tp.tv_nsec / 1000000) + (tp.tv_sec * 1000);
#else
  gettimeofday(&tp, NULL);
  msec1 = (tp.tv_usec / 1000) + (tp.tv_sec * 1000);
#endif /* ! __linux__ */
  seglocShowResults(seglocVerbose, msec1-msec0);


  /** ----------------------- EXPORTS DATA  -------------------------- **/

  /* Aspect map for debug */
#ifndef CLASSIFICATION
  matchedSegs = SDI_F->locReport.nbSuccesses > 0 ? TRUE : FALSE;
  if (SDI_F->segkitBest != -1) {
    seglocSgl2SegAspectPoster(Get_aspect_num(), matchedSegs,
			      &SDI_F->sickSeglocPos);
  }

#ifdef POSTER_MATCHED
   seglocSgl2MatchedSegPoster(imageSegmentNum, Get_aspect_num(),
			       &SDI_F->sickSeglocPos);
#endif

#else    
  /* CLASSIFICATION */
  if (SDI_F->offlineParams.setSegmentationOff) {

    /* Offline: display and classsify results */
    sprintf(fileName, "%s-segloc.%03d", 
	    SDI_F->offlineParams.baseName, SDI_F->offlineParams.fileNum);    
    if (!seglocSaveResults(fileName, imageSegmentNum, &SDI_F->segParams, 
			   locMode, &SDI_F->locParams[locMode], 
			   &initPos, &sickMeasures.header,
			   SDI_F->locReport.nbPossibilities, 
			   SDI_F->segkitBest, &newbest)) 
      printf ("Cannot save results in file %s\n", fileName);

    if (newbest != -1) {
      printf ("Need to change init pos\n");
      seglocChangeSegkitPos(&initPos);
    }
  } 
  /* RUN WITHOUT CLASSIFICATION */
  else {
    matchedSegs = SDI_F->locReport.nbSuccesses > 0 ? TRUE : FALSE;
    seglocSgl2SegAspectPoster(Get_aspect_num(), matchedSegs,
			      &SDI_F->sickSeglocPos);
  }
#endif /* CLASSIFICATION */

  return TRUE;
}

/**
 **  seglocInitLocalisation   -- Init localization parameters.
 ** 
 **  Reads compass.
 **  Computes Glob_pos in LOC_LOCAL mode. 
 **  Activates/desactivates filters (visibility, variance) and odo merging.
 **  
 **/

static BOOL seglocInitLocalisation(SEGLOC_LOC_MODE locMode, 
				   SEGLOC_POS_VAR *initPos,
				   BOOL *usingCompass)
{
  int compassDataValid;
  SEGLOC_COMPASS_RES *cmpRes=&SDI_F->compass.compassRes;
  SEGLOC_LOC_PARAMS *locParamsPtr;
  double angleUncertainty;

  /* Select parameters according to localisation mod */
  locParamsPtr = &SDI_F->locParams[locMode];

  /* Lecture du compas */
  compassDataValid = seglocUpdateCompass();

  /* Utilisation du compas pour l'estime de la position */
  if (compassDataValid &&
      locParamsPtr->useCompass == SEGLOC_TRUE &&
      !SDI_F->offlineParams.offlineOn) {
    SDI_F->initPos.pos.theta = 
      angleLimit(DEG_TO_RAD(cmpRes->correctedAngleDeg));
    SDI_F->initPos.unc.utheta = 
      angleLimit(DEG_TO_RAD(cmpRes->angleUncDeg));
    *usingCompass = TRUE;
  }
  else {
    *usingCompass = FALSE;
  }

  /* Init estimated pos */
  switch(locMode) {
    case SEGLOC_LOC_LOCAL:
    case SEGLOC_FORCE_LOC_LOCAL:

      MergeOdoOff();

      initPos->pos = SDI_F->initPos.pos;

      if (SDI_F->initPos.unc.ux==0.0)
	SDI_F->initPos.unc.ux=1000;
      if (SDI_F->initPos.unc.uy==0.0)
	SDI_F->initPos.unc.uy=1000;
      if (SDI_F->initPos.unc.utheta==0.0)
	SDI_F->initPos.unc.utheta=3.15;
      initPos->var.vxx = SEGLOC_VARIANCE(SDI_F->initPos.unc.ux);
      initPos->var.vyy = SEGLOC_VARIANCE(SDI_F->initPos.unc.uy);
      initPos->var.vtt = SEGLOC_VARIANCE(SDI_F->initPos.unc.utheta);
      initPos->var.vxy = 0.0;
      initPos->var.vxt = 0.0;
      initPos->var.vyt = 0.0;
      angleUncertainty = SDI_F->initPos.unc.utheta;

      /* Updates Glob_pos and re-computes Loc2Abs using rob_enc  */
      seglocChangeSegkitPos(initPos);

      break;

    case SEGLOC_LOC_ONCE:
    case SEGLOC_LOC_LOOP:

      MergeOdoOn();

      initPos->pos.x = Glob_pos[0];
      initPos->pos.y = Glob_pos[1];
      initPos->pos.theta = Glob_pos[2];
      /* XXX TRACKING MOINS LISSE MAIS + DE SOLUTIONS */
      if (SEGLOC_3SIGMA(Vpos[0]) < 0.1)
	Vpos[0] = SEGLOC_VARIANCE(0.1);
      if (SEGLOC_3SIGMA(Vpos[2]) < 0.1)
	Vpos[2] = SEGLOC_VARIANCE(0.1);
      if (SEGLOC_3SIGMA(Vpos[5]) < 0.1)
	Vpos[5] = SEGLOC_VARIANCE(0.1);
      initPos->var.vxx = Vpos[0];
      initPos->var.vyy = Vpos[2];
      initPos->var.vtt = Vpos[5];
      initPos->var.vxy = Vpos[1];
      initPos->var.vxt = Vpos[3];
      initPos->var.vyt = Vpos[4];
      angleUncertainty = locParamsPtr->angleUncertainty;
      /* XXX    angleUncertainty = SEGLOC_3SIGMA(Vpos[5]); */
      break;

    default:
      printf ("seglocLocalisation error: unknown mode %d\n", locMode);
      SDI_F->state.bilan = S_segloc_FAILED;
      return FALSE;
    }

#ifdef CLASSIFICATION
  MergeOdoOff();
#endif

  /* What we know about the angle */
  Set_loc_angle(RAD_TO_DEG(initPos->pos.theta), RAD_TO_DEG(angleUncertainty));

  /* Filter for the find out solutions using Glob_pos and Vpos */
  if (locParamsPtr->variancesFilter) {
    set_pos_jump_filter_fact(locParamsPtr->factUncertaintyTest);
    LocFilterOn();
  }
  else
    LocFilterOff();
  
  if (seglocVerbose > 3) {
    printf("--- INIT: %.2f %.2f %.2f",
	   initPos->pos.x, initPos->pos.y, initPos->pos.theta);
    printf("  Unc: %.3f %.3f %.3f (%.3f)",
	   SEGLOC_3SIGMA(initPos->var.vxx),
	   SEGLOC_3SIGMA(initPos->var.vyy), angleUncertainty,
	   SEGLOC_3SIGMA(initPos->var.vtt));
    printf ("(mode %d)\n", SDI_F->state.locMode);
  }
  return TRUE;
}


/**
 **  seglocLocQuality   -- analysis loc quality according to nbSuccesses, 
 **                        nbPossibilities, speed, uncertainties, ...
 ** 
 **                        updates : SDI_F->locReport.quality
 **                        and SDI_F->state.bilan
 **                        and SDI_F->stat
 **/
static void seglocLocQuality()
{
  BOOL motionFlag;
  SEGLOC_BOOL seglocConfidenceFlags[SEGLOC_NB_CONF_ATTRIBUTES] 
    = SEGLOC_DEFAULT_CONFIDENCE_FLAGS ;

  /* XXX pourrait aussi utiliser le nb de segments matches et 
     les longueurs matches - pas accessible directement ) */
  motionFlag = fabs(SDI_F->sickLocoPos.speed.v) > EPSILON_V ||
	       fabs(SDI_F->sickLocoPos.speed.w) > EPSILON_W;
  SDI_F->locReport.confidence = 0;

  if (SDI_F->locReport.nbSuccesses > 0) {
    SDI_F->locReport.quality = SEGLOC_LOC_QUALITY_OK;
    if (motionFlag) {
      SDI_F->stat.nbLocOK++;
    }

    if (SDI_F->segkitBest != -1)
      SDI_F->locReport.confidence 
	= seglocComputesConfidence(imageSegmentNum,
				   Glob_pos,
				   SDI_F->sickLocoPos.speed.w,
				   seglocConfidenceFlags,
				   SDI_F->confidenceAttributes);

  }
  else if (SDI_F->locReport.nbSuccesses == 0 &&
	   SDI_F->locReport.nbPossibilities == 1) {
    SDI_F->state.bilan = S_segloc_PARTIAL_LOC;
    SDI_F->locReport.quality = SEGLOC_LOC_QUALITY_MEDIUM;
    if (motionFlag) {
      SDI_F->stat.nbPartialLoc++;
    }
  }
  else {
    if (SDI_F->state.bilan != S_segloc_BIG_JUMP)
      SDI_F->state.bilan = S_segloc_LOC_FAILED;
    if (motionFlag) {
      SDI_F->stat.nbLocFailed++;
    }
    if (SDI_F->sickSeglocUncertainties.ux > SEGLOC_X_UNCERTAINTY_THRESHOLD ||
	SDI_F->sickSeglocUncertainties.uy > SEGLOC_Y_UNCERTAINTY_THRESHOLD ||
	SDI_F->sickSeglocUncertainties.utheta > 
	SEGLOC_THETA_UNCERTAINTY_THRESHOLD) {
      SDI_F->locReport.quality = SEGLOC_LOC_QUALITY_KO;
      if (motionFlag) {
	SDI_F->stat.nbLocCrashed++;
      }
    }
    else 
      SDI_F->locReport.quality = SEGLOC_LOC_QUALITY_BAD;
  }
  
  /*   if (SDI_F->locReport.nbPossibilities > 1) */
  /*     SDI_F->state.bilan = S_segloc_SEVERAL_SOLUTIONS; */
}  

/**
 **  seglocChangeSegkitPos  -- 
 **
 **  Remet a jour Glob_pos et sa variance Vpos,
 **  puis recalcul Loc2Abs et sa variance Vloc2abs  partir
 **  des dernires donnes odomtriques stockes dans rob_enc et Vrob_enc.
 **
 **
 **/

static void seglocChangeSegkitPos(SEGLOC_POS_VAR *initPos)
{
  double newPos[3];
  double newVPos[6];
  newPos[0] = initPos->pos.x;
  newPos[1] = initPos->pos.y;
  newPos[2] = initPos->pos.theta;
  newVPos[0] = initPos->var.vxx;
  newVPos[1] = initPos->var.vxy;
  newVPos[2] = initPos->var.vyy;
  newVPos[3] = initPos->var.vxt;
  newVPos[4] = initPos->var.vyt;
  newVPos[5] = initPos->var.vtt;
  /* Remet a jour Glob_pos, Loc2Abs ... */
  init_location_st(newPos, newVPos);
}

static void seglocRecordSegkitTransf()
{
  /* Record new transf */
  SDI_F->seglocTransf.pos.x     = Loc2Abs[0];
  SDI_F->seglocTransf.pos.y     = Loc2Abs[1];
  SDI_F->seglocTransf.pos.theta = angleLimit(Loc2Abs[2]);
  
  /* ... and its variances */
  SDI_F->seglocTransf.var.vxx = Vloc2abs[0];
  SDI_F->seglocTransf.var.vxy = Vloc2abs[1];
  SDI_F->seglocTransf.var.vyy = Vloc2abs[2];
  SDI_F->seglocTransf.var.vxt = Vloc2abs[3];
  SDI_F->seglocTransf.var.vyt = Vloc2abs[4];
  SDI_F->seglocTransf.var.vtt = Vloc2abs[5];
  
  /* Record corrected new pos */
  SDI_F->sickSeglocPos.pos.x     = Glob_pos[0];
  SDI_F->sickSeglocPos.pos.y     = Glob_pos[1];
  SDI_F->sickSeglocPos.pos.theta = Glob_pos[2];

  /* Get variances */
  SDI_F->sickSeglocPos.var.vxx = Vpos[0];
  SDI_F->sickSeglocPos.var.vxy = Vpos[1];
  SDI_F->sickSeglocPos.var.vyy = Vpos[2];
  SDI_F->sickSeglocPos.var.vxt = Vpos[3];
  SDI_F->sickSeglocPos.var.vyt = Vpos[4];
  SDI_F->sickSeglocPos.var.vtt = Vpos[5];

  SDI_F->sickSeglocUncertainties.ux = SEGLOC_3SIGMA(Vpos[0]);
  SDI_F->sickSeglocUncertainties.uy = SEGLOC_3SIGMA(Vpos[2]);
  SDI_F->sickSeglocUncertainties.utheta = SEGLOC_3SIGMA(Vpos[5]);


/*   printf ("REC: v0 %f u0 %f / v2 %f u2 %f / v5 %f u5 %f \n", */
/* 	  Vpos[0], SDI_F->sickSeglocUncertainties.ux, */
/* 	  Vpos[2], SDI_F->sickSeglocUncertainties.uy, */
/* 	  Vpos[5], SDI_F->sickSeglocUncertainties.utheta); */


  printf("  transf    %5.2f %5.2f %5.2f \n"
	 "  locoPos   %5.2f %5.2f %5.2f\n"
	 "  seglocPos %5.2f %5.2f %5.2f\n",
	 
	 SDI_F->seglocTransf.pos.x ,
	 SDI_F->seglocTransf.pos.y,
	 SDI_F->seglocTransf.pos.theta,
	 
	 SDI_F->currentLocoPos.pos.x,
	 SDI_F->currentLocoPos.pos.y,
	 SDI_F->currentLocoPos.pos.theta,

	 SDI_F->currentSeglocPos.pos.x,
	 SDI_F->currentSeglocPos.pos.y,
	 SDI_F->currentSeglocPos.pos.theta
	 );

  /** Exports it into the pom poster only and only if :
      - pomposter is defined
      - motion is being done OR it is the first cycle of a "No motion" phase */
  
  
}

static void recordSickSeglocPos()
{ 
  /* Record mainToOrigin at sick acquisition time
     (ie, the position of the robot in global frame while it 
     was acquiring the data) */
  SDI_F->sickSeglocPos.pos.x     = Glob_pos[0];
  SDI_F->sickSeglocPos.pos.y     = Glob_pos[1];
  SDI_F->sickSeglocPos.pos.theta = Glob_pos[2];

  /* Get variances */
  SDI_F->sickSeglocPos.var.vxx = Vpos[0];
  SDI_F->sickSeglocPos.var.vxy = Vpos[1];
  SDI_F->sickSeglocPos.var.vyy = Vpos[2];
  SDI_F->sickSeglocPos.var.vxt = Vpos[3];
  SDI_F->sickSeglocPos.var.vyt = Vpos[4];
  SDI_F->sickSeglocPos.var.vtt = Vpos[5];

  SDI_F->sickSeglocUncertainties.ux = SEGLOC_3SIGMA(Vpos[0]);
  SDI_F->sickSeglocUncertainties.uy = SEGLOC_3SIGMA(Vpos[2]);
  SDI_F->sickSeglocUncertainties.utheta = SEGLOC_3SIGMA(Vpos[5]);
}


/**
 ** seglocNewPosValidityTest  -  
 **
 ** Returns TRUE if the new computed position is compatible with 
 ** the variances of the old one.
 **
 ** This test should not be necessary if locParams[i].variancesFilter is TRUE
 ** and if locParams[i].posDistMahaThres is correctly set .
 **/ 

#ifdef DEBUG
static BOOL seglocPrintJump(SEGLOC_POS_VAR *old, 
			    SEGLOC_POS *new)
{
  double ecx, ecy, ect;
  double ux, uy, ut;

  if (old->var.vxx <= 0.0 ||
      old->var.vyy <= 0.0 ||
      old->var.vtt <= 0.0) {
    printf ("   seglocPrintJump:  variances negatives ou nulles (%.3f %.3f %.3f) refused.\n",
	    old->var.vxx, old->var.vyy, old->var.vtt);
  }

  ux = SEGLOC_3SIGMA(old->var.vxx);
  uy = SEGLOC_3SIGMA(old->var.vyy);
  ut = SEGLOC_3SIGMA(old->var.vtt);

  ecx = new->x - old->pos.x;
  ecy = new->y - old->pos.y;
  ect = angleLimit(new->theta - old->pos.theta);

  printf ("---------- seglocPrintJump:  \n"
	  "old (%.3f %.3f %.3f)\n"
	  "new (%.3f %.3f %.3f)\n"
	  "Jump = (%.3f %.3f %.3f)\n",
	  old->pos.x, old->pos.y, old->pos.theta,
	  new->x, new->y, new->theta,
	  ecx, ecy, ect);

  return TRUE;
}
#endif /* DEBUG */

static BOOL seglocTestJump(SEGLOC_POS_VAR *old, 
			   SEGLOC_POS *new)
{
  double ecx, ecy, ect;
  double ux, uy, ut;

  if (old->var.vxx <= 0.0 ||
      old->var.vyy <= 0.0 ||
      old->var.vtt <= 0.0) {
    printf ("   seglocTestJump:  variances negatives ou nulles (%.3f %.3f %.3f) refused.\n",
	    old->var.vxx, old->var.vyy, old->var.vtt);
    return FALSE;
  }

  ux = SEGLOC_3SIGMA(old->var.vxx);
  uy = SEGLOC_3SIGMA(old->var.vyy);
  ut = SEGLOC_3SIGMA(old->var.vtt);

  ecx = new->x - old->pos.x;
  ecy = new->y - old->pos.y;
  ect = angleLimit(new->theta - old->pos.theta);

  if ( fabs(ecx) > ux ||
       fabs(ecy) > uy ||
       fabs(ect) > ut) {
    printf ("   seglocTestJump:  pos (%.3f %.3f %.3f) refused. Jump = (%.3f %.3f %.3f)\n",
	    new->x, new->y, new->theta,
	    ecx, ecy, ect);
    return FALSE;
  }

  return TRUE;
}

/*------------------------------------------------------------------------
 * SurvLocKO
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_segloc_BAD_PARAMS
 */

/* seglocSurvLocExec  -  codel EXEC de SurvLoc
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT seglocSurvLocKOExec (SEGLOC_UNC *survUncertainty, 
				    SEGLOC_SURV_LOC_REPORT *survLocReport, 
				    int *bilan)
{
#if 0
  if (SDI_F->sickSeglocUncertainties.ux > survUncertainty->ux) {
    *survLocReport = SEGLOC_X_OUT;
    return ETHER;
  }
  if (SDI_F->sickSeglocUncertainties.uy > survUncertainty->uy) {
    *survLocReport = SEGLOC_Y_OUT;
    return ETHER;
  }
#endif
  if (SDI_F->sickSeglocUncertainties.ux > survUncertainty->ux &&
      SDI_F->sickSeglocUncertainties.uy > survUncertainty->uy) {
    *survLocReport = SEGLOC_X_OR_Y_OUT;
    return ETHER;
  }

  if (SDI_F->sickSeglocUncertainties.utheta > survUncertainty->utheta) {
    *survLocReport = SEGLOC_THETA_OUT;
    return ETHER;
  }
  return EXEC;
}


/*------------------------------------------------------------------------
 * SurvLocOK
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_segloc_BAD_PARAMS
 */

/* seglocSurvLocExec  -  codel EXEC de SurvLoc
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT seglocSurvLocOKExec (SEGLOC_UNC *survUncertainty,
				    int *bilan)
{
  if (SDI_F->sickSeglocUncertainties.ux < survUncertainty->ux
      && SDI_F->sickSeglocUncertainties.uy < survUncertainty->uy
      && SDI_F->sickSeglocUncertainties.utheta < survUncertainty->utheta) {
    return ETHER;
  }
  return EXEC;
}



/*------------------------------------------------------------------------
 * OfflineMode
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_segloc_FILE_NOT_FOUND
 */

/* seglocOfflineModeExec  -  codel EXEC de OfflineMode
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT seglocOfflineModeExec (SEGLOC_OFFLINE *offlineParams, int *bilan)
{
  char fileName[SEGLOC_MAX_STRLEN];
  FILE *fd;
  double rob_pos[4]= {1.0,0.0,0.0,0.0}; 

  /* Switch en mode offline */
  if (offlineParams->offlineOn) {

    /* Test acces fichier */
    sprintf(fileName, "%s.%03d", offlineParams->baseName, offlineParams->fileNum); 
    fd = fopen(fileName, "r");
    if (fd == NULL) {
      printf ("seglocOfflineModeExec: cannot open %s\n", fileName);
      *bilan = S_segloc_FILE_NOT_FOUND;
      return ETHER;
    }
    fclose(fd);

    if (offlineParams->setSegmentationOff)
      SDI_F->state.segmentationOn = FALSE;
  }
  /* Switch en mode normal */
  else {
    GetRobotEncoders(rob_pos,RobId);
  }
  return ETHER;
}


/**
 ** COMPASS
 **
 **/
static BOOL seglocUpdateCompass(void)
{
#ifdef XR4000
  XR4000_COMPASS_STR compass;
  SEGLOC_COMPASS_ENV *cmpEnv=&SDI_F->compass.compassEnv;
  SEGLOC_COMPASS_RES *cmpRes=&SDI_F->compass.compassRes;

  /* Lecture du compas */
  if (xr4000RobotCompassPosterRead(&compass) != OK) {
    printf ("seglocUpdateCompass: cannot read compass\n");
    return FALSE;
  }
  
  if (!compass.dataValid) {
    printf ("seglocUpdateCompass: compass not valid\n");
    return FALSE;
  }
  
  cmpRes->rawAngleDeg = RAD_TO_DEG(angleLimit(compass.heading));
  cmpRes->correctedAngleDeg = -cmpRes->rawAngleDeg + 
			      cmpEnv->environmentNorthDeg;
  cmpRes->angleUncDeg = RAD_TO_DEG(compass.precision) +
			cmpEnv->environmentAccuracyDeg;

#if 0
  printf ("compass: raw %5.2f deg   cor %5.2f deg = %5.2f rad\n", 
	  cmpRes->rawAngleDeg, cmpRes->correctedAngleDeg,
	  DEG_TO_RAD(cmpRes->correctedAngleDeg));
#endif
  return TRUE;
#else 
  return FALSE;
#endif
}


static void seglocRecordCompas(void)
{
  SEGLOC_COMPASS_RES *cmpRes=&SDI_F->compass.compassRes;
  double deltaCompassDeg;
  int compassDataValid;

  /* On relit le compas */
  compassDataValid = seglocUpdateCompass();

  if (!compassDataValid) {
    printf ("seglocRecordCompas: compass not valid \n");
    return;
  }

  if (cmpRes->nbData == SEGLOC_NB_MAX_COMPASS) {
    printf ("seglocRecordCompas: too many data\n");
    return;
  }

  /* Stockage donne */
  SDI_F->deltaCompassTab[cmpRes->nbData] = 
    angleLimit(SDI_F->currentSeglocPos.pos.theta - 
	       DEG_TO_RAD(cmpRes->correctedAngleDeg));

  deltaCompassDeg = RAD_TO_DEG(SDI_F->deltaCompassTab[cmpRes->nbData]);
  printf ("deltaCompassDeg %5.2f\n", deltaCompassDeg);


#define  AVERAGE(val,nb,new) (((val)*(nb)+(new))/(nb+1))

  /* Moyenne des erreurs */
  cmpRes->deltaAngleAverageDeg = 
    RAD_TO_DEG(angleLimit(AVERAGE(DEG_TO_RAD(cmpRes->deltaAngleAverageDeg),
				  cmpRes->nbData,
				  SDI_F->deltaCompassTab[cmpRes->nbData])));


  /* Moyenne des carrs des erreurs */
#if 0  /* XXXXXX C'est quoi ca ? Oeuvre inacheve ? */
  cmpRes->deltaAngleSquareAverageRad = 
    angleLimit(AVERAGE(cmpRes->deltaAngleSquareAverageRad,
		       cmpRes->nbData,
		       SDI_F->deltaCompassTab[cmpRes->nbData]));


  /* Variance */
  cmpRes->deltaAngle3sigmaDeg = RAD_TO_DEG(3*sqrt(cmpRes->deltaAngleSquareAverageRad));

#endif

  /* MIN et MAX */
  if (deltaCompassDeg < cmpRes->deltaAngleMinDeg)
    cmpRes->deltaAngleMinDeg = deltaCompassDeg;
  if (deltaCompassDeg > cmpRes->deltaAngleMaxDeg)
    cmpRes->deltaAngleMaxDeg = deltaCompassDeg;


  (cmpRes->nbData)++;
  
  seglocMiscCompassPosterWrite(SEGLOC_MISC_POSTER_ID, &SDI_F->compass);
}

/*------------------------------------------------------------------------
 * LocDynamicMapLoop
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_segloc_LOC_FAILED
 *              S_segloc_PARTIAL_LOC
 *              S_segloc_SEVERAL_SOLUTIONS
 *              S_segloc_FAILED
 */

/* seglocLocDynamicMapStart  -  codel START de LocDynamicMapLoop
   Retourne:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT seglocLocDynamicMapStart (int *bilan)
{
  if (!seglocStartLocalisation(bilan, SEGLOC_LOC_LOOP))
    return ETHER;
  return ETHER;
}

/* seglocLocDynamicMapExec  -  codel EXEC de LocDynamicMapLoop
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT seglocLocDynamicMapExec (int *bilan)
{
  igla *map;
#define NB_PTS 4
#define DIST_FRONT 30
#define DIST_SIDE  30
#define DIST_BACK   2
  const double polyLocal[2*NB_PTS] = {DIST_FRONT,  DIST_SIDE,
				      DIST_FRONT, -DIST_SIDE,
				      DIST_BACK,  -DIST_SIDE,
				      DIST_BACK,   DIST_SIDE};
  double poly[2*4];
  int i;
  double xR, yR, thetaR, cR, sR;

  /* if loc OK add new segments */
  seglocAddSegsInMapExec(bilan);

  /* remove segments behind */
  map = GetCurrentMapAddr();
  if (map == NULL) {
    printf ("seglocClipMapExec: empty global map\n");
    return ETHER;
  }

  xR = SDI_F->currentSeglocPos.pos.x;
  yR = SDI_F->currentSeglocPos.pos.y;
  thetaR = SDI_F->currentSeglocPos.pos.theta;
  sR = sin(thetaR);
  cR = cos(thetaR);

  /* CALCUL POLY */
  for (i=0; i<NB_PTS; i+=2) {
    poly[i]   = xR + polyLocal[i] * cR - polyLocal[i+1] * sR;
    poly[i+1] = yR + polyLocal[i] * sR + polyLocal[i+1] * cR;
  }

  /* clip */
  seglocPolygonClipMap(map, NB_PTS, poly);
  
  /* Ca semble important .... (cf RetrieveMap) */
  UpdateLocalMap();

  /* Enregistrement de la carte dans le poster */
  if(!seglocSgl2MapPoster()) {
    printf ("seglocLoadMapExec: record map poster error\n");
    *bilan = S_segloc_FAILED;
    return ETHER;
  }


  return ETHER;
}

/* seglocLocDynamicMapInter  -  codel INTER de LocDynamicMapLoop
   Retourne:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT seglocLocDynamicMapInter (int *bilan)
{
  seglocEndLocalisation(bilan);
  return ETHER;
}

/*------------------------------------------------------------------------
 * RemoveSegment
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_segloc_UNKNOWN_SEG_ID
 *              S_segloc_SEG_NOT_FOUND
 *              S_segloc_BAD_MAP_ID
 *              S_segloc_ERROR
 */

/* seglocRemoveSegmentExec  -  codel EXEC de RemoveSegment
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT seglocRemoveSegmentExec (SEGLOC_SEG *seg, int *bilan)
{
  /* Remove segment */
  if (!RemoveSegment(seg->x1,seg->y1,seg->x2,seg->y2)) {
    *bilan = S_segloc_SEG_NOT_FOUND;
    return ETHER;
  }

  /* Record map in poster */
  if(!seglocSgl2MapPoster()) {
    printf ("seglocRemoveSegmentExec: record map poster error\n");
    *bilan = S_segloc_FAILED;
    return ETHER;
  }

  return ETHER;
}

/*------------------------------------------------------------------------
 * MergeSegments
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_segloc_UNKNOWN_SEG_ID
 *              S_segloc_BAD_MAP_ID
 *              S_segloc_ERROR
 */

/* seglocMergeSegmentExec  -  codel EXEC de MergeSegments
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT seglocMergeSegmentExec (SEGLOC_2SEGS *segs, int *bilan)
{
  double segsTab[9];

  /* Record the number of segements */
  segsTab[0] = 2;

  /* Get segment 1 */
  segsTab[1] = segs->seg1.x1;
  segsTab[2] = segs->seg1.y1;
  segsTab[3] = segs->seg1.x2;
  segsTab[4] = segs->seg1.y2;

  /* Get segment 2 */
  segsTab[5] = segs->seg2.x1;
  segsTab[6] = segs->seg2.y1;
  segsTab[7] = segs->seg2.x2;
  segsTab[8] = segs->seg2.y2;

  /* Remove segment */
  if (!MergeSegments(segsTab)) {
    *bilan = S_segloc_SEG_NOT_FOUND;
    return ETHER;
  }

  /* Record map in poster */
  if(!seglocSgl2MapPoster()) {
    printf ("seglocRemoveSegmentExec: record map poster error\n");
    *bilan = S_segloc_FAILED;
    return ETHER;
  }

  return ETHER;
}

/*------------------------------------------------------------------------
 * IntersectSegments
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_segloc_UNKNOWN_SEG_ID
 *              S_segloc_BAD_MAP_ID
 *              S_segloc_ERROR
 */

/* seglocIntersectSegmentExec  -  codel EXEC de IntersectSegments
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT seglocIntersectSegmentExec (SEGLOC_2SEGS *segs, int *bilan)
{

  /* Intersect segments */
  if (!IntersectsSegmentsD(segs->seg1.x1, segs->seg1.y1,
			   segs->seg1.x2, segs->seg1.y2,
			   segs->seg2.x1, segs->seg2.y1,
			   segs->seg2.x2, segs->seg2.y2)) {
    *bilan = S_segloc_SEG_NOT_FOUND;
    return ETHER;
  }

  /* Record map in poster */
  if(!seglocSgl2MapPoster()) {
    printf ("seglocRemoveSegmentExec: record map poster error\n");
    *bilan = S_segloc_FAILED;
    return ETHER;
  }

  return ETHER;
}

/*------------------------------------------------------------------------
 * CutSegment
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_segloc_UNKNOWN_SEG_ID
 *              S_segloc_BAD_MAP_ID
 *              S_segloc_ERROR
 */

/* seglocCutSegmentExec  -  codel EXEC de CutSegment
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT seglocCutSegmentExec (SEGLOC_CUT_SEG *segCut, int *bilan)
{
  printf ("seglocCutSegmentExec: not implemented yet\n");
  return ETHER;
}


/*------------------------------------------------------------------------
 * ForceLocLocal
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_segloc_BIG_JUMP
 */

/* seglocForceLocalLocStart  -  codel START de ForceLocLocal
   Retourne:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT seglocForceLocalLocStart (SEGLOC_UNCERTAIN_POS *initPos, SEGLOC_POS *pos, int *bilan)
{
  printf (">>>>>>>>>>>>>>>  FORCE_LOC_LOCAL   >>>>>>>>>>>>>>>\n" 
	  "   current  : %5.2f %5.2f %5.2f\n"
	  "   force to : %5.2f %5.2f %5.2f\n",
	  SDI_F->currentSeglocPos.pos.x, SDI_F->currentSeglocPos.pos.y,
	  SDI_F->currentSeglocPos.pos.theta,
	  initPos->pos.x, initPos->pos.y, initPos->pos.theta);
  if (!seglocStartLocalisation(bilan, SEGLOC_FORCE_LOC_LOCAL))
    return ETHER;
  return END;
}

/* seglocForceLocalLocEnd  -  codel END de ForceLocLocal
   Retourne:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT seglocForceLocalLocEnd (SEGLOC_UNCERTAIN_POS *initPos, SEGLOC_POS *pos, int *bilan)
{
  seglocEndLocalisation(bilan);
  
  printf ("   output  : %5.2f %5.2f %5.2f  (%s)\n"
	  "<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<\n",
	  pos->x, pos->y, pos->theta, h2getMsgErrno(*bilan));

  return ETHER;
}

/* seglocForceLocalLocEnd  -  codel INTER de ForceLocLocal
   Voir plus haut (requete ForceLocLocal) */


/*------------------------------------------------------------------------
 * ResetFromRflex
 *
 * Description: 
 *
 * Bilans:      OK 
 */

/* seglocResetFromRflexExec  -  codel EXEC de ResetFromRflex
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT seglocResetFromRflexExec (int *bilan)
{
  SEGLOC_POS_VAR_SPEED pos;
  double newPos[3];
  double newVPos[6];
  int date;

  if (!seglocGetCurrentOdoPos(&pos, &date)) {
    *bilan = S_segloc_POSTER_NOT_FOUND;
    return ETHER;
  }

  /* update encoders */
  rob_enc[0] = pos.pos.x;
  rob_enc[1] = pos.pos.y;
  rob_enc[2] = pos.pos.theta;
  rob_enc[3] = pos.pos.theta;

  Vrob_enc[5]= pos.var.vtt;
  Vrob_enc[0]= pos.var.vxx;
  Vrob_enc[2]= pos.var.vyy;
  Vrob_enc[1]= pos.var.vxy;
  Vrob_enc[4]= pos.var.vyt;
  Vrob_enc[3]= pos.var.vxt;

  /* update global pos */
  newPos[0]  = Glob_pos[0];
  newPos[1]  = Glob_pos[1];
  newPos[2]  = Glob_pos[2];
  newVPos[0] = Vpos[0];
  newVPos[1] = Vpos[1];
  newVPos[2] = Vpos[2];
  newVPos[3] = Vpos[3];
  newVPos[4] = Vpos[4];
  newVPos[5] = Vpos[5];

  /* Remet a jour Glob_pos, Loc2Abs .. */
  init_location_st(newPos, newVPos);

  /* Update segloc transf and positions in SDI_F */
  seglocRecordSegkitTransf();

/*   seglocGetCurrentOdoPos(&SDI_F->currentLocoPos, &date); */

  /* update pos for pom */
  sickPomDate_at_loc = date;
  seglocPomExports();

  return ETHER;
}


