/**
 ** seglocCntrlTaskCodels.c
 **
 ** Codels used by the control task seglocCntrlTask
 **
 ** Auteur: 
 ** Date: 
 **
 **/

#ifdef VXWORKS
#include <vxWorks.h>
#else
#ifdef XR4000
#include <portLib.h>
#else
#include "portLib.h"
#endif
#endif
#include <stdio.h>
#include <string.h>
#include "server/seglocHeader.h"
#include "server/seglocPrint.h"
#include "h2mathLib.h"

/* From sick */
#include "odometer.h"
#include "err_pos_p.h"
/*------------------------------------------------------------------------
 * seglocGetLocParams  -  control codel of CONTROL request GetLocParams
 *
 * Description:    
 * Report (bilan): OK 
 *
 * Return:    OK ou ERROR
 */

STATUS seglocGetLocParams (SEGLOC_LOC_MODE *locMode, int *bilan)
{
  if (*locMode < 0 || *locMode >= SEGLOC_NB_LOC_MODES) {
    *bilan = S_segloc_BAD_PARAMS;
    return ERROR;
  }

  memcpy(&SDI_F->locParamsRqst, &SDI_F->locParams[*locMode], 
	 sizeof(SEGLOC_LOC_PARAMS));
 
  return OK;
}

/*------------------------------------------------------------------------
 * seglocSetLocParams  -  control codel of CONTROL request SetLocParams
 *
 * Description:    
 * Report (bilan): OK 
 *
 * Return:    OK ou ERROR
 */

STATUS seglocSetLocParams (SEGLOC_LOC_PARAMS *locParams, int *bilan)
{
  if (locParams->locMode < 0 || locParams->locMode >= SEGLOC_NB_LOC_MODES) {
    *bilan = S_segloc_BAD_PARAMS;
    return ERROR;
  }

#ifndef DYNAMIC_SET_POS
  if (locParams->updatePosMode == SEGLOC_UPDATE_POS) {
    printf ("seglocSetLocParams : no dynamic pos for now\n");
    *bilan = S_segloc_BAD_PARAMS;
    return ERROR;
  }   
#endif /* DYNAMIC_SET_POS */

  memcpy(&SDI_F->locParams[locParams->locMode], locParams,
	 sizeof(SEGLOC_LOC_PARAMS));

  if (locParams->locMode == SEGLOC_LOC_LOOP) {
    RealVariance = SDI_F->locParams[SEGLOC_LOC_LOOP].segkitSimulErr ? 0 : 1;
    SetAccuracyMove(SDI_F->locParams[SEGLOC_LOC_LOOP].simulErrLin,
		    SDI_F->locParams[SEGLOC_LOC_LOOP].simulErrAng,
		    SDI_F->locParams[SEGLOC_LOC_LOOP].simulErrLinAng);
  }
  return OK;
}


/*------------------------------------------------------------------------
 * seglocSegmentationOnCntrl  -  control codel of CONTROL request SegmentationOn
 *
 * Description:    
 * Report (bilan): OK 
 *
 * Return:    OK ou ERROR
 */

STATUS seglocSegmentationOnCntrl (int *bilan)
{
  SDI_F->state.segmentationOn = SEGLOC_TRUE;
  return OK;
}

STATUS seglocSegmentationOffCntrl (int *bilan)
{
  SDI_F->state.segmentationOn = SEGLOC_FALSE;
  return OK;
}


/*------------------------------------------------------------------------
 * seglocSetSegmentationParamsCntrl  -  control codel of CONTROL request SetSegmentationParams
 *
 * Description:    
 * Report (bilan): OK 
 *                 S_segloc_BAD_PARAMS
 *
 * Return:    OK ou ERROR
 */

STATUS seglocSetSegmentationParamsCntrl (SEGLOC_SEG_PARAMS *segmentationParams, int *bilan)
{

#ifndef SEG_VARIANCES
  if (segmentationParams->getVariances) {
    printf ("segloc error : this module must be compiled with SEG_VARIANCES\n"
	    "                to get variances on segments\n");
    *bilan = S_segloc_BAD_PARAMS;
    return ERROR;
  }
#endif 

  return OK;
}


/*------------------------------------------------------------------------
 * seglocSetNorthEnv  -  control codel of CONTROL request SetNorthEnv
 *
 * Description:    
 * Report (bilan): OK 
 *
 * Return:    OK ou ERROR
 */

STATUS seglocSetNorthEnv (SEGLOC_COMPASS_ENV *compassEnv, int *bilan)
{
  int i;

  /* Affiche bilan compas */
  print_struct_SEGLOC_COMPASS_STR(stdout, &SDI_F->compass, 0, 0, NULL, stdin);
  
  for (i=0; i<SDI_F->nbCompassData; i++) 
    printf("%5.3f\n", RAD_TO_DEG(SDI_F->deltaCompassTab[i]));

  /* Reset tout */
  SDI_F->compass.compassRes.rawAngleDeg = 0.0;
  SDI_F->compass.compassRes.correctedAngleDeg = 0.0;
  SDI_F->compass.compassRes.angleUncDeg = 0.0;
  SDI_F->compass.compassRes.deltaAngleAverageDeg = 0.0;
  SDI_F->compass.compassRes.deltaAngleMinDeg = 0.0;
  SDI_F->compass.compassRes.deltaAngleMaxDeg = 0.0;

  SDI_F->nbCompassData = 0;
  return OK;
}




