/**
 ** seglocSegsToPosters.c
 **
 ** Functions to record segkit Map or Segments to posters
 **
 ** Auteur: Sara Fleury
 ** Date:   15/05/99
 **
 **/

#ifdef VXWORKS
#include <vxWorks.h>
#include "h2sysLibProto.h"
#else
#ifdef XR4000
#include <portLib.h>
#include "xes.h"
#else
#include <portLib.h>
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <h2mathLib.h>
#include <math.h>

#include "server/sickPosterLib.h"
#include "server/seglocPosterLib.h"

#include "seglocSaveStat.h"

#include <lasdef.h>
#include <dbliste.h>

#define MATCHED_SEGS
#ifdef MATCHED_SEGS /* XXX*/
#include <localize.h> /* Pour Hyp ... */
#include <disambiguate_p.h> /* Pour GetBestHyp ... */
#include <map_mgr_p.h> /* Pour GetListValidHypo ... */
#include <image_mgr_p.h> /* Pour PtoIm */
#endif

#define GAUSS_99 3.0 

#include "seglocSegsToPosters.h"

#include "libmap/seglocStatsIO.h"

static char *seglocClassName[] = SEGLOC_CLASS_NAME;

static int pre_classify_hyps(SEGLOC_HYPS hypTbl[], int nbHyps, int numBest);
static BOOL analyse_hyp_list (PDBL *lhyp, Hyp *best, 
			      int numbest, int nbhyp,
			      SEGLOC_HYPS hypTbl[],
			      int *newbest);
static void displayHyp(int numHyp, SEGLOC_HYPS *hypTbl, int numBest);
static BOOL getName(char *prompt, char *name);
static BOOL getNumber(char *prompt, int *num);
static double computePointsAngle(p2d *p1, p2d *p2, p2d *p3);

BOOL seglocSaveResults(char *fileName,
		       int imageSegmentNum,
		       SEGLOC_SEG_PARAMS *segParams,
		       SEGLOC_LOC_MODE locMode,
		       SEGLOC_LOC_PARAMS *locParams,
		       SEGLOC_POS_VAR *initPos,
		       SICK_MEASURES_HEADER_STR *header,
		       int nbSol, int best, int *newbest)
{
  FILE *file;
  PDBL *possibleTransfs;
  Hyp *Best;
  BOOL result;
  char answer[64];
  SEGLOC_HYPS *hypTbl=NULL;
  int nbAcqSegs; double acqLength; double acqAngleView;
  int nbAspectSegs; double aspectLength; double aspectAngleView;
  SEGLOC_POS nulpos = {0,0,0};
  SEGLOC_POS_VAR newPos;

  *newbest = -1;
  if (nbSol == 0 || best == -1) {
    printf ("seglocSaveResults: no hyp !\n");
    *newbest = -1;
    nbSol = 0;
  }

  /* EXPORTS SOLUTION IN POSTER (for visualisation) AND CLASSIFIES */
  else {
    possibleTransfs = GetListValidHypo();  /* GetPossibleTranfs(); */
    Best = GetBestHyp();
    
    if (hypTbl != NULL) {
      free(hypTbl);
    }
    hypTbl = malloc(nbSol*sizeof(SEGLOC_HYPS));
    if (hypTbl == NULL) {
      printf("seglocSaveResults: cannot alloc hypTbl\n");
      return FALSE;
    }
    
    do {
      result = analyse_hyp_list(possibleTransfs, Best, 
				best, nbSol, hypTbl, newbest);

      /* Par dfaut on conserve la prcdente solution */
      newPos = *initPos;

      if (result) {
	/* Au moins un best */
	if (*newbest != -1) {
	  newPos.pos.x = hypTbl[*newbest].hyp->Tr[0];
	  newPos.pos.y = hypTbl[*newbest].hyp->Tr[1];
	  newPos.pos.theta = hypTbl[*newbest].hyp->Tr[2];
	  newPos.var.vxx = hypTbl[*newbest].hyp->Vtr[0];
	  newPos.var.vxy = hypTbl[*newbest].hyp->Vtr[1];
	  newPos.var.vyy = hypTbl[*newbest].hyp->Vtr[2];
	  newPos.var.vxt = hypTbl[*newbest].hyp->Vtr[3];
	  newPos.var.vyt = hypTbl[*newbest].hyp->Vtr[4];
	  newPos.var.vtt = hypTbl[*newbest].hyp->Vtr[5];
	}
      }
      else {
	strcpy(answer, "n");
	getName("Analysis aborted. Retry (y/n) ?", answer);
	if (strcmp(answer, "n") == 0) {
	  printf ("bye\n");
	  return FALSE;
	}
      }
    } while(!result);
    
    /* DATA ON ACQ SEGS AND ASPECT MAP */
    nbAcqSegs = computeSegsListAngleView(imageSegmentNum,
					 &nulpos,
					 &acqLength,
					 &acqAngleView);
    nbAspectSegs = computeSegsListAngleView(Get_aspect_num(),
					    &newPos.pos,
					    &aspectLength,
					    &aspectAngleView);
    /* XXXXX Depend de l'hypothese !!!! */
    aspectAngleView = 0.0;
  }

  /* RECORD RESULTS */
  /* Ouverture fichier pour sauvegarder les donnes */
  file = fopen(fileName, "w");
  if (file == NULL) {
    printf("seglocSaveResults: cannot open %s\n", fileName);
    return FALSE;
  }

  if (!writeDataFile(file, segParams, locMode, locParams, 
		     initPos,
		     header, hypTbl, nbSol, best,
		     nbAcqSegs, acqLength, acqAngleView,
		     nbAspectSegs, aspectLength, aspectAngleView)) {
    printf ("seglocSaveResults: cannot save results\n");
  }
  printf ("Results recorded in %s\n", fileName);

  fclose(file);

  /* Le 'best' a chang */
#ifdef BEST
  if (*newbest != -1) {
#else
    *newbest = 1000;
#endif
    *initPos = newPos;
#ifdef BEST
  }
#endif
  return TRUE;
}

/* ---------------------------------------------------------------------- */
/*                             LOCAL FUNCTIONS                            */
/* ---------------------------------------------------------------------- */

int computeSegsListAngleView(int imageNum, SEGLOC_POS *pos, 
			     double *length, double *angle)
{  
  PDBL *l;
  int i;
  double locLength=0;
  double locAngle=0;
  p2d p1, p2, pr;
  igla *ig;
  Sgl *s;

  pr.x = pos->x;
  pr.y = pos->y;
  ig = PtoIm(imageNum);
  for(l = ig->lsgl, i=0; l != NULL; l = dls(l)){
    s = (Sgl *)dlel(l);
    p1 = s->first;
    p2 = s->last;
    locLength += sqrt((p2.x-p1.x)*(p2.x-p1.x) + (p2.y-p1.y)*(p2.y-p1.y));
    locAngle += computePointsAngle(&p1, &pr, &p2);
    i++;
  }
  *length = locLength;
  *angle = locAngle;
  return i;
}



static BOOL analyse_hyp_list (PDBL *lhyp, Hyp *best, 
			      int numbest, int nbhyp,
			      SEGLOC_HYPS hypTbl[],
			      int *newbest)
{
  PDBL *l;
  Hyp *hyp;
  int i=0;
  SEGLOC_CLASS class;
  int nbHyp=0;
  int numBest=-1;
#ifdef BEST
  int myBest=-1;
#endif
  int num;
  BOOL noanswer, bestnotok;
  char answer[8];
  double goodPos[3];

  *newbest = -1;

  /**
   ** Record hyp list
   **/
  for (nbHyp=0, l = lhyp; l != NULL; l = dls(l)) {

    hyp = (Hyp *)dlel(l);

    if (hyp->nmatched == 0)
      continue;


    if (best->lsm == hyp->lsm) {
      numBest = nbHyp;
    }
    hypTbl[nbHyp].hyp = hyp;
    hypTbl[nbHyp].class = SEGLOC_CLASS_UNKNOWN;

    nbHyp++;
  }

  if (nbhyp != nbHyp) {
    printf ("analyse_hyp_list: not same nb hyps  (%d != %d)\n", nbhyp, nbHyp);
    return FALSE;
  }
  if (numbest != numBest) {
    printf ("analyse_hyp_list: not same numbest  (%d != %d)\n", 
	    numbest, numBest);
    return FALSE;
  }

  printf ("           %d hypothesis\n", nbHyp);

  numBest = pre_classify_hyps(hypTbl, nbHyp, numBest);


  /**
   ** Visualize and Classify 
   **/
  for (i=0; i<nbHyp; i++) {

    /* Visualize */
    displayHyp(i, hypTbl, numBest);

    /* Pre-selection bad hyps */
    if (*newbest != -1 &&
	(sqrt(SQR(hypTbl[i].hyp->Tr[0]-goodPos[0]) +
	     SQR(hypTbl[i].hyp->Tr[1]-goodPos[1]))>1.0 ||
	    fabs(angleLimit(hypTbl[i].hyp->Tr[2]-goodPos[2]))>0.1)) {
      printf ("    -> hyp auto %d removed (too far)\n", i);
      hypTbl[i].class = SEGLOC_CLASS_BAD;
    }

    else {

      /* Get class */
      do {
	strcpy(answer, "g");
	getName("\t(g/b/a) ?", answer);
	
	if (strcmp(answer, "g") == 0) {
	  class = SEGLOC_CLASS_GOOD;
	  printf ("Class %s\n", seglocClassName[class]);
#ifndef BEST
	  *newbest = i;
	  goodPos[0] = hypTbl[i].hyp->Tr[0];
	  goodPos[1] = hypTbl[i].hyp->Tr[1];
	  goodPos[2] = hypTbl[i].hyp->Tr[2];
#endif
	  noanswer = FALSE;
	}
	else if (strcmp(answer, "b") == 0) {
	  class = SEGLOC_CLASS_BAD;
	  printf ("Class %s\n", seglocClassName[class]);
	  noanswer = FALSE;
	}
	else if (strcmp(answer, "a") == 0) {
	  printf ("ABORT (no record)\n");
	  return FALSE;
	}
	else {
	  noanswer = TRUE;
	  printf ("Select between:  g (good)\n"
		  "                 b (bad)\n"
		  "                 a (abort)\n");
	}
	
	/* Record class */
	hypTbl[i].class = class;
	
      } while(noanswer);
    }
  } /* for */
  
  /**
   ** Visualize again ?
   **/
#ifdef BEST
  bestnotok = FALSE;
#else
  bestnotok = TRUE;
#endif

  do {

    /* Visualize selected sol */
    if (*newbest != -1) {
	printf ("Selected: %d\n\7", *newbest);
	displayHyp(*newbest, hypTbl, numBest);
    }

    strcpy(answer, "n");
    getName("Want to control (y/n) ? ", answer);

    if (strcmp(answer, "n") == 0 && bestnotok == TRUE)
      break;
	  
    if (strcmp(answer, "y") == 0) {
      num = -1;
      
      /* Select hyp */
      while(1) {
	
	num++;
	getNumber("Num ? ", &num);
	if (num < 0 || num>=nbHyp) 
	  break;
	
	/* Visualize */
	displayHyp(num, hypTbl, numBest);
	
	/* Get class */
	do {
	  if (hypTbl[num].class == SEGLOC_CLASS_BAD)
	    strcpy(answer, "b");
	  else
	    strcpy(answer, "g");
	  getName("\t(g/b/a) ?", answer);
	  
	  noanswer = TRUE;
	  if (strcmp(answer, "g") == 0) {
	    class = SEGLOC_CLASS_GOOD;
#ifndef BEST
	    *newbest = num;
#endif
	    printf ("Class %s\n", seglocClassName[class]);
	    noanswer = FALSE;
	  }
	  else if (strcmp(answer, "b") == 0) {
	    class = SEGLOC_CLASS_BAD;
	    printf ("Class %s\n", seglocClassName[class]);
	    noanswer = FALSE;
	  }
	  else if (strcmp(answer, "a") == 0) {
	  printf ("ABORT (no record)\n");
	  return FALSE;
	  }
	  else {
	    noanswer = TRUE;
	    printf ("Select between:  g (good)\n"
		    "                 b (bad)\n"
		    "                 a (abort)\n");
	  }
	  
	  /* Record class */
	  hypTbl[num].class = class;
	
	} while(noanswer);  

      } /* while(1) */
    } /* want to control */

    /** 
     ** Best ?
     **/
#ifdef BEST
    *newbest = -1;
    if (numBest != -1 && nbHyp > 1) {
      bestnotok = FALSE;
      myBest = numBest;
      getNumber("Num best ? ", &myBest);
      if (myBest >= 0 && myBest<nbHyp) {
	hypTbl[myBest].class = SEGLOC_CLASS_BEST;
	if (numBest != myBest) {
	  *newbest = myBest;
	  bestnotok = TRUE;
	  printf ("** DISAGREE WITH SEGLOC **\n");
	}
      }
    }
#endif
  } while (bestnotok);

  return TRUE;
}



static void displayHyp(int numHyp, SEGLOC_HYPS *hypTbl, int numBest)
{
  SEGLOC_POS_VAR pos;
  Hyp *hyp = hypTbl[numHyp].hyp;

  /* Met a jour le poster */
  pos.pos.x = hyp->Tr[0];
  pos.pos.y = hyp->Tr[1];
  pos.pos.theta = hyp->Tr[2];
  pos.var.vxx = hyp->Vtr[0];
  pos.var.vxy = hyp->Vtr[1];
  pos.var.vyy = hyp->Vtr[2];
  pos.var.vxt = hyp->Vtr[3];
  pos.var.vyt = hyp->Vtr[4];
  pos.var.vtt = hyp->Vtr[5];

  seglocHypSgl2SegAspectPoster(Get_aspect_num(), hyp, &pos);
  
  if (hypTbl[numHyp].class != SEGLOC_CLASS_UNKNOWN)
    printf ("   Class %s\n", seglocClassName[hypTbl[numHyp].class]);

  printf("   %d (%6.3f %6.3f %6.3f) nm %ld lm %-2.2f %c ", 
	 numHyp, pos.pos.x, pos.pos.y, pos.pos.theta,
	 hyp->nmatched, hyp->length_matched,
	 numHyp==numBest ? '*' : ' '); 
}

static int pre_classify_hyps(SEGLOC_HYPS hypTbl[], int nbHyps, int numBest)
{
  SEGLOC_HYPS *hypTblTmp;
  int i,j;
  double maxlength;
  int maxlength_j;
  int newNumBest=-1;
  
  hypTblTmp = (SEGLOC_HYPS *)malloc(nbHyps*sizeof(SEGLOC_HYPS));
  
  for(i=0; i<nbHyps; i++)
    hypTblTmp[i].class = SEGLOC_CLASS_UNKNOWN;
  
  for(i=0; i<nbHyps; i++) {
    maxlength = -1000;
    maxlength_j = -1;
    for(j=0; j<nbHyps; j++) {
      if (hypTblTmp[j].class == SEGLOC_CLASS_UNKNOWN) {
	if (hypTbl[j].hyp->length_matched >= maxlength) {
	  maxlength = hypTbl[j].hyp->length_matched;
	  maxlength_j = j;
	}
      }
    }
    hypTblTmp[i].hyp = hypTbl[maxlength_j].hyp;
    hypTblTmp[maxlength_j].class = SEGLOC_CLASS_GOOD;
    if (maxlength_j == numBest)
	newNumBest = i;
  }
  
  for(i=0; i<nbHyps; i++) {
    hypTbl[i].hyp = hypTblTmp[i].hyp;
  }
  free(hypTblTmp);
  return newNumBest;
}

/*
 * Recuperation d'un nom
 */
static BOOL getName(char *prompt, char *name)
{
  char newname[128]="";
  
  if (name == NULL) {
    printf ("getName: unallocated string\n");
    return FALSE;
  }

  while(1) {
    if(name[0] != '\0') {
      strcpy(newname, name);
      printf("%s (%s) ", prompt, newname);
    }
    else
      printf("%s ", prompt);
    
    fflush (stdout);
    scanf ("%[^\n]", newname);
    if (newname[0] == '.' && newname[1] == '\0') {
      printf ("interrupted\n");
      return FALSE;
    }
    if (newname[0] != '\0')
      strcpy(name, newname);
    else
      if (name[0] == '\0') 
	continue;
    break;
  }

  return TRUE;
}
static BOOL getNumber(char *prompt, int *num)
{
  int newnum;

  if(num != NULL) {
    newnum = *num;
    printf("%s (%d) ", prompt, newnum);
  }
  else
    printf("%s> ", prompt);
  fflush (stdout);
  scanf ("%d", &newnum);
  *num = newnum;
  return TRUE;
}




/*****************
  computePointsAngle  -  computes the angle (p1,p2,p3)   
 *****************/

static double computePointsAngle(p2d *p1, p2d *p2, p2d *p3)
{
  double l1, l2;
  double angle, scal;
  p2d v1;
  p2d v2;

  v1.x = p1->x - p2->x;
  v1.y = p1->y - p2->y;
  v2.x = p3->x - p2->x;
  v2.y = p3->y - p2->y;

  scal = v1.x * v2.x + v1.y * v2.y;
  
  l1 = sqrt(v1.x*v1.x + v1.y*v1.y);
  l2 = sqrt(v2.x*v2.x + v2.y*v2.y);

  angle = acos(scal/l1/l2);

  /* Une bonne approximation: */
/*   printf ("(~ %.4f) ", sqrt(l13_2 - (l12-l32) * (l12-l32)) / MIN(l12,l32)); */
  return angle;
}


