/**
 ** seglocSegsToPosters.c
 **
 ** Functions to record segkit Map or Segments to posters
 **
 ** Auteur: Sara Fleury
 ** Date:   15/05/99
 **
 **/

#ifdef VXWORKS
#include <vxWorks.h>
#else
#ifdef XR4000
#include <portLib.h>
#else
#include "portLib.h"
#endif
#endif

#include <stdio.h>
#include <string.h>
#include <h2mathLib.h>
#include <math.h>

#include "server/seglocHeader.h"
#include "libutils/seglocUtils.h"
#include "seglocConst.h"
#include "seglocSaveStat.h"

#include <lasdef.h>
#include <lasegm_p.h>
#include <dbliste.h>
#include <image_mgr_p.h>

/* From sick */
#include <odometer_p.h>

#include <user.h>
#include <map_mgr_p.h>
#include <map_builder_p.h>
#include <map_util.h>

#define MATCHED_SEGS
#ifdef MATCHED_SEGS /* XXX*/
#include <localize.h> /* Pour Hyp ... */
#include <disambiguate_p.h> /* Pour GetBestHyp ... */
#endif

#include "seglocSegsToPosters.h"

/*------------------------------------------------------------------------
 *
 * seglocSgl2MapPoster   - exporte les segments de la carte
 *
 * Description: from GetCurrentMap in Map_builder.c
 * 
 * Retourne:    nb segs or 0 if problem
 */

static SEGLOC_MAP * seglocMapPosterAddr=NULL;

int 
seglocSgl2MapPoster(void)
{
  PDBL *l;
  Sgl *s;
  int i=0;
  int nb=0;
  igla *Map;

  
  /* Get poster address */
  if (seglocMapPosterAddr==NULL) {
    seglocMapPosterAddr = (SEGLOC_MAP *)posterAddr(SEGLOC_MAP_POSTER_ID);
    if (seglocMapPosterAddr==NULL) {
      printf ("seglocUpdateMapPoster: no map poster address\n");
      return FALSE;
    }
    seglocMapPosterAddr->header.mapNum = -1;
  }

  Map = GetCurrentMapAddr();
  if (Map == NULL) {
    printf ("seglocUpdateMapPoster: empty global map\n");
    return FALSE;
  }

  /* Fill poster */
  posterTake(SEGLOC_MAP_POSTER_ID, POSTER_WRITE);

/*   strcpy(seglocMapPosterAddr->header.envName, SDI_F->seglocInit.envName); */
  (seglocMapPosterAddr->header.mapNum)++;
  seglocMapPosterAddr->header.pos.x = Map->x;
  seglocMapPosterAddr->header.pos.y = Map->y;
  seglocMapPosterAddr->header.pos.theta = atan2(Map->sn,Map->cs);

  for(l = Map->lsgl; l != NULL; l = dls(l)){
    s = (Sgl *)dlel(l);
    seglocMapPosterAddr->segs[i].x1 = s->first.x;
    seglocMapPosterAddr->segs[i].y1 = s->first.y;
    seglocMapPosterAddr->segs[i].x2 = s->last.x;
    seglocMapPosterAddr->segs[i].y2 = s->last.y;
    i++;
    if (i==SEGLOC_MAX_MAP_SEGS) {
      printf ("*** WARNING : seglocLoadMapExec: too many segs Keep only %d/%d\n",
	      SEGLOC_MAX_MAP_SEGS, nb);
      break;
    }
  }
  nb = i;
  seglocMapPosterAddr->header.nbSegs = nb;

  posterGive(SEGLOC_MAP_POSTER_ID);


  printf ("seglocUpdateMapPoster: %d segs recorded\n",nb);

  return nb;
}


/*------------------------------------------------------------------------
 *
 * seglocSgl2SegPoster  - exporte les segments acquis
 *
 * Description: from Copy_image_segment in image_mgr.c
 * 
 * Retourne:    nb segs or 0 if problem
 */

static SEGLOC_ACQ_SEGS *seglocSegPosterAddr=NULL;
static Sgl *segsAddress[SEGLOC_MAX_ACQ_SEGS];

int 
seglocSgl2SegPoster(int imageNum)
{
  igla *ig;
  PDBL *l;
  Sgl *s;

  int i=0;
  int nb=0;

  /* Poster adresse */
  if (seglocSegPosterAddr==NULL) {
    seglocSegPosterAddr = (SEGLOC_ACQ_SEGS *)posterAddr(SEGLOC_SEGS_POSTER_ID);
    if (seglocSegPosterAddr==NULL) {
      printf ("seglocSgl2SegPoster: no seg poster address\n");
      return FALSE;
    }
  }

  /* Segment Image */
  if(imageNum < 0){
    printf("seglocSgl2SegPoster: can't access to negative image number %d\n",
	   imageNum);
    return FALSE;
  }
  if((ig = PtoIm(imageNum)) == NULL) {
    printf("seglocSgl2SegPoster: in copying image number %d: no structure allocated !\n",imageNum);
    return FALSE;
  }  

  /* Save to poster */
  posterTake(SEGLOC_SEGS_POSTER_ID, POSTER_WRITE);

  /* Acquisition time */
  seglocSegPosterAddr->header.sickTime = SDI_F->seglocTimes.sickTime;

  /* Position */
  seglocSegPosterAddr->header.locoPos.x     = SDI_F->sickLocoPos.pos.x;
  seglocSegPosterAddr->header.locoPos.y     = SDI_F->sickLocoPos.pos.y;
  seglocSegPosterAddr->header.locoPos.theta = SDI_F->sickLocoPos.pos.theta;

  /* Get Position variances */
  seglocSegPosterAddr->header.variances.vxx = SDI_F->sickLocoPos.var.vxx;
  seglocSegPosterAddr->header.variances.vxy = SDI_F->sickLocoPos.var.vxy;
  seglocSegPosterAddr->header.variances.vxt = SDI_F->sickLocoPos.var.vxt;
  seglocSegPosterAddr->header.variances.vyy = SDI_F->sickLocoPos.var.vyy;
  seglocSegPosterAddr->header.variances.vyt = SDI_F->sickLocoPos.var.vyt;
  seglocSegPosterAddr->header.variances.vtt = SDI_F->sickLocoPos.var.vtt;

  /* Pom pose */
  seglocSegPosterAddr->header.sickPomPos = SDI_F->sickPomPos;

  /* All segments */
  for(l = ig->lsgl, i=0; l != NULL; l = dls(l)){
    s = (Sgl *)dlel(l);
    seglocSegPosterAddr->segs[i].x1 = s->first.x;
    seglocSegPosterAddr->segs[i].y1 = s->first.y;
    seglocSegPosterAddr->segs[i].x2 = s->last.x;
    seglocSegPosterAddr->segs[i].y2 = s->last.y;

    segsAddress[i] = s;

    i++;
    if (i==SEGLOC_MAX_ACQ_SEGS) {
      printf ("seglocSgl2SegPoster: too many segs. Keep only %d/%ld.\n",
	      SEGLOC_MAX_ACQ_SEGS, ig->nsg);
      break;
    }
  }
  nb = i;
  seglocSegPosterAddr->header.nbSegs = nb;

#ifdef SEG_VARIANCES
  /* Variances */
  if (SDI_F->segParams.getVariances) {
    for(l = ig->lsgl, i=0; l != NULL; l = dls(l)){
      s = (Sgl *)dlel(l);
      seglocSegPosterAddr->var[i].r = s->r;
      seglocSegPosterAddr->var[i].o = s->o;
      seglocSegPosterAddr->var[i].vr = s->vr;
      seglocSegPosterAddr->var[i].vo = s->vo;
      seglocSegPosterAddr->var[i].vro = s->vro;
      i++;
      if (i==SEGLOC_MAX_ACQ_SEGS) {
	printf ("seglocSgl2SegPoster: too many segs. Keep only %d/%ld.\n",
		SEGLOC_MAX_ACQ_SEGS, ig->nsg);
	break;
      }
    }
  }
#endif /* SEG_VARIANCES */

  posterGive(SEGLOC_SEGS_POSTER_ID);

  if (nb != ig->nsg) {
    printf ("seglocSgl2SegPoster: incoherent number of segs\n");
    return FALSE;
  }

/*   printf ("seglocSgl2SegPoster: %d segs recorded (%d @ %d)\n", nb, ig->nsg, ig);  */

  return nb;
}

   
#ifdef POSTER_MATCHED

/*------------------------------------------------------------------------
 *
 * seglocSgl2MatchedSegPoster  - exporte les segments acquis
 *
 * Description: from Copy_image_segment in image_mgr.c
 * 
 * Retourne:    nb segs or 0 if problem
 */

static SEGLOC_MATCHED_SEGS *seglocMatchedSegPosterAddr=NULL;

int 
seglocSgl2MatchedSegPoster(int imageNum, int imageAspectNum,
			  SEGLOC_POS_VAR *seglocPos)
{
  igla *ig, *iga;
#if 0
  PDBL *l;
#endif
  PDBL *lmap, *l_end, *lmatch;
  Sgl *s, *sm, *sp, *smap;
  int i=0, nbMatch=0;
  igla *Map;
  Smatch *match; /* 1 matching */
  Hyp *hyp;
  double xR, yR, cosR, sinR;
  double x1, y1, x2, y2;

  hyp = GetBestHyp();

  Map = GetCurrentMapAddr();
  if (Map == NULL) {
    printf ("seglocUpdateMapPoster: empty global map\n");
    return FALSE;
  }

  /* Poster adresse */
  if (seglocMatchedSegPosterAddr==NULL) {
    seglocMatchedSegPosterAddr = (SEGLOC_MATCHED_SEGS *)posterAddr(SEGLOC_MATCHED_POSTER_ID);
    if (seglocMatchedSegPosterAddr==NULL) {
      printf ("seglocSgl2MatchedSegPoster: no seg poster address\n");
      return FALSE;
    }
  }

  /* Segment Image */
  if(imageNum < 0){
    printf("seglocSgl2MatchedSegPoster: can't access to negative image number %d\n",
	   imageNum);
    return FALSE;
  }
  if((ig = PtoIm(imageNum)) == NULL) {
    printf("seglocSgl2MatchedSegPoster: in copying image number %d: no structure allocated !\n",imageNum);
    return FALSE;
  }  

  if((iga = PtoIm(imageAspectNum)) == NULL) {
    printf("seglocSgl2MatchedSegPoster: in copying aspect image number %d: no structure allocated !\n",imageAspectNum);
    return FALSE;
  }  

  /* Save to poster */
  posterTake(SEGLOC_MATCHED_POSTER_ID, POSTER_WRITE);

  /* Acquisition time */
  seglocMatchedSegPosterAddr->header.sickTime = SDI_F->seglocTimes.sickTime;

  /* Position */
  seglocMatchedSegPosterAddr->header.seglocPos.x     = seglocPos->pos.x;
  seglocMatchedSegPosterAddr->header.seglocPos.y     = seglocPos->pos.y;
  seglocMatchedSegPosterAddr->header.seglocPos.theta = seglocPos->pos.theta;

  xR =  seglocPos->pos.x;
  yR =  seglocPos->pos.y;
  cosR = cos(seglocPos->pos.theta);
  sinR = sin(seglocPos->pos.theta);

  /* Get Position variances */
  seglocMatchedSegPosterAddr->header.variances.vxx = seglocPos->var.vxx;
  seglocMatchedSegPosterAddr->header.variances.vxy = seglocPos->var.vxy;
  seglocMatchedSegPosterAddr->header.variances.vxt = seglocPos->var.vxt;
  seglocMatchedSegPosterAddr->header.variances.vyy = seglocPos->var.vyy;
  seglocMatchedSegPosterAddr->header.variances.vyt = seglocPos->var.vyt;
  seglocMatchedSegPosterAddr->header.variances.vtt = seglocPos->var.vtt;


  /* All perceived segments */
  i=0;
#if 0    /* XXX NNOT OK car liste des segments modifie' par la loc */
  for(l = ig->lsgl; l != NULL; l = dls(l)){
    s = (Sgl *)dlel(l);
    /* segs in global fram */
    seglocMatchedSegPosterAddr->segs[i].acqSeg.x1 = 
      xR + s->first.x * cosR - s->first.y * sinR;
    seglocMatchedSegPosterAddr->segs[i].acqSeg.y1 = 
      yR + s->first.x * sinR + s->first.y * cosR;
    seglocMatchedSegPosterAddr->segs[i].acqSeg.x2 = 
      xR + s->last.x * cosR - s->last.y * sinR;
    seglocMatchedSegPosterAddr->segs[i].acqSeg.y2 = 
      yR + s->last.x * sinR + s->last.y * cosR;
#else   /* XXX On repart donc des segments du  poster Segs */
  for (i=0; i<seglocSegPosterAddr->header.nbSegs; i++)  {
    /* segs in global fram */
    x1 = seglocSegPosterAddr->segs[i].x1;
    y1 = seglocSegPosterAddr->segs[i].y1;
    x2 = seglocSegPosterAddr->segs[i].x2;
    y2 = seglocSegPosterAddr->segs[i].y2;
    seglocMatchedSegPosterAddr->segs[i].acqSeg.x1 = xR + x1 * cosR - y1 * sinR;
    seglocMatchedSegPosterAddr->segs[i].acqSeg.y1 = yR + x1 * sinR + y1 * cosR;
    seglocMatchedSegPosterAddr->segs[i].acqSeg.x2 = xR + x2 * cosR - y2 * sinR;
    seglocMatchedSegPosterAddr->segs[i].acqSeg.y2 = yR + x2 * sinR + y2 * cosR;

    s = segsAddress[i];
#endif

    /* look for matching seg */
    if (hyp != NULL) {
      /* Les segments percus apparis */
      l_end = dlfin(hyp->lsm);
      sm = NULL;
      for(lmatch = l_end; lmatch != NULL; lmatch = dlp(lmatch)){
	match = (Smatch *)(lmatch->el);
	sp = match->sgp;
	/* found */
	if (sp == s) {
	  sm = match->sgm;
	  break;
	}
      }
      /* matched */
      if (sm) {
	/* look for aspect seg */
	for(lmap = iga->lsgl; lmap != NULL; lmap = dls(lmap)){
	  smap = (Sgl *)dlel(lmap);
	  if (smap == sm) {
	    break;
	  }
	}
	if (lmap == NULL)
	  seglocMatchedSegPosterAddr->segs[i].mapSeg = -2;
	else {
	  seglocMatchedSegPosterAddr->segs[i].mapSeg = smap->ifirst;
	  nbMatch++;
	}
      }
      /* not matched */
      else {
	seglocMatchedSegPosterAddr->segs[i].mapSeg = -1;
      }
    }
    else {
      seglocMatchedSegPosterAddr->segs[i].mapSeg = -1;
    }

    /* next perceived seg */
#if 0
    i++;
    if (i==SEGLOC_MAX_ACQ_SEGS) {
      printf ("seglocSgl2MatchedSegPoster: too many segs. Keep only %d/%ld.\n",
	      SEGLOC_MAX_ACQ_SEGS, ig->nsg);
      break;
    }
#endif
  }
  seglocMatchedSegPosterAddr->header.nbMatchSegs = i;

  posterGive(SEGLOC_MATCHED_POSTER_ID);

/*   if (i != ig->nsg) { */
/*     printf ("seglocSgl2MatchedSegPoster: incoherent number of segs\n"); */
/*     return FALSE; */
/*   } */

  printf ("seglocMatchedPoster: %d segs, %ld for loc, %d match\n", 
	  i, ig->nsg, nbMatch);

  return i;
}
#endif /* POSTER_MATCHED */


/*------------------------------------------------------------------------
 *
 * seglocSgl2SegAspectPoster - exporte la carte d'aspect et, 
 *                             si  matchedSegs=TRUE les segments matchs 
 *                             pour la meilleur hypothese 
 *
 * Description: from Copy_image_segment in image_mgr.c
 * 
 * Retourne:    nb segs or 0 if problem
 */

static SEGLOC_ACQ_SEGS *seglocAspectSegPosterAddr=NULL;

int 
seglocSgl2SegAspectPoster(int imageNum, BOOL matchedSegs, 
			  SEGLOC_POS_VAR *seglocPos)
{

#ifdef MATCHED_SEGS
  Hyp *bestHyp=NULL;
#endif

  if (matchedSegs)
    bestHyp = GetBestHyp();
  return seglocHypSgl2SegAspectPoster(imageNum, bestHyp, 
				      seglocPos);
}

/*------------------------------------------------------------------------
 *
 * seglocHypSgl2SegAspectPoster - exporte la carte d'aspect et
 *                                les segments matchs de l'hypothese
 *                                'hyp'  (si hyp != NULL)
 *
 * Description: from Copy_image_segment in image_mgr.c
 * 
 * Retourne:    nb segs or 0 if problem
 */

int 
seglocHypSgl2SegAspectPoster(int imageNum, Hyp *hyp,
			     SEGLOC_POS_VAR *seglocPos)
{
  igla *ig;
  PDBL *l;
  Sgl *s;
#ifdef MATCHED_SEGS
  Smatch *match; /* 1 matching */
  PDBL *l_end;
  Sgl *sp;  /* Le segment "percu" du matching */
  Sgl *sm;  /* Le segment "model" du matching */
  int nbMatched;
#endif

  int i=0;
  int nb=0;

  /* Poster adresse */
  if (seglocAspectSegPosterAddr==NULL) {
    seglocAspectSegPosterAddr = (SEGLOC_ACQ_SEGS *)posterAddr(SEGLOC_ASPECT_POSTER_ID);
    if (seglocAspectSegPosterAddr==NULL) {
      printf ("seglocSgl2SegAspectPoster: no seg poster address\n");
      return FALSE;
    }
  }

  /* Segment Image */
  if(imageNum < 0){
    printf("seglocSgl2SegAspectPoster: can't access to negative image number %d\n",
	   imageNum);
    return FALSE;
  }
  if((ig = PtoIm(imageNum)) == NULL) {
    printf("seglocSgl2SegAspectPoster: in copying image number %d: no structure allocated !\n",imageNum);
    return FALSE;
  }  

  /* Save to poster */
  posterTake(SEGLOC_ASPECT_POSTER_ID, POSTER_WRITE);

  /* Record new position and variance */
  seglocAspectSegPosterAddr->header.locoPos = seglocPos->pos;
  seglocAspectSegPosterAddr->header.variances = seglocPos->var;


#ifdef MATCHED_SEGS
  /** 
   ** Les segments "matchs" 
   **/
  if (hyp != NULL) {
    nbMatched = hyp->nmatched;

    seglocAspectSegPosterAddr->header.nbMatchSegs = nbMatched;
    
    /* Les segments percus apparis */
    l_end = dlfin(hyp->lsm);
    for(l = l_end; l != NULL && i<SEGLOC_MAX_ASPECT_SEGS; l = dlp(l)){
      match = (Smatch *)(l->el);
      sp = match->sgp;
      
      seglocAspectSegPosterAddr->segs[i].x1 = sp->first.x;
      seglocAspectSegPosterAddr->segs[i].y1 = sp->first.y;
      seglocAspectSegPosterAddr->segs[i].x2 = sp->last.x;
      seglocAspectSegPosterAddr->segs[i].y2 = sp->last.y;
      i++;
    }
    
    /* Segments modeles apparis
       (duplication avec les segments de la carte d'aspect qui suivent) */
    l_end = dlfin(hyp->lsm);
    for(l = l_end; l != NULL && i<SEGLOC_MAX_ASPECT_SEGS; l = dlp(l)){
      match = (Smatch *)(l->el);
      sm = match->sgm;
      
      seglocAspectSegPosterAddr->segs[i].x1 = sm->first.x;
      seglocAspectSegPosterAddr->segs[i].y1 = sm->first.y;
      seglocAspectSegPosterAddr->segs[i].x2 = sm->last.x;
      seglocAspectSegPosterAddr->segs[i].y2 = sm->last.y;
      i++;
    }
  }
  else
    seglocAspectSegPosterAddr->header.nbMatchSegs = 0;
#endif  /* MATCHED_SEGS */

  /* All aspect segments */
  nb = 0;
  for(l = ig->lsgl; l != NULL; l = dls(l)){
    s = (Sgl *)dlel(l);
    seglocAspectSegPosterAddr->segs[i].x1 = s->first.x;
    seglocAspectSegPosterAddr->segs[i].y1 = s->first.y;
    seglocAspectSegPosterAddr->segs[i].x2 = s->last.x;
    seglocAspectSegPosterAddr->segs[i].y2 = s->last.y;

    i++;
    nb++;
    if (i==SEGLOC_MAX_ASPECT_SEGS) {
      printf ("seglocSgl2SegAspectPoster: too many segs. Keep only %d/%ld.\n",
	      SEGLOC_MAX_ASPECT_SEGS, ig->nsg);
      break;
    }
  }

  seglocAspectSegPosterAddr->header.nbSegs = nb;

  posterGive(SEGLOC_ASPECT_POSTER_ID);

  if (nb != ig->nsg) {
    printf ("seglocSgl2SegAspectPoster: incoherent number of segs\n");
    return FALSE;
  }

  return nb;
}


/*------------------------------------------------------------------------
 *
 * seglocComputesConfidence
 *
 * Retourne:    la confiance
 */

double
seglocComputesConfidence(int imageSegmentNum,
			 double globPos[4],
			 double w,
			 SEGLOC_BOOL flags[SEGLOC_NB_CONF_ATTRIBUTES],
			 double values[SEGLOC_NB_CONF_ATTRIBUTES])
{
  Hyp *bestHyp=NULL;
  PDBL *l;
#ifdef MATCHED_SEGS
  Smatch *match; /* 1 matching */
  PDBL *l_end;
  Sgl *sp;  /* Le segment "percu" du matching */
  Sgl *sm;  /* Le segment "model" du matching */
#endif
  double angle_matched=0.0;
  double angle_not_matched=0.0;
  double angleUnderlap;
  double length_matched=0.0;
  double length_not_matched=0.0;
  double lengthUnderlap;
  double confidence, confidenceNorm;
  int nmatched = 0;
  igla pos;
  double vx, vy, length;
  double angle;
  double vo_sp=0.0;
  int numSeg, nummatch;
  int i=0;
  int nbAcqSegs;
  double acqLength;
  double acqAngleView;
  SEGLOC_POS nulpos = {0,0,0};
  static double attributesConfidenceWeigth[SEGLOC_NB_CONF_ATTRIBUTES] =
    SEGLOC_CONFIDENCE_ATT_WEIGTH;
  double tot=0.0, min = 0.0, max = 0.0;

#define NB_MAX_SEG 200
  int acqSegUsed[NB_MAX_SEG];

  for (i=0;i<SEGLOC_NB_CONF_ATTRIBUTES; i++) {
    values[i] = 0.0;
    if (flags[i]) {
      if (attributesConfidenceWeigth[i] < 0) {
	min += attributesConfidenceWeigth[i];
      }
      else {
	max += attributesConfidenceWeigth[i];
      }
    }
  }
  tot = max - min;

  bestHyp = GetBestHyp();
  if (!bestHyp)
    return 0;

  /* DATA ON ACQ SEGS AND ASPECT MAP */
  nbAcqSegs = computeSegsListAngleView(imageSegmentNum,
				       &nulpos,
				       &acqLength,
				       &acqAngleView);

  /* Init tableau-flags pour controler qu'un segment acquis n'est pas
     utilise plusieurs fois */
  for(numSeg=0; numSeg<NB_MAX_SEG; numSeg++) {
    acqSegUsed[numSeg] = -1;
  }

  pos.x = globPos[0];
  pos.y = globPos[1];
  pos.sn = pos.snt = sin(globPos[2]);
  pos.cs = pos.cst = cos(globPos[2]);
  pos.coosys = 'S';

/*     distMaha = pos_dist(initPosTab, initVPosTab, bestHyp->Tr, bestHyp->Vtr, */
/* 			NULL, NULL); */

  /**
   ** Calcul d'attributs
   **/
  
  /* Pour chaque couple de segment 'matchs' */
  nummatch=0;
  l_end = dlfin(bestHyp->lsm);
  for(l = l_end; l != NULL && i<SEGLOC_MAX_ASPECT_SEGS; 
      l = dlp(l), nummatch++){
    
    /* Couple segments apparis  */
    match = (Smatch *)(l->el);

    /* Segment acquis */
    sp = match->sgp;
    
    /* Segment modle */
    sm = match->sgm;
    
    /* Test segment acquis pas deja utilise dans cette hypothese */
    if (acqSegUsed[sp->ifirst] != -1) {
      printf ("\nacq seg %ld already used (mod segs %d and %ld)", 
	      sp->ifirst, acqSegUsed[sp->ifirst], sm->ifirst);
      acqSegUsed[sp->ifirst] = sm->ifirst;
      continue;
    }
    else {
      acqSegUsed[sp->ifirst] = sm->ifirst;
    }
    
    /* Vecteur tangentiel normalis */
    vx = sm->last.x - sm->first.x;
    vy = sm->last.y - sm->first.y;
    length = sqrt(vx*vx+vy*vy);
    sm->vx = -vy/length;
    sm->vy = vx/length;
    
    /* Calcul des longueurs cumules */
    if (flags[SEGLOC_CONF_LENGTH_MATCH] || 
	flags[SEGLOC_CONF_LENGTH_NOT_MATCH] ||
	flags[SEGLOC_CONF_LENGTH_PERCENT_NOT_MATCH]) {
      length = lengthOverlap(&pos, sp, sm, &lengthUnderlap);
      length_matched += length;
      length_not_matched += lengthUnderlap;
    }

    /* Calcul des angles cumuls */
    if (flags[SEGLOC_CONF_ANGLE_MATCH]) {
      angle = angleOverlap(&pos, sp, sm, &angleUnderlap);
      angle_matched += angle;
      angle_not_matched += angleUnderlap;
    }

    vo_sp = vo_sp + sp->vo;
    nmatched++;
  }
  
  /* Moyenne de la variance sur l'orientation des segmetns percus */
  vo_sp = vo_sp/nummatch;
  
  
  /**
   ** On enregistre les attributs
   **/
  
  /* Nombre de segments matchs normalise */
  if (flags[SEGLOC_CONF_NB_MATCH]) {
    values[SEGLOC_CONF_NB_MATCH] = (double)(nmatched-1)/(double)nbAcqSegs;
    if (values[SEGLOC_CONF_NB_MATCH] >= 1.0)
      printf ("\nWARNING nb seg match/acq = %d/%d = %f >= 1.0!\n",
	      nmatched, nbAcqSegs, values[SEGLOC_CONF_NB_MATCH]);
  }

  /* Longueur matche cumule */
  if (flags[SEGLOC_CONF_LENGTH_MATCH]) {
    values[SEGLOC_CONF_LENGTH_MATCH] = length_matched/acqLength;
    if (values[SEGLOC_CONF_LENGTH_MATCH] >= 1.0)
      printf ("\nWARNING length match/acq = %f/%f = %f >= 1.0!\n",
	      length_matched, acqLength, values[SEGLOC_CONF_LENGTH_MATCH]);
  }

  /* Angle du vu depuis robot match cumul */
  if (flags[SEGLOC_CONF_ANGLE_MATCH]) {
    values[SEGLOC_CONF_ANGLE_MATCH] = angle_matched/acqAngleView;
    if (values[SEGLOC_CONF_ANGLE_MATCH] >= 1.0)
      printf ("\nWARNING angle match/acq = %f/%f = %f >= 1.0!\n",
	      angle_matched, acqAngleView, values[SEGLOC_CONF_ANGLE_MATCH]);
  }

  /* Longueur non-matche cumule */
  if (flags[SEGLOC_CONF_LENGTH_NOT_MATCH]) {
    values[SEGLOC_CONF_LENGTH_NOT_MATCH] = length_not_matched/acqLength;
    if (values[SEGLOC_CONF_LENGTH_NOT_MATCH] >= 1.0)
      printf ("\nWARNING length not match/acq = %f/%f = %f >= 1.0!\n",
	      length_not_matched, acqLength, values[SEGLOC_CONF_LENGTH_NOT_MATCH]);
  }

  /* Longueur non-matche cumule en pourcentage */
  if (flags[SEGLOC_CONF_LENGTH_PERCENT_NOT_MATCH]) {
    values[SEGLOC_CONF_LENGTH_PERCENT_NOT_MATCH] =
      length_not_matched/(length_matched+length_not_matched);
  }

  /* Vitesse angulaire */
  if (flags[SEGLOC_CONF_W]) {
    values[SEGLOC_CONF_W] = fabs(w);
  }
  
  /* Incertitude sur la position rsultante
     XXX ce calcul me semble faux 
     ce serait plutot : GAUSS_99*sqrt(bestHyp->Vtr[0] + bestHyp->Vtr[2]) */
  if (flags[SEGLOC_CONF_POS_UNC]) {
    values[SEGLOC_CONF_POS_UNC] = SEGLOC_3SIGMA(SDI_F->sickSeglocPos.var.vxx +
						SDI_F->sickSeglocPos.var.vyy);
  }

  /* Incertitude sur l'angle rsultant */
  if (flags[SEGLOC_CONF_ANGLE_UNC]) {
    values[SEGLOC_CONF_ANGLE_UNC] = SEGLOC_3SIGMA(SDI_F->sickSeglocPos.var.vtt);
  }

  /* Distance de mahalanobis */

  /* Distribution angulaire des segments matchs 
     (eg, ecart-type de robotAngle) */

  /* Incertitude moyenne ou max sur les segments (sm->vo ou sp->vo) */

  /* Le critre de SEGKIT: la longueur apparie cumule */
/*     fprintf(pfileout, "  %5.4f", bestHyp->length_matched); */

  /* Pour le moment juste le premier attribut resultant 
     de l'analyse en composantes principales */
  /* Voir le rapport de karim fakhfakh */
  confidence = 0;
  for (i=0;i<SEGLOC_NB_CONF_ATTRIBUTES;i++) {
    values[i] = attributesConfidenceWeigth[i] * values[i];
    confidence += values[i];
  }

  /* Normalised confidence */
  confidenceNorm = (confidence - min)/tot;

#if 0
  printf ("conf %4.2f (", confidenceNorm);
  for (i=0;i<SEGLOC_NB_CONF_ATTRIBUTES;i++) {
    values[i] = values[i]/tot;
    printf ("%4.2f ", values[i]);
  }
  printf (")\n");
#endif

  return confidenceNorm;
}

#ifdef FUSED_SEGS

/* XXXX I thing I should add this following fonction in Map_builder.c */
/* Computes segs fusion */
/* The problem with Append2Map is that it also add new segs and it 
   changes the map - 2 things that we don't want */
/* From Append2Map in Map_builder.c */
static void segkitMergeSeg(long imageNum,PDBL *lsm)
{
  igla *data_ig;
  PDBL *l,*lm,*l_end;
  Sgl *sg,*sgp,*sgm;
  Smatch *m;
  Sgl smap;
  static double test[4];
  double *xp = Glob_pos;
  double *yp = Vpos;
  Hyp *hyp;

  hyp = GetBestHyp();

  data_ig = PtoIm(imageNum);
  
  /* merging matched segments  */

  l_end = dlfin(hyp->lsm);
  for(lm = l_end; lm != NULL; lm = dlp(lm)){
    m = (Smatch *)(dlel(lm));

    sgp = m->sgp;

    /* retreiving the global segment from the aspect */
    sgm = (m->sgm)->sfirst;
    
    /** sg = projection of sgp in xp coord. syst. **/
    project_sg_unc(sgp,&smap,xp,vxp);
    
    /* for now we don't deal with connections */
    
    /* merging sgm+smap -> sgm*/
    abs_segments_fusion(sgm,&smap,sgm);
    
    /* computing extremities leading to the maximum length of sgm */
    overlap_abs(sgm,&smap,test);

    /* in Mab_builder.c (static -> extern) */
    Maximum_length_seg(sgm,&smap,test);
  }
}

#endif /* FUSED_SEGS */

