/**
 ** Copyright (c) 1999-2003 LAAS/CNRS 
 **
 ** Matthieu Herrb - Mon Mar  8 1999
 **
 ** $Source: /cvs/openrobots/modules/segloc/codels/seglocSickToSegkit.c,v $
 ** $Revision: 1.1 $
 ** $Date: 2006/10/16 15:31:02 $
 ** 
 **/

#ident "$Id: seglocSickToSegkit.c,v 1.1 2006/10/16 15:31:02 sara Exp $"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "lasdef.h"
#include "las_struct_p.h"
#include "mat.h"
#include "image_mgr_p.h"

#include "lasers_robots.h"
#include "las_unc_p.h"

#ifndef VXWORKS
# include "portLib.h"
#else
#include <vxWorks.h>
#endif
 
#include "pomStruct.h"
#include "sickStruct.h"
#include "sick_save.h"

#include "sickConst.h"

#include "seglocSickToSegkit.h"
#include "../libutils/seglocUtils.h"

/* #define VERBOSE */

/* Specific variance matrix for the LMS 200. Needs to be updated since the 
   values are from the laser at LAAS. */
static double var_pol[3] = {SICK_VAR_R,
			    SICK_VAR_RO,
			    SICK_VAR_O};

/* ------------------------------------------------------------
 *
 *  seglocSetSickConfig : convert laser parameters frokm sick module
 *                        to segkit format
 *              (ie, SICK_MEASURES_HEADER_STR to laserConfiguration)
 */
static int seglocSetSickConfig(SICK_MEASURES_HEADER_STR *header,
			       laserConfiguration *sickConfig)
{
  static BOOL first=TRUE;

   double angleRange;
   sickConfig->pos[0] = header->sickPomPos.sensorToMain.euler.x;
   sickConfig->pos[1] = header->sickPomPos.sensorToMain.euler.y;
   sickConfig->pos[2] = header->sickPomPos.sensorToMain.euler.yaw;

   /* Angle min-max */
   angleRange = header->dtheta * (header->np - 1);
   sickConfig->angle_min = header->sickPomPos.sensorToMain.euler.yaw 
     - angleRange/2.;
   sickConfig->angle_max = header->sickPomPos.sensorToMain.euler.yaw
     + angleRange/2.;

   /* OTHERS SICK PARAMETERS ... XXXXXX */
   sickConfig->range = SICK_SEG_RHO_MAX;
   sickConfig->accuracy = SEGLOC_3SIGMA(SICK_VAR_R);

   /* Angular resolution */
   sickConfig->resolution = header->dtheta;

   /* Init OK */
   if (first) {
     printf ("seglocSetSickConfig: pos (%g %g %g)\n"
	     "           range %g meters (+/- %g)\n"
	     "           angle %g to %g radian (step of %g rd)\n",
	     sickConfig->pos[0], sickConfig->pos[1], sickConfig->pos[2], 
	     sickConfig->range, sickConfig->accuracy,
	     sickConfig->angle_min, sickConfig->angle_max, 
	     sickConfig->resolution);
     first = FALSE;
   }
   return 0;
}

/* ------------------------------------------------------------
 *
 *  setSegkitLaserImage : convert a laser image from Sick module 
 *                        to a segkit laser image
 *
 */
/******** 
   gets an image from the sick laser. Creates an image structure and returns 
   its index in the Glob_images array.
   performs a gs().
   Stores the robot position in the field ig->enc[] 
   The image position is set at the origin
   Coordinate system is set as Steering
*********/

int setSegkitLaserImage(SICK_MEASURES_STR *measures,
			laserConfiguration *sickConfig)
{
    long i; 
    long numig;
    p2d *pt;
    double inc, alpha;
    igla * ig;
    int j = 0;
   
    /* convert sick config */
    seglocSetSickConfig(&measures->header, sickConfig);

    /* Allocate the image */
    numig = Get_Image();
    ig = PtoIm(numig);
    if(ig == NULL) {
      printf("setSegkitLaserImage: cannot alloc laser image\n");
      return -1;
    }

#ifdef VERBOSE 
    printf ("setSegkitLaserImage: np %d, theta_min %f dtheta %f \n"
	    "sickPos x %f y %f th %f\n"
	    "robotPos x %f y %f th %f\n"
	    "var %f %f %f %f %f %f\n",
	    measures->header.np,
	    measures->header.theta_min,
	    measures->header.dtheta,
	    measures->header.sickPomPos.sensorToMain.euler[3],
	    measures->header.sickPomPos.sensorToMain.euler[4],
	    measures->header.sickPomPos.sensorToMain.euler[0],
	    measures->header.sickPomPos.mainToBase.euler[3],
	    measures->header.sickPomPos.mainToBase.euler[4],
	    measures->header.sickPomPos.mainToBase.euler[0],
	    measures->header.sickPomPos.mainToBase.v[9],
	    measures->header.sickPomPos.mainToBase.v[13],
	    measures->header.sickPomPos.mainToBase.v[14],
	    measures->header.sickPomPos.mainToBase.v[6],
	    measures->header.sickPomPos.mainToBase.v[10],
	    measures->header.sickPomPos.mainToBase.v[0]);

/*     for (i=0; i<SICK_N_PT_MAX; i++) { */
/*       printf ("%d %f\n", i, rho[i]); */
/*     } */
#endif /* VERBOSE */

    ig->coosys = 'S';
    ig->x = 0.;
    ig->y = 0.;
    ig->cs = 1.;
    ig->sn = 0.;
    ig->cst = 1.;
    ig->snt = 0.;

    ig->enc[0] = measures->header.sickPomPos.mainToBase.euler.x;
    ig->enc[1] = measures->header.sickPomPos.mainToBase.euler.y;
    ig->enc[2] = ig->enc[3] = measures->header.sickPomPos.mainToBase.euler.yaw;

    ig->var[0]=measures->header.sickPomPos.mainToBase.var.vxx;
    ig->var[1]=measures->header.sickPomPos.mainToBase.var.vxy;
    ig->var[2]=measures->header.sickPomPos.mainToBase.var.vyy;
    ig->var[3]=measures->header.sickPomPos.mainToBase.var.vox;
    ig->var[4]=measures->header.sickPomPos.mainToBase.var.voy;
    ig->var[5]=measures->header.sickPomPos.mainToBase.var.voo;

    pt = ig->tbp = Pt_cree(measures->header.np);
    if (pt == NULL) {
      printf("InitSick: cannot alloc %d pts\n", measures->header.np);
      return -1;
    }
    ig->lsgl = NULL;
    ig->nsg = 0;
    
    inc = measures->header.dtheta;
    alpha = sickConfig->pos[2] + sickConfig->angle_min;
/*     printf("Alpha %f (%f %f)\n",alpha,sickConfig->pos[2],
       sickConfig->angle_min); */
    for (i = 0;  i < measures->header.np; i++){
          if (measures->rho[i] == 0)
	    measures->rho[i] = SICK_RHO_MAX;
	  (pt+j)->r =  measures->rho[i];

	  /* !! ATTENTION !! : on n'enregistre pas tous les points 
	     L'angle est donc important. */
	  if((pt+j)->r < SICK_RHO_MAX){
	    (pt+j)->o = alpha;
	    (pt+j)->co = cos((pt+j)->o);
	    (pt+j)->so = sin((pt+j)->o);
	    (pt+j)->vr = var_pol[0];
	    (pt+j)->vro = var_pol[1];
	    (pt+j)->vo = var_pol[2];
	    (pt+j)->x = (pt+j)->r * (pt+j)->co;
	    (pt+j)->y = (pt+j)->r * (pt+j)->so;
	    point_inc_roxy(pt+j);
	    /* dealing with offset of the position of the sick on the robot */
	    (pt+j)->x += measures->header.sickPomPos.sensorToMain.euler.x;
	    (pt+j)->y += measures->header.sickPomPos.sensorToMain.euler.y;
	    (pt+j)->r = sqrt(  ((pt+j)->x) *  ((pt+j)->x) +  ((pt+j)->y) *  ((pt+j)->y));
	    (pt+j)->o = atan2((pt+j)->y,(pt+j)->x);
	    (pt+j)->co = cos((pt+j)->o);
	    (pt+j)->so = sin((pt+j)->o);
	    	    j += 1;
	    ig->nbp++;
	  }
	  alpha += inc;
	  
	}
    return numig;
}

