/**
 **  Translate la partie infrieur de la grille  (y < -41.6) de 1metre
 **  vers le bas
 **
 **  corr_map(filein, fileout)
 **
 **/


#include <stdio.h>
#include <sys/types.h>
#ifndef VXWORKS
#include <time.h>
#include <termios.h>
#endif
#include <unistd.h>
#include <stdlib.h>

#include <dbliste.h>
#include <lasdef.h>    /* igla ... */
#include <las_struct_p.h>  /* proto */
/* #include "las_prot.h" */
/* #include "image_mgr_p.h" */
#include <las_io_p.h>  /* rd_file_sg_c .. */
#include "seglocMapUtils.h"

#include "h2mathLib.h"

#ifdef VXWORKS
#include <vxWorks.h>
#endif
#include <stdio.h>

static void  seglocMoveSeg(igla *igin);

#ifdef DEBUG
static long wrt_file_sg_c_stdout(igla *igl);
#endif

void main(int argc, char **argv)
{
  igla *igin;
  FILE *ptr;

#ifdef UNIX
  char *filein;
  char *fileout;

  if (argc != 3) {
    printf ("usage: corr_map(char *filein, "
	    "char *fileout)\n");
    exit(1);
  }
  filein = argv[1];
  fileout  = argv[2];

  printf ("filein %s  -> fileout %s\n",
	  filein, fileout);
#endif /* UNIX */

  if((ptr = fopen(fileout,"w+")) == NULL){
    printf("can't open file %s\n",fileout);
    return;
  }
  fclose(ptr);

  igin = Imgl_cree();

  rd_file_sg_c(igin, filein);

  seglocMoveSeg(igin);

  /* Nouveau fichier */
#ifdef DEBUG
  wrt_file_sg_c_stdout(igin);
#else
  wrt_file_sg_c(igin, fileout);
#endif

  /* On libre tout */
  Imgl_free(igin);

}

static void  seglocMoveSeg(igla *igin)
{
  PDBL *l;
  Sgl *sinput;
  int i=0;

#define BOUNDS
#define ROT

#ifdef ROT /* pivot toto */
  double xRot=-27.2298;
  double yRot=-57.8698;
  double angleRot=-0.017;
  double ca=cos(angleRot);
  double sa=sin(angleRot);
  double dx, dy;
  double length;
#endif

  double xmin = -74.0;
  double xmax = -27.24;
  double ymin = -71.7;
  double ymax = -44.5;

  /* Parcous les segments de la carte */
  l = igin->lsgl;
  do {

    sinput = (Sgl *)l->el;

#ifdef BOUNDS
   if (((sinput->first.x > xmin && sinput->first.y < ymax) &&
	 (sinput->first.x < xmax && sinput->first.y > ymin)) ||

	((sinput->last.x > xmin && sinput->last.y < ymax) &&
	 (sinput->last.x < xmax && sinput->last.y > ymin))) {
#endif

#ifdef ROT      
      dx = sinput->first.x - xRot;
      dy = sinput->first.y - yRot;
      sinput->first.x = xRot + dx*ca - dy*sa;
      sinput->first.y = yRot + dx*sa + dy*ca;

      dx = sinput->last.x - xRot;
      dy = sinput->last.y - yRot;
      sinput->last.x = xRot + dx*ca - dy*sa;
      sinput->last.y = yRot + dx*sa + dy*ca;      

      dx = sinput->last.x - sinput->first.x;
      dy = sinput->last.y - sinput->first.y;
      length = sqrt(dx*dx+dy*dy);
      sinput->vx = -dy/length;
      sinput->vy = dx/length;
      sinput->o = atan2(sinput->vy,sinput->vx);

#else /* TRANSLATION */

    /* Extremite 1 */
    sinput->first.y += 0.23;
    sinput->first.r = sqrt(SQR(sinput->first.x) + SQR(sinput->first.y));
    /* Extremite 2 */
    sinput->last.y += 0.23;
    sinput->last.r = sqrt(SQR(sinput->last.x) + SQR(sinput->last.y));
#endif

#ifdef BOUNDS
   }
#endif
    l = dls(l);
    i++;

  } while (l != NULL);

  printf ("%d segs in file\n", i);
  return;
}

 
#ifdef DEBUG
static long wrt_file_sg_c_stdout(igla *igl)
     
{
  PDBL *l;
  Sgl *s;
  FILE *ptr=stdout;

  if(igl->lsgl == NULL) {
    printf ("nothing \n");
    return(-1);
  }

    
  fprintf(ptr,"%ld \n",igl->nsg);
  printf ("%d segs\n", igl->nsg);
    
  fprintf(ptr,"%.8f %.8f %.8f %.8f %.8f %.8f %c \n",
	  igl->x,igl->y,igl->cs,igl->sn,igl->cst,igl->snt,igl->coosys);
  fprintf(ptr,"%.8f %.8f %.8f %.8f \n",igl->enc[0],igl->enc[1],
	  igl->enc[2],igl->enc[3]);
  for(l = igl->lsgl; l != NULL; l = dls(l)){
    
    s = (Sgl *)dlel(l);
    
    fprintf(ptr, "%.8f ",s->first.x);
    fprintf(ptr, "%.8f \n",s->first.y);
    fprintf(ptr, "%.8f ",s->last.x);
    fprintf(ptr, "%.8f \n",s->last.y);
    fprintf(ptr, "%.9f %.9f %.9f \n",s->vx,s->vy,s->d);
    fprintf(ptr, "%.9f %.9f \n",s->r,s->o);
    fprintf(ptr, "%.9f %.9f %.9f \n",s->vr,s->vro,s->vo);
    fprintf(ptr, "%c %c \n",s->typf,s->typl);
    

  }

  return(1);
}
#endif
