/**
filter-stats.c  : Function that computes attributs (for instance the
                  number and the cumulated length of the matched
                  segments) from data recorded seglocStatsIO.c
                  
                  Get in input a list of files produced with
                  writeDataFile.

                  Produces its results in a file with '-results'
                  extension name.
                  See for instance: example-segloc-results.

                  From these data and using probabilistics procedures we
                  can computes the probability that a solution computed
                  by segkit is a correct solution.
                  (SEE KARIM FAKHFAKH WORK)

 **/


#include <stdio.h>
#include <sys/types.h>
#include <time.h>
#include <termios.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include <dbliste.h>
#include <lasdef.h>
#include <las_struct_p.h>
/* #include "las_prot.h" */
/* #include "image_mgr_p.h" */
#include <las_io_p.h>
#include <odom_funct_p.h>

#include <map_util.h>
#include <localize.h>

#include "sickStruct.h"
#include "seglocStruct.h"
#include "seglocStatsIO.h"


#define GAUSS_99 3.0 

/* ---------------------------------------------------------------------- */


static int writeStatFile(FILE *pfileout, SEGLOC_SEG_PARAMS *segParams,
			 SEGLOC_LOC_MODE *locMode,
			 SEGLOC_LOC_PARAMS *locParams,
			 SEGLOC_POS_VAR *initPos,
			 SICK_MEASURES_HEADER_STR *header,
			 SEGLOC_HYPS_STR *hypTbl[], int nbSol,
			 int nbAcqSegs, double acqLength, double acqAngleView,
			 int nbAspectSegs, double aspectLength, double aspectAngleView)
{			
  int numhyp, nummatch, numSeg;
  Hyp *hyp;
  Sgl *sp;  /* Le segment "percu" du matching */
  Sgl *sm;  /* Le segment "model" du matching */
  double angle_matched;
  double angle_not_matched;
  double angleUnderlap;
  double length_matched;
  double length_not_matched;
  double lengthUnderlap;
  int undermatched=0;
  int nmatched = 0;
  igla pos;
  double vx, vy, length;
  double angle;

  double robotAngle;
  double vo_sp;
  double norm;
  double initPosTab[3];
  double initVPosTab[6];
  double max_segkit_length=-1000;
  int bestSegkit=-1;

#define NB_MAX_SEG 400
  int acqSegUsed[NB_MAX_SEG];
  int modSegUsedGood[NB_MAX_SEG];
  int modSegUsedBad[NB_MAX_SEG];

  initPosTab[0] = initPos->pos.x;
  initPosTab[1] = initPos->pos.y;
  initPosTab[2] = initPos->pos.theta;
  initVPosTab[0] =  initPos->var.vxx;
  initVPosTab[1] =  initPos->var.vxy;
  initVPosTab[2] =  initPos->var.vyy;
  initVPosTab[3] =  initPos->var.vxt;
  initVPosTab[4] =  initPos->var.vyt;
  initVPosTab[5] =  initPos->var.vtt;

  /* Init les points attribues au segments du modele */
  for(numSeg=0; numSeg<NB_MAX_SEG; numSeg++) {
    modSegUsedGood[numSeg] = 0;
    modSegUsedBad[numSeg] = 0;
  }

  /* Recherche du best de SEGKIT qu'on a betement perdu */
  for (numhyp = 0; numhyp<nbSol; numhyp++) {
    if (hyp->length_matched > max_segkit_length) {
      max_segkit_length = hyp->length_matched;
      bestSegkit = numhyp;
    }
  }
  if (bestSegkit!=0)
    printf ("(best SEGKIT = %d)\n",bestSegkit);



  /* Pour chaque hypothese */
  for (numhyp = 0; numhyp<nbSol; numhyp++) {

    /* Init tableau-flags pour controler qu'un segment acquis n'est pas
       utilise plusieurs fois */
    for(numSeg=0; numSeg<NB_MAX_SEG; numSeg++) {
      acqSegUsed[numSeg] = -1;
    }

    hyp = &((*hypTbl)[numhyp].hyp);
    
    pos.x = hyp->Tr[0];
    pos.y = hyp->Tr[1];
    pos.sn = pos.snt = sin(hyp->Tr[2]);
    pos.cs = pos.cst = cos(hyp->Tr[2]);
    pos.coosys = 'S';

    angle_matched = 0;
    angle_not_matched = 0;
    length_matched = 0;
    length_not_matched = 0;
    nmatched = 0;

    vo_sp = 0;
    robotAngle = 0;
/*     distMaha = pos_dist(initPosTab, initVPosTab, hyp->Tr, hyp->Vtr, */
/* 			NULL, NULL); */

    /**
     ** Calcul d'attributs
     **/
    
    /* Pour chaque couple de segment 'matchs' */
    for(nummatch=0; nummatch<hyp->nmatched; nummatch++) {

      /* Segment acquis */
      sp = &(*hypTbl)[numhyp].segacq[nummatch];

      /* Segment modle */
      sm = &(*hypTbl)[numhyp].segmod[nummatch];

      /* Test segment acquis pas deja utilise dans cette hypothese */
      if (acqSegUsed[sp->ifirst] != -1) {
	printf ("\nHyp %d: acq seg %ld already used (mod segs %d and %ld)", 
		numhyp, sp->ifirst, acqSegUsed[sp->ifirst], sm->ifirst);
	acqSegUsed[sp->ifirst] = sm->ifirst;
	continue;
      }
      else {
	acqSegUsed[sp->ifirst] = sm->ifirst;
      }

      if (sm->ifirst >=0 && sm->ifirst<NB_MAX_SEG) {

	if ((*hypTbl)[numhyp].class == SEGLOC_CLASS_GOOD ||
	    (*hypTbl)[numhyp].class == SEGLOC_CLASS_BEST) {
	  (modSegUsedGood[sm->ifirst]) ++;
	} 
	else if ((*hypTbl)[numhyp].class == SEGLOC_CLASS_BAD) {
	  (modSegUsedBad[sm->ifirst]) ++;
	} 
      }
	
      /* Vecteur tangentiel normalis */
      vx = sm->last.x - sm->first.x;
      vy = sm->last.y - sm->first.y;
      length = sqrt(vx*vx+vy*vy);
      sm->vx = -vy/length;
      sm->vy = vx/length;

      /* Calcul des longueurs cumules */
      length = lengthOverlap(&pos, sp, sm, &lengthUnderlap);
      length_matched += length;
      length_not_matched += lengthUnderlap;

      /* Calcul des angles cumuls */
      angle = angleOverlap(&pos, sp, sm, &angleUnderlap);
      angle_matched += angle;
      angle_not_matched += angleUnderlap;

      vo_sp = vo_sp + sp->vo;
      robotAngle = robotAngle + (*hypTbl)[numhyp].robotAngle[nummatch];
      nmatched++;
    }

    /* Moyenne de l'angle du robot 
       (pas interessant ! calculer plutot l'ecart type) */
    robotAngle = robotAngle/ nummatch;

    /* Moyenne de la variance sur l'orientation des segmetns percus */
    vo_sp = vo_sp/nummatch;

    /* Pourcentage de non-superposition des couples de segments */
    undermatched = length_not_matched/(length_matched+length_not_matched)*100;


    /**
     ** On enregistre les attributs
     **/
/*     fprintf(pfileout, "class  nb-m  l-m  ang-m  l-out  out  w\n"); */

    /* la classe */
    fprintf(pfileout, "  %d", 	    (*hypTbl)[numhyp].class);

    /* Nombre de segments matchs normalise */
    norm = (double)nmatched/(double)nbAcqSegs;
    fprintf(pfileout, "   %5.3f",  norm);
    if (norm >= 1.0)
      printf ("\nWARNING hyp %d: nb seg match/acq = %d/%d = %f >= 1.0!\n",
	      numhyp, nmatched, nbAcqSegs, norm);

    /* Longueur matche cumule */
    norm = length_matched/acqLength;
    fprintf(pfileout, "  %5.2f",     norm);
    if (norm >= 1.0)
      printf ("\nWARNING hyp %d: length match/acq = %f/%f = %f >= 1.0!\n",
	      numhyp, length_matched, acqLength, norm);

    /* Angle match cumul */
    norm = angle_matched/acqAngleView;
    fprintf(pfileout, "  %5.2f",     norm); 
    if (norm >= 1.0)
      printf ("\nWARNING hyp %d: angle match/acq = %f/%f = %f >= 1.0!\n",
	      numhyp, angle_matched, acqAngleView, norm);

    /* Longueur non-matche cumule */
    norm = length_not_matched/acqLength;
    fprintf(pfileout, "  %5.2f",     norm);
    if (norm >= 1.0)
      printf ("\nWARNING hyp %d: length not match/acq = %f/%f = %f >= 1.0!\n",
	      numhyp, length_not_matched, acqLength, norm);

    /* Pourcentage non-match */
    fprintf(pfileout, "  %2d", 	     undermatched);

    /* Vitesse angulaire */
    fprintf(pfileout, "  %5.2f",     header->sickPomPos.VLocal.euler[0]);

    /* Incertitude sur la position rsultante 
       XXX ce calcul me semble faux 
       ce serait plutot : GAUSS_99*sqrt(hyp->Vtr[0] + hyp->Vtr[2]) */
    fprintf(pfileout, "  %5.4f",     GAUSS_99*sqrt(hyp->Vtr[0]*hyp->Vtr[0] +
						   hyp->Vtr[2]*hyp->Vtr[2]));
    /* Incertitude sur l'angle rsultant */
    fprintf(pfileout, "  %5.4f", hyp->Vtr[5]);

    /* Distance de mahalanobis */
/*     fprintf(pfileout, "  %5.4f", distMaha); */

    /* Distribution angulaire des segments matchs 
       (eg, ecart-type de robotAngle) */

    /* Incertitude moyenne ou max sur les segments (sm->vo ou sp->vo) */

    /* Le critre de SEGKIT: la longueur apparie cumule */
    fprintf(pfileout, "  %5.4f", hyp->length_matched);

    /* Le best de SEGKIT */
    if (numhyp==bestSegkit)
      fprintf(pfileout, "  1");
    else
      fprintf(pfileout, "  0");

    fprintf(pfileout, "\n");
  } /* Pour chaque hypothese */
  

  /* Libre les segments et le tableau d'hypotheses */
  for(numhyp=0;numhyp<nbSol;numhyp++) {
    free((*hypTbl)[numhyp].segmod);
    free((*hypTbl)[numhyp].segacq);
    free((*hypTbl)[numhyp].robotAngle);
  }
  free(*hypTbl);

  return 1;
}


/* ---------------------------------------------------------------------- */

int main(int argc, char **argv)
{
  char *basein;
  char *baseout;
  int numfirst, numlast;
  char filename[256];
  FILE *pfilein;
  FILE *pfileout;
  int numfile;
  int nbSol;

  int nbAcqSegs; double acqLength; double acqAngleView;
  int nbAspectSegs; double aspectLength; double aspectAngleView;

  SEGLOC_HYPS_STR *hypTbl=NULL;

  SEGLOC_SEG_PARAMS segParams;
  SEGLOC_LOC_MODE locMode;
  SEGLOC_LOC_PARAMS locParams;
  SICK_MEASURES_HEADER_STR header;
  SEGLOC_POS_VAR initPos;

  /**
   **  Lecture parametres
   **/
  if (argc != 4 && argc != 5) {
    printf ("usage: segloc-filter <filein> <first> <last> [<fileout>]\n");
    exit(1);
  }
  basein = argv[1];
  numfirst = atof(argv[2]);
  numlast = atof(argv[3]);

  if (argc == 5) 
    baseout = argv[4];
  else
    baseout = argv[1];

  printf ("args : filein %s  first %d  last %d\n",
	  basein, numfirst, numlast);


  /**
   **  Ouverture fichiers
   **/
  sprintf (filename, "%s-segloc.%03d", basein,  numfirst);
  if((pfilein = fopen(filename,"r")) == NULL){
    printf("can't open file %s\n", filename);
    fclose(pfilein);
    return 0;
  }
  
/*   sprintf (filename, "%s-segloc-results", basein); */
/*   if((pfileout = fopen(filename,"r")) != NULL) { */
/*     printf("%s exists\n", filename); */
/*     fclose(pfileout); */
/*     return; */
/*   } */
  sprintf (filename, "%s-stats", baseout);
  if((pfileout = fopen(filename,"w")) == NULL) {
    printf("can't open file %s\n", filename);
    return 0;
  }


  /**
   **  BOUCLE LECTURE/ECRITURE FICHIERS
   **/

  fprintf(pfileout, "class  nb-m  l-m  ang-m  l-out  out  w\n");

  for (numfile=numfirst; numfile<numlast; numfile++) {

    /* LECTURE FICHIER */

    /* Ouverture fichier */
    sprintf (filename, "%s-segloc.%03d", basein,  numfile);
    if((pfilein = fopen(filename,"r")) == NULL){
      printf("can't open file %s\n", filename);
      return 1;
    }
/*     printf ("%s\n", filename); */
    printf ("%d..", numfile);

    if(!(nbSol = readDataFile(pfilein, &segParams, &locMode, &locParams, 
			      &initPos, &header, &hypTbl,
			      &nbAcqSegs, &acqLength, &acqAngleView,
			      &nbAspectSegs, &aspectLength, &aspectAngleView
			      ))) {
      fclose(pfilein);
      continue;
/*       fclose(pfileout); */
/*       return; */
    }
    fclose(pfilein);

    /* ECRITURE FICHIER */
    if(!writeStatFile(pfileout, &segParams, &locMode, &locParams, &initPos,
		      &header, &hypTbl, nbSol,
		      nbAcqSegs, acqLength, acqAngleView,
		      nbAspectSegs, aspectLength, aspectAngleView)) {
      fclose(pfileout);
      return 1;
    }

  } /* FOR Fichier */

  fclose(pfileout);
  return 0;
}


