/**
 ** 
 **
 **
 **/


#include <stdio.h>
#include <sys/types.h>
#ifndef VXWORKS
#include <time.h>
#include <termios.h>
#endif
#include <unistd.h>
#include <stdlib.h>

#include "portLib.h"
#include "posterLib.h"
#include "server/pomPosterLib.h"
#include "seglocStruct.h"
#include "seglocMapIO.h"

BOOL
seglocReadSegkit2SeglocMap2(char *filename, int *nbSegs, SEGLOC_SEG *map)
{
  SEGKIT_MAP_HEADER header;
  SEGKIT_SEG *segs, *seg;
  int i;

  if ((segs = seglocReadSegkitMap(filename, &header)) == NULL) {
    printf ("seglocReadSegkit2SeglocMap : seglocReadSegkitMap failed\n");
    return FALSE;
  }
  
  *nbSegs = header.nbSegs;
  for (i=0,seg=segs; i<header.nbSegs; i++,seg++) {
    map[i].x1 = seg->x1;
    map[i].y1 = seg->y1;
    map[i].x2 = seg->x2;
    map[i].y2 = seg->y2;
  }

  free(segs);
  return TRUE;
}

BOOL
seglocReadSegkit2SeglocMap(char *filename, SEGLOC_MAP *map)
{
  SEGKIT_MAP_HEADER header;
  SEGKIT_SEG *segs, *seg;
  int i;

  if ((segs = seglocReadSegkitMap(filename, &header)) == NULL) {
    printf ("seglocReadSegkit2SeglocMap : seglocReadSegkitMap failed\n");
    return FALSE;
  }
  
  map->header.nbSegs = header.nbSegs;
  for (i=0,seg=segs; i<header.nbSegs; i++,seg++) {
    map->segs[i].x1 = seg->x1;
    map->segs[i].y1 = seg->y1;
    map->segs[i].x2 = seg->x2;
    map->segs[i].y2 = seg->y2;
  }

  free(segs);
  return TRUE;
}


SEGKIT_SEG *
seglocReadSegkitMap(char *filename, SEGKIT_MAP_HEADER *mapHeader)
{
  FILE *pfile;
  int i;
  int nbSegs;
  SEGKIT_SEG *segs=NULL;
  SEGKIT_SEG *seg;

  /* Lecture fichier */
  if ((pfile = fopen(filename,"r")) == NULL) {
    printf ("seglocReadSegkitMap: cannot open %s\n", filename);
    return NULL;
  }
  
  /* Nombre de segments ? */
  if (fscanf(pfile, "%d", &nbSegs) != 1) {
    printf ("seglocReadSegkitMap: cannot read nbsegs\n");
    fclose(pfile);
    return NULL;
  }
  mapHeader->nbSegs = nbSegs;

  /* Entete */
  if (fscanf(pfile, "%lf %lf %lf %lf %lf %lf %c", 
	     &mapHeader->x, &mapHeader->y, 
	     &mapHeader->cs, &mapHeader->sn,
	     &mapHeader->cst, &mapHeader->snt, 
	     &mapHeader->coosys) != 7) {
    printf ("seglocReadSegkitMap: cannot read header\n");
    fclose(pfile);
    return NULL;
  }
  if (fscanf(pfile, "%lf %lf %lf %lf", 
	     &mapHeader->enc[0], &mapHeader->enc[1],
	     &mapHeader->enc[2], &mapHeader->enc[3]) != 4) {
    printf ("seglocReadSegkitMap: cannot read enc\n");
    fclose(pfile);
    return NULL;
  }
	     
  /* Les segments */
  if ((segs = (SEGKIT_SEG *) malloc (sizeof(SEGKIT_SEG) * nbSegs)) == NULL) {
    printf ("seglocReadSegkitMap: cannot alloc segs\n");
    fclose(pfile);
    return NULL;
  }

  seg=segs;
  for(i=0; i<nbSegs; i++,seg++) {
    
    /* Premier point */
    if (fscanf(pfile, "%lf %lf", 
	       &seg->x1, &seg->y1) != 2) {
      printf ("seglocReadSegkitMap: cannot read first point (seg %d)\n", i);
      fclose(pfile);
      free(segs);
      return NULL;
    }
    /* second point */
    if (fscanf(pfile, "%lf %lf", 
	       &seg->x2, &seg->y2) != 2) {
      printf ("seglocReadSegkitMap: cannot read second point (seg %d)\n", i);
      fclose(pfile);
      free(segs);
      return NULL;
    }
    /* vecteur */
    if (fscanf(pfile, "%lf %lf %lf", 
	       &seg->vx, &seg->vy, &seg->d) != 3) {
      printf ("seglocReadSegkitMap: cannot read vx ... (seg %d)\n", i);
      fclose(pfile);
      free(segs);
      return NULL;
    }
    /* o,r */
    if (fscanf(pfile, "%lf %lf", 
	       &seg->r, &seg->o) != 2) {
      printf ("seglocReadSegkitMap: cannot read o, r (seg %d)\n", i);
      fclose(pfile);
      free(segs);
      return NULL;
    }
    /* variances */
    if (fscanf(pfile, "%lf %lf %lf", 
	       &seg->vr, &seg->vro, &seg->vo) != 3) {
      printf ("seglocReadSegkitMap: cannot read variances ... (seg %d)\n", i);
      fclose(pfile);
      free(segs);
      return NULL;
    }
    /* type extremite */
    if (fscanf(pfile, "%s %s", 
	       &seg->typ1, &seg->typ2) != 2) {
      printf ("seglocReadSegkitMap: cannot read extremity types (seg %d)\n", 
	      i);
      fclose(pfile);
      free(segs);
      return NULL;
    }
  }
    
  fclose(pfile);
  return segs;
}

BOOL
seglocWriteSeglocMap(char *filename, SEGLOC_MAP_HEADER *mapHeader,
		     SEGLOC_SEG *segs)
{
  FILE *pfile;
  int i, res;
  SEGLOC_SEG *seg;

  /* Ouverture fichier */
  if ((pfile = fopen(filename,"w")) == NULL) {
    printf ("seglocWriteSeglocMap: cannot open %s\n", filename);
    return FALSE;
  }
    
  /* Nombre de segments ? */
  res = fprintf(pfile, "%d \n", mapHeader->nbSegs);
  if (res <= 0) {
    printf ("seglocWriteSeglocMap: cannot write nbsegs (res = %d)\n", res);
    fclose(pfile);
    return FALSE;
  }

  /* Entete */
  res = fprintf(pfile, "0.0 0.0 0.0 0.0 0.0 0.0 T\n");
  if (res <= 0) {
    printf ("seglocWriteSeglocMap: cannot write header (res = %d)\n", res);
    fclose(pfile);
    return FALSE;
  }
  res = fprintf(pfile, "0.0 0.0 0.0 0.0\n");
  if (res <= 0) {
    printf ("seglocWriteSeglocMap: cannot write enc (res = %d)\n", res);
    fclose(pfile);
    return FALSE;
  }
	     
  seg=segs;

  for(i=0; i<mapHeader->nbSegs; i++,seg++) {
    
    /* Premier point */
    res = fprintf(pfile, "%.4f %.4f\n", 
	       seg->x1, seg->y1);
    if (res <= 0) {
      printf ("seglocWriteSeglocMap: cannot write first point (seg %d - res %d)\n", i, res);
      fclose(pfile);
      return FALSE;
    }
    /* second point */
    res = fprintf(pfile, "%.4f %.4f\n", 
	       seg->x2, seg->y2);
    if (res <= 0) {
      printf ("seglocWriteSeglocMap: cannot write second point (seg %d - res %d)\n",
	      i, res);
      fclose(pfile);
      return FALSE;
    }
    /* vecteur */
    res = fprintf(pfile, "0.0 0.0 0.0\n");
    if (res <= 0) {
      printf ("seglocWriteSeglocMap: cannot write vx ... (seg %d - res %d)\n",
	      i, res);
      fclose(pfile);
      return FALSE;
    }
    /* o,r */
    res = fprintf(pfile, "0.0 0.0\n");
    if (res <= 0) {
      printf ("seglocWriteSeglocMap: cannot write o, r (seg %d - res %d)\n",
	      i, res);
      fclose(pfile);
      return FALSE;
    }
    /* variances */
    res = fprintf(pfile, "0.0 0.0 0.0\n");
    if (res <= 0) {
      printf ("seglocWriteSeglocMap: cannot write variances ... (seg %d - res %d)\n",
	      i, res);
      fclose(pfile);
      return FALSE;
    }
    /* type extremite */
    res = fprintf(pfile, "B B\n");
    if (res <= 0) {
      printf ("seglocWriteSeglocMap: cannot write extremity types (seg %d - rs %d)\n", 
	      i, res);
      fclose(pfile);
      return FALSE;
    }
  }
    
  fclose(pfile);
  return TRUE;
}

BOOL
seglocWriteSegkitMap(char *filename, SEGKIT_MAP_HEADER *mapHeader,
		     SEGKIT_SEG *segs)
{
  FILE *pfile;
  int i, res;
  SEGKIT_SEG *seg;

  /* Ouverture fichier */
  if ((pfile = fopen(filename,"w")) == NULL) {
    printf ("seglocWriteSegkitMap: cannot open %s\n", filename);
    return FALSE;
  }
    
  /* Nombre de segments ? */
  res = fprintf(pfile, "%d \n", mapHeader->nbSegs);
  if (res <= 0) {
    printf ("seglocWriteSegkitMap: cannot write nbsegs (res = %d)\n", res);
    fclose(pfile);
    return FALSE;
  }

  /* Entete */
  res = fprintf(pfile, "%.4f %.4f %.5f %.5f %.5f %.5f %c\n", 
	     mapHeader->x, mapHeader->y, 
	     mapHeader->cs, mapHeader->sn,
	     mapHeader->cst, mapHeader->snt, 
	     mapHeader->coosys);
  if (res <= 0) {
    printf ("seglocWriteSegkitMap: cannot write header (res = %d)\n", res);
    fclose(pfile);
    return FALSE;
  }
  res = fprintf(pfile, "%.4f %.4f %.4f %.4f\n", 
	     mapHeader->enc[0], mapHeader->enc[1],
	     mapHeader->enc[2], mapHeader->enc[3]);
  if (res <= 0) {
    printf ("seglocWriteSegkitMap: cannot write enc (res = %d)\n", res);
    fclose(pfile);
    return FALSE;
  }
	     
  seg=segs;

  for(i=0; i<mapHeader->nbSegs; i++,seg++) {
    
    /* Premier point */
    res = fprintf(pfile, "%.4f %.4f\n", 
	       seg->x1, seg->y1);
    if (res <= 0) {
      printf ("seglocWriteSegkitMap: cannot write first point (seg %d - res %d)\n", i, res);
      fclose(pfile);
      return FALSE;
    }
    /* second point */
    res = fprintf(pfile, "%.4f %.4f\n", 
	       seg->x2, seg->y2);
    if (res <= 0) {
      printf ("seglocWriteSegkitMap: cannot write second point (seg %d - res %d)\n",
	      i, res);
      fclose(pfile);
      return FALSE;
    }
    /* vecteur */
    res = fprintf(pfile, "%.7f %.7f %.7f\n", 
	       seg->vx, seg->vy, seg->d);
    if (res <= 0) {
      printf ("seglocWriteSegkitMap: cannot write vx ... (seg %d - res %d)\n",
	      i, res);
      fclose(pfile);
      return FALSE;
    }
    /* o,r */
    res = fprintf(pfile, "%.7f %.7f\n", 
	       seg->r, seg->o);
    if (res <= 0) {
      printf ("seglocWriteSegkitMap: cannot write o, r (seg %d - res %d)\n",
	      i, res);
      fclose(pfile);
      return FALSE;
    }
    /* variances */
    res = fprintf(pfile, "%.7f %.7f %.7f\n", 
	       seg->vr, seg->vro, seg->vo);
    if (res <= 0) {
      printf ("seglocWriteSegkitMap: cannot write variances ... (seg %d - res %d)\n",
	      i, res);
      fclose(pfile);
      return FALSE;
    }
    /* type extremite */
    res = fprintf(pfile, "%c %c\n", 
	       seg->typ1, seg->typ2);
    if (res <= 0) {
      printf ("seglocWriteSegkitMap: cannot write extremity types (seg %d - rs %d)\n", 
	      i, res);
      fclose(pfile);
      return FALSE;
    }
  }
    
  fclose(pfile);
  return TRUE;
}
