/**
 ** 2 functions to clip map, ie, to remove segments out of a given rectangle.
 **
 ** seglocClipMap : the real, nice one
 **
 ** seglocSimpleClipMap :
 **  A very simple and approximative cliping ! (but conservative)
 **  for now just remove segments for which both extremities are
 **  outside the boundaries and in the same side of a boudary
 **
 **/


#include <stdio.h>
#include <sys/types.h>
#ifndef VXWORKS
#include <time.h>
#include <termios.h>
#endif
#include <unistd.h>
#include <stdlib.h>

#include "h2mathLib.h"
#include "libpolygons/polygons.h"

#include <dbliste.h>
#include <lasdef.h>
#include <las_struct_p.h>
/* #include "las_prot.h" */
/* #include "image_mgr_p.h" */
#include <las_io_p.h>

static int Xmin, Xmax, Ymin, Ymax;
static int testPointIn(struct pt2d *point);

int seglocClipMap(igla *igin, 
		  double xmin, double xmax, 
		  double ymin, double ymax)
{
  PDBL *l, *l2;
  Sgl *sinput;
  int remove = 0;
  double ptSeg1[2];
  double ptSeg2[2];
  double ptRec[4];

  int clip;
  ptRec[0] = xmin;
  ptRec[1] = ymin;
  ptRec[2] = xmax;
  ptRec[3] = ymax;

  if (xmin > xmax || ymin > ymax) {
    printf ("seglocClipMap: incoherent data\n");
    return FALSE;
  }

  Xmin = xmin;
  Ymin = ymin;
  Xmax = xmax;
  Ymax = ymax;

  /* Parcous les segments de la carte */
  l = igin->lsgl;
  do {

    sinput = (Sgl *)l->el;

    ptSeg1[0] = sinput->first.x;
    ptSeg1[1] = sinput->first.y;
    ptSeg2[0] = sinput->last.x;
    ptSeg2[1] = sinput->last.y;

    /* Hors du rectangle : on enleve le segment */
    if (!(clip = PLclipSegment(ptSeg1, ptSeg2, ptRec, ptSeg1, ptSeg2))) {
      l2 = dls(l);
      igin->lsgl = dbdel_el(igin->lsgl,l,Sgl_free);
      igin->nsg -= 1;
      remove++;
      l = l2;
      continue;
    }
    /* Clip */
    else if (clip == 2) {
      /* On coupe le point 1 du segment */
      sinput->first.x = ptSeg1[0];
      sinput->first.y = ptSeg1[1];
      sinput->first.r = sqrt(SQR(sinput->first.x) + SQR(sinput->first.y));
      
      /* On coupe le point 2 du segment */
      sinput->last.x = ptSeg2[0];
      sinput->last.y = ptSeg2[1];
      sinput->last.r = sqrt(SQR(sinput->last.x) + SQR(sinput->last.y));
    }

    l = dls(l);
  } while (l != NULL);

  printf ("Removed %d segs (remain %ld)\n", remove, igin->nsg);

  return TRUE;
}

int seglocPolygonClipMap(igla *igin,
			 int nbPts, double *poly)
{
  PDBL *l, *l2;
  Sgl *sinput;
  int remove = 0;
  double pt1[2];
  double pt2[2];

  int clip;

  /* Parcous les segments de la carte */
  l = igin->lsgl;
  do {

    sinput = (Sgl *)l->el;

    pt1[0] = sinput->first.x;
    pt1[1] = sinput->first.y;
    pt2[0] = sinput->last.x;
    pt2[1] = sinput->last.y;

    /* 2 extremites hors polygone : on enleve */
    if (!PLpointInPolygon(pt1, nbPts, poly) &&
	!PLpointInPolygon(pt2, nbPts, poly)) {
      l2 = dls(l);
      igin->lsgl = dbdel_el(igin->lsgl,l,Sgl_free);
      igin->nsg -= 1;
      remove++;
      l = l2;
      continue;
    }

    l = dls(l);
  } while (l != NULL);

  printf ("Removed %d segs (remain %ld)\n", remove, igin->nsg);

  return TRUE;
}

int seglocRemoveRectangle(igla *igin, 
			  double xmin, double xmax, 
			  double ymin, double ymax)
{
  PDBL *l, *l2;
  Sgl *sinput;
  int remove = 0;
  double ptSeg1[2];
  double ptSeg2[2];
  double ptRec[4];

  int clip;
  ptRec[0] = xmin;
  ptRec[1] = ymin;
  ptRec[2] = xmax;
  ptRec[3] = ymax;

  if (xmin > xmax || ymin > ymax) {
    printf ("seglocRemoveRectangle: incoherent data\n");
    return FALSE;
  }

  Xmin = xmin;
  Ymin = ymin;
  Xmax = xmax;
  Ymax = ymax;

  /* Parcous les segments de la carte */
  l = igin->lsgl;
  do {

    sinput = (Sgl *)l->el;

    ptSeg1[0] = sinput->first.x;
    ptSeg1[1] = sinput->first.y;
    ptSeg2[0] = sinput->last.x;
    ptSeg2[1] = sinput->last.y;

    /* clip le segment */
    clip = PLclipSegment(ptSeg1, ptSeg2, ptRec, ptSeg1, ptSeg2);

    /* Dans le rectangle : on enleve */
    if (clip == 1) {
      l2 = dls(l);
      igin->lsgl = dbdel_el(igin->lsgl,l,Sgl_free);
      igin->nsg -= 1;
      remove++;
      l = l2;
      continue;
    }
    /* Clip */
    else if (clip == 2) {
      /* On coupe le point 2 du segment */
      sinput->last.x = ptSeg1[0];
      sinput->last.y = ptSeg1[1];
      sinput->last.r = sqrt(SQR(sinput->last.x) + SQR(sinput->last.y));
      
      /* On coupe le point 1 du segment */
      sinput->first.x = ptSeg2[0];
      sinput->first.y = ptSeg2[1];
      sinput->first.r = sqrt(SQR(sinput->first.x) + SQR(sinput->first.y));
    }

    /* next */
    l = dls(l);
  } while (l != NULL);

  printf ("Removed %d segs (remain %ld)\n", remove, igin->nsg);

  return TRUE;
}

 /** Simple clipping: just remove segments for which both extremities are
  **  outside the boundaries and in the same side of a boudary
  **/
igla * seglocSimpleClipMap(igla *igin, 
			   double xmin, double xmax, 
			   double ymin, double ymax)
{
  igla *igout;
  PDBL *l;
  Sgl *sinput;
  Sgl *sout;
  int i;
  int remove = 0;

  igout = Imgl_cree();

  Xmin = xmin;
  Ymin = ymin;
  Xmax = xmax;
  Ymax = ymax;

  /* Recopie ce qui ne change pas */
  igout->x = igin->x;
  igout->y = igin->y;
  igout->cs = igin->cs;
  igout->sn = igin->sn;
  igout->cst = igin->cst;
  igout->snt = igin->snt;
  igout->coosys = igin->coosys;

  for (i=0; i<4; i++)
    igout->enc[i] = igin->enc[i];

  /* Parcous les segments de la carte */
  for(l = igin->lsgl; l != NULL; l = dls(l)){
    
    sinput = (Sgl *)dlel(l);

    /* Les cas triviaux :
       les 2 extremites du segments sont hors des limites ET du mme cot */
    if ((sinput->first.x < xmin && sinput->last.x < xmin) ||
        (sinput->first.x > xmax && sinput->last.x > xmax) ||
        (sinput->first.y < ymin && sinput->last.y < ymin) ||
        (sinput->first.y > ymax && sinput->last.y > ymax)) {
      remove++;
      continue;
    }
        
    /* Et on garde le reste */
    /* XXXX faudra peut etre faire attention aux types des extremites */
    sout = Sgl_cree();
    sout = sinput;
    igout->lsgl = dlhcons(igout->lsgl,sout);
    igout->nsg = igin->nsg - remove;
  } /* for */

  printf ("Removed %d segs (remain %ld)\n", remove, igout->nsg);

  return igout;
}

