/**************************************************************************
 *
 * testIntersect -
 *
 * Test functions:  polygonInsidePolygon and polygonInterPolygon
 *
 * The 2 polygons have to be describe in a file named 
 * "polygons" that must have the following format (H2 obstacles format) :
 *
 *      OBSTACLES
 *      x1 y1                      (1st polygon coordinates)
 *      x2 y2
 *      ...
 *      xn yn
 *      END
 *      x1 y1                      (2nd polygon coordinates)
 *      x2 y2
 *      ...
 *      xm ym
 *      END
 *
 *  WARNING: polygon intersection means intersection of their surfaces.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include "polygons.h"


int main()
{
  char *filename = "polygons";
  FILE *pfile;
  double *poly1, *poly2;
  int nb1, nb2;
  int numPoly=0;

  /*
   * Lectures des polygones dans le fichier
   */
  if ((pfile = fopen(filename, "r")) == NULL) {
    printf ("Can not open file \"polygons\"\n");
    return -1;
  }


  if (!(poly1 = PLscanOneH2Polygon(pfile, &nb1))) {
    printf ("Can not read 1st polygon\n");
    return -1;
  }
  printf ("-- POLYGON %d: ", numPoly);
  PLprintPolygon(stdout, nb1, poly1, 1);
  printf ("\n");
  numPoly++;

  do {

    /* Read next polygon */
    if (!(poly2 = PLscanOneH2Polygon(pfile, &nb2))) {
      printf ("No more polygon\n");
      free(poly1);
      fclose(pfile);
      return 0;
    }
    printf ("-- POLYGON %d: ", numPoly);
    PLprintPolygon(stdout, nb2, poly2, 1);
    
    /* Tests */
    printf ("\n");
    if (PLpolygonInsidePolygon(nb1, poly1, nb2, poly2))
      printf ("  Polygon %d INSIDE polygon %d\n", numPoly-1, numPoly);
    else
      printf ("  Polygon %d NOT INSIDE polygon %d\n", numPoly-1, numPoly);
    
    if (PLpolygonInsidePolygon(nb2, poly2, nb1, poly1))
      printf ("  Polygon %d INSIDE polygon %d\n", numPoly, numPoly-1);
    else
      printf ("  Polygon %d NOT INSIDE polygon %d\n", numPoly, numPoly-1);
    
    if(PLpolygonInterPolygon(nb1, poly1, nb2, poly2))
      printf ("  Polygon %d and %d INTERSECT\n", numPoly-1, numPoly);
    else
      printf ("  Polygon %d and %d DO NOT INTERSECT\n", numPoly-1, numPoly);
    printf ("\n");

    /* Next */
    free(poly1);
    numPoly++;
    poly1 = poly2;
    nb1 = nb2;
  } while(1);

}
