#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <h2mathLib.h>
#include "seglocUtils.h"

static int dms_plus(const double *m1, const double *m2, double *m_res, 
		int l, int c);
static int dmfsf_jmult2(const double *m1, const double *m2, double *m_res, 
		int l1, int c1, int l2, int c2);

#define MATELT(mat,i,j,c)   ( *((mat) + (c) * (i) + (j)) )
#define INDEX_SMATELT(i,j)   \
((i)<(j) ? ((j)*((j)+1))/2 + (i) : ((i)*((i)+1))/2 +(j)) 
#define SMATELT(mat,i,j,n) *((mat) + INDEX_SMATELT((i),(j)))
#define MATSIZE(l,c)        ((l) * (c))
#define SMATSIZE(n,c)       (((n) * ((n) + 1)) / 2)

static int 
dms_plus(const double *m1, const double *m2, double *m_res, int l, int c)
{
	int k;

	if (l != c)
	{
		fprintf (stderr, "dms : matrice symetrique non carree !!!\n");
		return -1;
	}

	for (k = 0; k < SMATSIZE(l, c); k++)
		*m_res++ = *m1++ + *m2++;
	return 0;
}

static int
dmfsf_jmult2 (const double *m1, const double *m2, double *m_res,
		int l1, int c1, int l2, int c2)
{

	double *temp, *res;
	double *aux = NULL;
	double *aux2 = NULL;
	int i,j,k,flag = 0;
	double val;

	if (l2 != c2 || c1 != c2)
	{
		fprintf(stderr, "%s : tailles incorrectes \n", __func__);
		return -1;
	}

	if (m_res == m1 || m_res == m2)
	{
		temp = malloc((l1*c2 + l1*(l1+1)/2) * sizeof(double));
		if (temp == NULL) {
			fprintf(stderr, "%s : Can't allocate memory for temp\n", __func__);
			return -1;
		}
		aux = temp;
		res = temp + l1*c2;
		aux2 = res;
		flag = 1;
	}
	else
	{
		temp = malloc((l1 * c2) * sizeof(double));
		if (temp == NULL) {
			fprintf(stderr, "%s : Can't allocate memory for temp\n", __func__);
			return -1;
		}
		aux = temp;
		res = m_res;
	}


	for (i = 0; i < l1; i++)
		for (j = 0; j < c2; j++)
		{
			val = 0.;
			for (k = 0; k < c1; k++)
				val += MATELT (m1, i, k, c1) * SMATELT (m2, k, j, c2);
			*temp++ = val;
		}
	temp = aux;

	for (j = 0; j < l1; j++)
		for (i = 0; i <= j; i++)
		{
			val = 0.;
			for (k = 0; k < c2; k++)
				val += MATELT (temp, i, k, c2) * MATELT (m1, j, k, c1);
			*res++ = val;
		}

	if (flag)
	{
		res = aux2;
		for (i = 0; i < SMATSIZE (l1, l1); i++)
			*m_res++ = *res++;
	}

	free(temp);

	return 0;
}

void seglocConvertVar2Matrix(double vX[6], SEGLOC_VARIANCES *var)
{
  vX[0] = var->vxx;
  vX[1] = var->vxy;
  vX[2] = var->vyy;
  vX[3] = var->vxt;
  vX[4] = var->vyt;
  vX[5] = var->vtt;
}
void seglocConvertMatrix2Var(SEGLOC_VARIANCES *var, double vX[6])
{
  var->vxx = vX[0];
  var->vxy = vX[1];
  var->vyy = vX[2];
  var->vxt = vX[3];
  var->vyt = vX[4];
  var->vtt = vX[5]; 
}


/**
 **  CalcAbsPos -  
 **
 **
 **/

void CalcAbsPos(SEGLOC_POS_VAR *odo, 
		SEGLOC_POS_VAR *transf,
		SEGLOC_POS_VAR *pos)
{
  double rob_enc[3];
  double rob_encV[6];
  double tr[3];
  double trV[6];
  double posCorr[3];
  double posCorrV[6];

  rob_enc[0] = odo->pos.x;
  rob_enc[1] = odo->pos.y;
  rob_enc[2] = odo->pos.theta;

  rob_encV[0] = odo->var.vxx;
  rob_encV[1] = odo->var.vxy;
  rob_encV[2] = odo->var.vyy;
  rob_encV[3] = odo->var.vxt;
  rob_encV[4] = odo->var.vyt;
  rob_encV[5] = odo->var.vtt;

  tr[0] = transf->pos.x;
  tr[1] = transf->pos.y;
  tr[2] = transf->pos.theta;

  trV[0] = transf->var.vxx;
  trV[1] = transf->var.vxy;
  trV[2] = transf->var.vyy;
  trV[3] = transf->var.vxt;
  trV[4] = transf->var.vyt;
  trV[5] = transf->var.vtt;

  calc_abs_pos(rob_enc, rob_encV, tr, trV, posCorr, posCorrV);

  pos->pos.x = posCorr[0];
  pos->pos.y = posCorr[1];
  pos->pos.theta = posCorr[2];

  /*  */
  pos->var.vxx = posCorrV[0];
  pos->var.vxy = posCorrV[1];
  pos->var.vyy = posCorrV[2];
  pos->var.vxt = posCorrV[3];
  pos->var.vyt = posCorrV[4];
  pos->var.vtt = posCorrV[5];
}

/**
 **
 ** Fonction tire de $SEGKIT_DIR/mapkit/src/odom_funct.c
 **
 **/
void calc_abs_pos(double xrob[3], double vrob[6],
		  double tr[3], double vtr[6],
		  double abs[3], double vabs[6])
{
  double Jo2a[9] = {1.,0.,0.,
		    0.,1.,0.,
		    0.,0.,1.};

  double Jr2a[9] = {1.,0.,0.,
		    0.,1.,0.,
		    0.,0.,1.};

  double aux[9];

  double  cloc2abs,sloc2abs;

  cloc2abs = cos(tr[2]);
  sloc2abs = sin(tr[2]);

  abs[0] =  cloc2abs * xrob[0] - sloc2abs * xrob[1] + tr[0]; 
  abs[1] =  sloc2abs * xrob[0] + cloc2abs * xrob[1] + tr[1];

  abs[2] =  angleLimit(xrob[2] + tr[2]);
  if(vtr){
    Jr2a[0] = cloc2abs; Jr2a[1] = -sloc2abs;
    Jr2a[3] = sloc2abs; Jr2a[4] = cloc2abs;
    
    Jo2a[2] = tr[1] - abs[1];
    Jo2a[5] = abs[0] - tr[0];

    dmfsf_jmult2(Jr2a,vrob,aux,3,3,3,3);
    dmfsf_jmult2(Jo2a,vtr,vabs,3,3,3,3);
    
    /* les 2 et 3 eme parametre etaient vtr mais c'est vabs qu'on calcul ??? */
    dms_plus(aux,vabs,vabs,3,3);
  }
}

