#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>

#include "h2mathLib.h"
#include <portLib.h>
#include <h2timerLib.h>

#ifdef H2
#include "locoStruct.h"
#endif

#ifdef XR4000
#include "xr4000Struct.h"
#include "locoCompat.h"
#endif

#if defined(ATRV) || defined(B21R) || defined(JIDO)
#include "rflexStruct.h"
#include "locoCompat.h"
#endif

#include "auto/gridPosterLib.h"
#include "gridConst.h"
#include "gridStruct.h"

#include "auto/us2matPosterLib.h"
#include "us2matConst.h"
#include "us2matStruct.h"

#include "affSegloc.h"
#include "affSeglocOptions.h"

#include "dessin_seg_p.h"

/* typedef struct COL_RGB { */
/*   unsigned int blue : 8; */
/*   unsigned int green : 8; */
/*   unsigned int red : 8; */
/* } COL_RGB; */

#define COL_RGB int

BOOL affUs2MatProbaGrid(US2MAT_PROBAGRID_POSTER_STR *grid, 
			int followFlag,
			int colorGridTab [NB_GRID_COUL][NB_GRID_COUL],
			GC mygc)
{
  int i, j;
  int posi, posj;
  double posx, posy, startPosx, startPosy;
  int sizei, sizej;
  double size;
  int occupe, vide, inconnu;
  int colRgb=0;
  int prevColRgb=-1;
  int red, green;
  int draw;

  /* Dimensions d'une cellule */
  size = grid->gridParams.cellSize;
  sizei = r_m_to_pixel(size);
  sizej = sizei;
  if (sizei==0) {
    printf ("affUs2MatProbaGrid: small cell\n");
    return FALSE;
  }

  /* Position premire cellule qui se trouve en bas a gauche */
  startPosx = grid->gridPosition.x - grid->gridParams.xGridSize/2.0;
  startPosy = grid->gridPosition.y - grid->gridParams.yGridSize/2.0;

  /* Parcours des lignes de cellule */
  for (i=0; i<grid->gridParams.nbXCells; i++) {

    /* Ordonnee de la ligne consideree */
    posy = startPosy + i * grid->gridParams.cellSize;

    /* Parcours des cellules de chaque ligne */
    for (j=0; j<grid->gridParams.nbYCells; j++) {  
      
      occupe = grid->probaGrid.occupe[i][j];
      inconnu = grid->probaGrid.inconnu[i][j];
      vide = 100 - (occupe+inconnu);
      
      /* Incoherent : bleu */
      if (occupe  < 0 || inconnu < 0 || vide < 0 ||
	  occupe > 100 || inconnu > 100 || vide > 100) {
	printf ("cell(%d,%d) = %d %d %d\n", i,j, occupe, vide, inconnu);
	red = 0;
	green = 0;
	draw = TRUE;
      }
      /* Inconnu : on ne dessine rien */
      else if (inconnu >= 99) {
	draw = FALSE;
      }
      /* Cellule valide: calcul couleur */
      else {
	/* Plus c'est occup plus c'est rouge 
	   (valeur entre 0 et NB_GRID_COUL) */
	red = (int)(((double)occupe)/NB_GRID_COUL);
	if (red<0) red = 0;
	if (red>=NB_GRID_COUL) red = NB_GRID_COUL-1;

	/* Plus c'est libre plus c'est vert 
	   (valeur entre 0 et NB_GRID_COUL) */
	green = (int)(((double)vide)/NB_GRID_COUL);
	if (green<0) green = 0;
	if (green>=NB_GRID_COUL) green = NB_GRID_COUL-1;
	draw = TRUE;
      }
      
      /* Dessin */
      if (draw) {

	/* Activation couleur */
	colRgb = colorGridTab[red][green];
	if (colRgb != prevColRgb) {
	  XSetForeground (dpy, mygc, colRgb);
	  prevColRgb = colRgb;
	}
      
	/* Absice de la cellule */
	posx = startPosx + j * grid->gridParams.cellSize;

	/* Coord pixels de la premiere cellule */
	metrePix(posx, posy, &posi, &posj);
	XFillRectangle (dpy, win, mygc, posi, posj, sizei, sizej);
      }
    }  /* for j */
  } /* for i */
  return TRUE;
}


BOOL affUs2MatTernarGrid(US2MAT_GRID_POSTER_STR *grid, 
			 int followFlag,
			 unsigned long c_occupee,
			 unsigned long c_libre, 
			 unsigned long c_inconnu,
			 GC mygc)
{
  int i, j;
  int posi, posj;
  double posx, posy, startPosx, startPosy;
  int sizei, sizej;
  double size;
  US2MAT_CELL_TYPE type;
  US2MAT_CELL_TYPE prevType = -1;
  
  /* Dimensions d'une cellule */
  size = grid->gridParams.cellSize;
  sizei = r_m_to_pixel(size);
  sizej = sizei;
  if (sizei==0) {
    printf ("affUs2MatTernarGrid: small cell\n");
    return FALSE;
  }

  /* Position premire cellule qui se trouve en bas a gauche */
  startPosx = grid->gridPosition.x - grid->gridParams.xGridSize/2.0;
  startPosy = grid->gridPosition.y - grid->gridParams.yGridSize/2.0;

  /* Parcours des cellules */
  for (i=0; i<grid->gridParams.nbXCells; i++) {

    /* Ordonnee de la ligne consideree */
    posy = startPosy + i * grid->gridParams.cellSize;

    for (j=0; j<grid->gridParams.nbYCells; j++) {  
      
      type = grid->grid.matrice[i][j];

      /* On ne dessine pas l'inconnu */
      if (type != US2MAT_CELL_UNKNOWN) {

	/* Le type de cellule ou la ligne change : on dessine prevType */
	if (type != prevType) {
	  switch (type) { 
	  case US2MAT_CELL_FULL : 
	    XSetForeground (dpy, mygc, c_occupee);
	    prevType = type;
	    break;
	  case US2MAT_CELL_FREE : 
	    XSetForeground (dpy, mygc, c_libre);
	    prevType = type;
	    break;
	  default:
	    printf ("affUs2MatTernarGrid: unknown cell(%d,%d) type %d \n",
		    i,j,type);
	    break;
	  }
	}

	/* Absice de la cellule */
	posx = startPosx + j * grid->gridParams.cellSize;

	/* Coord pixels de la premiere cellule */
	metrePix(posx, posy, &posi, &posj);
	XFillRectangle (dpy, win, mygc, posi, posj, sizei, sizej);
      }
    }  /* for j */
  } /* for i */
  return TRUE;
}





/*-----------------------------------------------------------------*/
void affiche_matrice_resultat (GRID_GRID_POSTER_STR *grid,
			       unsigned long c_occupee,
			       unsigned long c_libre, 
			       unsigned long c_mysterieux, 
			       unsigned long c_inconnu,
			       GC mygc) 
{
  int i, j;
  int posi, posj;
  double posx, posy;
  int sizei, sizej;
  double sizex, sizey;
  GRID_CELL_TYPE type, prevType=GRID_CELL_UNKNOWN;
  int nbSameType = 0;
  
  /* Dimensions d'une cellule */
  sizex = grid->params.xCellSize;
  sizey = grid->params.yCellSize;
  sizei = r_m_to_pixel(sizex);
  sizej = r_m_to_pixel(sizey);

  /* Position premire cellule */
  posx = grid->gridPos.x;
  posy = grid->gridPos.y;
  type = grid->grid.cells[0][0].type;

  /* Parcours des cellules */
  for (i=0; i<grid->params.nbXCells; i++) {
    for (j=0; j<grid->params.nbYCells; j++) {  
      
      nbSameType++;
      type = grid->grid.cells[i][j].type;

      /* Le type de cellule ou la ligne change : on dessine prevType */
      if (1||type != prevType || j == grid->params.nbYCells-1) {

	switch (prevType) { 
	case GRID_CELL_FULL : 	  
	  XSetForeground (dpy, mygc, c_occupee);
	  break;
	case GRID_CELL_FREE : 
	  XSetForeground (dpy, mygc, c_libre);
	  break;
	case GRID_CELL_UNKNOWN : 
	  XSetForeground (dpy, mygc, c_inconnu);
	  break;
	case GRID_CELL_MYSTERIOUS :
	  XSetForeground (dpy, mygc,c_mysterieux);
	  break;
	default:
	  break;
	}

	/* Coord pixels de la premiere cellule */
	metrePix(posx, posy, &posi, &posj);
	XFillRectangle (dpy, win, mygc, posi, posj, sizei, nbSameType*sizej);

	/* Position premiere cellule de la serie suivante */
	posx = grid->gridPos.x + j*grid->params.xCellSize;
	posy = grid->gridPos.y + (grid->params.nbXCells-i)*grid->params.yCellSize;

	prevType = type;
	nbSameType=0;
      }


    }  /* for j */
  } /* for i */
}

