#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <sys/time.h>
#include <ctype.h>

#include "h2mathLib.h"

#include <portLib.h>
#include <taskLib.h>
#include "h2initGlob.h"
#include <h2timerLib.h>

#ifdef H2
#include "locoStruct.h"
#include "usStruct.h"
#include "usXR4000Compat.h"
#endif

#ifdef XR4000
#include "xr4000Struct.h"
#include "locoCompat.h"
#endif

#if defined(ATRV) || defined(B21R)
#include "rflexStruct.h"
#include "server/rflexPosterLib.h"
#include "locoCompat.h"
#include "usXR4000Compat.h"
#endif

#include "affSegloc.h"
#include "affSeglocOptions.h"
#include "dessin_seg.h"
#include "ellipses_p.h"

#include "../server/seglocPosterLib.h"
#include "server/sickPosterLib.h"
#include "sick_save.h"

#ifdef M2D
#include "server/m2dPosterLib.h"
#endif

#ifdef VSTP
#include "server/vstpPosterLib.h"
#endif

#ifdef ASPECT
#include "server/aspectPosterLib.h"
#endif

#ifdef GRID
#include "server/gridPosterLib.h"
/* #include "server/us2matPosterLib.h" */
#endif

#ifdef ZONE
#include "zoneIO.h"
#include "zoneStruct.h"
#endif

#include "affSeglocCommProto.h"

#include "mat.h" /* mxfs_mmulttr, mxs_plus */
#include "libutils/seglocUtils.h"
#include "libmap/seglocMapIO.h"

#ifndef min
#define min(x,y) (x>y ? y : x)
#endif

/* ----------------------- Variables globales ----------------------- */

/* Parametres fenetre affichage */
Window win;
Display *dpy=NULL;
GC gc, gcgris;
int screen;

/* carte */
char mapFileName[256]="\0";
static BOOL readMap=FALSE;

#ifdef ZONE
/* zones */
char zonesFileName[256]="\0";
#endif
static BOOL readZones=FALSE;

/* Fichier donnees laser */
char laserBaseFileName[256]="\0";
int laserFileBaseNum = 0;
int laserFileNum = 0;

/* poster position */
char pomMEposPosterName[64];

/* Tableaux des coord des obstacles rectangulaires */
double xMapObst[SEGLOC_MAX_MAP_SEGS*4];
double yMapObst[SEGLOC_MAX_MAP_SEGS*4];

/* Tous les flags et leur valeur par defaut */
static DRAW_FLAGS drawFlags;

/* couleur */
static XColor colorTab[MAX_COLOR];

static  int colorGridTab[NB_GRID_COUL][NB_GRID_COUL];

#define STD_EVENT_MASK (ButtonPressMask | KeyPressMask | \
			StructureNotifyMask | ExposureMask | \
			ButtonReleaseMask | \
                        ButtonMotionMask )

/* ------------------ Prototypes des fonctions locales ------------------ */
#ifdef USRFLEX
static void dessin_points_us_rflex(SEGLOC_POS *posRobot, 
				   RFLEX_SONARSCART_POSTER_STR *us,
				   unsigned long colTop,
				   unsigned long colDown,
				   GC gc);
#endif

extern void dessin_robot_rond_segloc(SEGLOC_POS_VAR *pos, double radius, BOOL drawUnc, GC gc);

static unsigned long get_mtime (void);
static int enlargeSegs(int nbSegs, SEGLOC_SEG *mapSegs,  
		       double *xMapObst, double *yMapObst);

static int draw_all(GC gc);

static void dessin_segloc_segments(SEGLOC_POS *segpos, int nbsegs, 
				   SEGLOC_SEG *segs, BOOL extremity, GC gc);

static int readMapSegs(SEGLOC_MAP *mapSegs,  
		       double *xMapObst, double *yMapObst);

static int readMapZoneSegs(SEGLOC_MAP *mapSegs,  
			   double *xMapObst, double *yMapObst);

#if defined M2D || defined VSTP
static BOOL draw_m2d_trajectory (GC gc);
#endif

static void draw_uncertainties(XAFF_POS *pos, SEGLOC_VARIANCES *svar,
			       GC gc);

static void dessin_point_laser(SEGLOC_POS *posRobot, 
			       SICK_MEASURES_HEADER_STR *header, 
			       double rho[SICK_N_PT_MAX],
			       GC gc);

#ifdef US
static void dessin_points_us_odo(XR4000_MEASURES_IN_GLOBAL_CART_TAB *us,
				 unsigned long colTop,
				 unsigned long colDown,
				 GC gc);
static void dessin_points_us_segloc(SEGLOC_POS *posRobot, 
				    XR4000_MEASURES_IN_ROBOT_CART_TAB *us,
				    unsigned long colTop,
				    unsigned long colDown,
				    GC gc);
#endif

static BOOL readLaserPoints(char *baseFileName, int fileNum,
			    SICK_MEASURES_HEADER_STR *header, 
			    double rho[SICK_N_PT_MAX], 
			    SEGLOC_POS_VAR *pos);

#ifdef ZONE
void dessin_zones(ZONES_MAP *zones, GC gc);
#endif

/* -------------------------------  MAIN  ------------------------------- */

/* Main */
int
main(int argc, char **argv)
{
  char *dpy_name=NULL;
  XWindowAttributes win_att;
  int window_x = XSIZE;
  int window_y = YSIZE;
  int window_width  = XSIZE;
  int window_height = YSIZE/2;
  int border_width = 5;
  XEvent event;
  XSizeHints   size_hints;
  int end = 0;
  H2TIMER_ID timer;
  int timerPeriod = 10;
  unsigned long debut, fin_period, fin_calcul;
  int init_decompte = 5;
  int decompte = init_decompte;
  int firsttime = 1;
  int numarg;

  static const char *help = HELP;
  static const char *usage = "affSegloc [periode en tics (defaut: %d tics)] [-f <fichier segloc>]  [-l <sick file name> <sick file num>]\n";

  /** ------------------------------------------------------------
   ** TRAITEMENT DES ARGUMENTS 
   **/

  numarg = 1;
  while(numarg<argc) {
    /* period */
    if (isdigit(argv[numarg][0])) {
      timerPeriod = atoi(argv[numarg]);
      numarg++;
    }
    /* help */
    else if (argv[numarg][0] == 'h') {
      printf (usage, timerPeriod);
      printf(help);
      exit(0);
    }
    /* option */
    else if (argv[numarg][0] == '-') {
      switch(argv[numarg][1]) {
	/* help */
      case 'h':
	printf (usage, timerPeriod);
	printf(help);
	exit(0);
	/* map file */
      case 'f' :
	strcpy(mapFileName, argv[numarg+1]);
	printf ("mapFileName: %s\n", mapFileName);
	drawFlags.drawSeglocFlag = FALSE;
	drawFlags.drawAcqSegsFlag = FALSE;
	drawFlags.drawAspectSegsFlag = FALSE;
	drawFlags.drawAspectMapFlag = FALSE;
	drawFlags.drawMatchedPosFlag = FALSE;
	drawFlags.drawMatchedAcqSegFlag = FALSE;
	drawFlags.drawMatchedModelSegFlag = FALSE;
	drawFlags.drawUsSeglocFlag = FALSE;
	break;
	/* zone file */
      case 'z' :
#ifdef ZONE
	strcpy(zonesFileName, argv[numarg+1]);
	printf ("zonesFileName: %s\n", zonesFileName);
#else
	printf ("WARNING: To display ZONE, edit Makefile.in and recompile...\n");
#endif
	break;
      case 'l' :
	strcpy(laserBaseFileName, argv[numarg+1]);
	printf ("laserBaseFileName: %s\n", laserBaseFileName);
	numarg += 1;
	laserFileBaseNum = atoi(argv[numarg+1]);
	printf ("laserFileBaseNum: %d\n", laserFileBaseNum);
	break;	

      case 'p' :
	strcpy(pomMEposPosterName, argv[numarg+1]);
	printf ("pomMEposPosterName: %s\n", pomMEposPosterName);
	break; 

      default:
	printf (usage, timerPeriod);
	printf(help);
	exit(0);
      }
      numarg += 2;
    }
    else {
      printf (usage, timerPeriod);
      printf(help);
      exit(0);
    }
  } /* while arg */


  /* Defini la valeur d'un tic (en nombre de tic par seconde) */
#ifndef SOLARIS
  h2initGlob(20);
#endif


  /** ------------------------------------------------------------
   ** CREATION DE FENETRE 
   **/

  printf ("XOpenDisplay...");

  dpy = XOpenDisplay(dpy_name);
  if ( dpy == NULL ) {
    fprintf(stderr, "Error Open Display!!\n");
    exit(-1);
  }
  printf ("dpy=0x%x\n", (unsigned int)dpy);

  screen = DefaultScreen(dpy);
  XGetWindowAttributes(dpy, RootWindow(dpy, screen), &win_att);

  win = XCreateSimpleWindow(dpy, RootWindow(dpy, screen),
			    window_x, window_y, window_width, window_height,
			    border_width,    
			    BlackPixel(dpy, screen),
			    WhitePixel(dpy, screen));
  size_hints.flags = PSize;
/*   size_hints.flags = PPosition | PSize | PMinSize; */
  size_hints.x = window_x;
  size_hints.y = window_y;
  size_hints.width  = window_width;
  size_hints.height = window_height;
  size_hints.min_width  = XSIZE + 2 * border_width;
  size_hints.min_height = YSIZE + 2 * border_width;
  XSetStandardProperties(dpy, win, "SEGKIT", "SEGKIT",
			 None, argv, argc, &size_hints); 

  XSelectInput (dpy, win, STD_EVENT_MASK);
  XMapWindow(dpy, win);

  gc = XCreateGC(dpy, win, 0, NULL);
  XSetForeground(dpy, gc, BlackPixel(dpy, screen));
  gcgris = XCreateGC(dpy, win, 0, NULL);

  /** ------------------------------------------------------------
   ** TIMER
   **/

#if 0
#ifndef SOLARIS
  timer = h2timerAlloc ();
  if (!timer) {
    fprintf (stderr, "Cannot alloc timer!\n");
    XDestroyWindow(dpy, win);
    XCloseDisplay(dpy);
    exit (-1);
  }

  /* Start timer */
  h2timerStart(timer, timerPeriod, 1);
#else
  printf ("solaris: no timer\n");
#endif

#endif

/*     ((double) 200 / (double) NTICKS_PER_SEC); */
  printf ("   <<<   TIMER of %d tics :  %g SECOND  >>> \n"
	  "Can be set with : \n"
	  "       %s [period in tics]\n",
	  timerPeriod,
	  (double)timerPeriod/(double)NTICKS_PER_SEC,
/* 	  sysClkRateGet()); */
	  argv[0]);
  printf ("Right mouse button for help.\n\n");

  printf ("Wait timer ...");
  fflush(stdout);


  /** ------------------------------------------------------------
   ** LES OPTIONS ET LES COULEURS 
   **/

  initColorTab(dpy, screen, colorTab, colorGridTab);

  affSeglocDrawOptionsInit(&drawFlags);
  
  /* On trace demo une fois au depart */
  XClearArea(dpy,win,0,0,0,0,1);

  printf ("affSegloc: INFO: to get help, type 'h' in the X window display\n");

  /** ------------------------------------------------------------
   ** INIT COMM
   **/

  /* Init loco and segloc */
  if (!affInitSegloc()) {
    drawFlags.drawSeglocFlag = FALSE;
    drawFlags.drawAcqSegsFlag = FALSE;
    drawFlags.drawAspectSegsFlag = FALSE;
    drawFlags.drawAspectMapFlag = FALSE;
    drawFlags.drawMatchedPosFlag = FALSE;
    drawFlags.drawMatchedAcqSegFlag = FALSE;
    drawFlags.drawMatchedModelSegFlag = FALSE;
    drawFlags.drawUsSeglocFlag = FALSE;
  }
  if (drawFlags.drawOdoFlag) {
    if (!affInitLoco()) {
      drawFlags.drawOdoFlag = FALSE;
    }
  }
#if 0
  if (!affInitComm()) {
    XDestroyWindow(dpy, win);
    XCloseDisplay(dpy);
    exit (-2);
  }
#endif
  fprintf (stderr, "ok\n");

#define STEP (5.)

  /**  ------------------------------------------------------------
   **  MAIN LOOP
   **/

  while(!end) {
    
    /**
     ** TIME
     **/

    /* Timer */
#ifndef SOLARIS
#if 0
    if (h2timerPause(timer) != OK) {
      fprintf (stderr, "Cannot Pause timer!\n");
      end = 1;
      break;
    }
#else
    taskDelay(timerPeriod);
#endif

#else
/*     printf (".\n"); */
/*     sleep(1); */
#endif

    if (firsttime) {
	printf (" OK\n");
	firsttime = 0;
    }
    
    /* Time */
    if (!decompte) {
	fin_period = get_mtime();
	printf ("Temps Calcul / Periode :  %g sec / %g sec\n", 
		(fin_calcul - debut)/1000.0,
		(fin_period - debut)/1000.0);
	debut = fin_period;
/* 	decompte = init_decompte; */
	decompte--;
    }

    
    /**
     ** GESTION DES EVENEMENTS
     **/
    if (XCheckWindowEvent(dpy, win, STD_EVENT_MASK, &event)) {

      /* XNextEvent(dpy, &event); */

      switch(event.type) {
	/* A faire quand la souris bouge */
      case MotionNotify:
	/* Affiche coord position souris en metres */
	/* XXX Pas compatible avec une tache periodique */
	/* Ou du moins il faudrait :
	   1. consommer tous les evts en attente (ou utiliser 'timer X')
	   2. afficher directemtn dans la figure (evite xterm qui defile 
	*/
/* 	if(event.xmotion.state && Button1Mask) */
/* 	  printf ("pos %g %g\n", x_pix_to_metre(event.xbutton.x),  */
/* 		  y_pix_to_metre(event.xbutton.y)); */
	/* Nothing */
	break;

	/* A faire quand on appuie sur une touche du clavier */
      case KeyPress:
	if (!draw_option(&drawFlags, &event.xkey, 
			 &echelle, &offset_x, &offset_y)) {
	  end = 1;
	}
	XClearArea(dpy,win,0,0,0,0,False);
	break;

	/* A faire quand on appuie sur un bouton de la souris */
      case ButtonPress:

	switch(event.xbutton.button) {
	case Button1:
	  /* Affiche la position dans le shell */
	  printf ("pos %g %g\n", x_pix_to_metre(event.xbutton.x), 
		  y_pix_to_metre(event.xbutton.y));
	  break;
	case Button2:
	  printf (help);
	  break;  
	case Button3:
/* 	  end=1; */
	  printf (help);
	  break;
	}	/* switch */

      }	/* switch */
    } /* if */


    /**   ------------------------------------------------------------
     **   DRAW ALL : on redessine tout
     **/
    if (!drawFlags.drawStopFlag) {

      /* Nettoyage eventuel */
      if (drawFlags.drawRedrawFlag)
	XClearArea(dpy,win,0,0,0,0,False);
      
      /* On affiche tout */
      if (draw_all(gc) == ERROR) {
#ifndef SOLARIS
	h2timerPause(timer);
#endif
      }
    }


    fin_calcul = get_mtime();
  } /* while */

  if (timer)
    h2timerFree (timer);
  XDestroyWindow(dpy, win);
  XCloseDisplay(dpy);

  return 1;
}


/**  ----------------------------------------------------------------------
 **  draw_all : fonction de dessin principale + lecture donnees
 **
 **
 **
 **/

static int draw_all(GC gc)
{
  SEGLOC_POS_VAR posLocoSick;  /* Pos.  robot associée aux données du sick */
#ifdef SICK2
  SEGLOC_POS_VAR posLocoSick2;  /* Pos.  robot associée aux données du sick */
#endif
  SEGLOC_POS_VAR posLocoSegloc;  /* Pos.  robot associée à la segmentation */
  SEGLOC_POS_VAR posLocoSickCorr; /* posLocoSick corrigée 
				 (calculée avec la transformée) */
  SEGLOC_POS_VAR posPomGlobalSick;
#ifdef SICK2
  SEGLOC_POS_VAR posLocoSickCorr2; /* posLocoSick corrigée 
				      (calculée avec la transformée) */
#endif
  SEGLOC_POS_VAR posLocoSeglocCorr; /* posLocoSegloc corrigée 
				 (calculée avec la transformée) */
  SEGLOC_POS_VAR posMatchedAcqSeg;  /* Position calculée par seloc 
				   (retournée avec les segments matchés ->
				   pas de risque de dé-synchro des posters) */
  SEGLOC_POS_VAR tr;

  SEGLOC_POS_VAR seglocPos;
#ifdef POM
  SEGLOC_POS_VAR pomPos;
  SEGLOC_POS_VAR odoPomPos;
#endif

  SEGLOC_POS_VAR posRecent;
  SEGLOC_POS_VAR posRecentCorr;

#ifdef US
  SEGLOC_POS_VAR posUs;
  SEGLOC_POS_VAR posUsCorr;
#endif

  SEGLOC_POS posOrigine={0.0, 0.0, 0.0};

  SEGLOC_ACQ_SEGS acqSegsSick;
  SEGLOC_ACQ_SEGS acqSegsSegloc;
  SEGLOC_ASPECT_SEGS aspectSegs;

#ifdef M2D
  SEGLOC_ACQ_SEGS m2dObstSegs;
#endif

#ifdef ZONE
  ZONES_MAP zonesSegsObst;
  ZONES_MAP zonesSegsTarget;
  ZONES_MAP zonesSegsSpecial;
  ZONES_DESC zonesDesc;
  ZONES_DESC zonesTargetDesc;
#endif

#ifdef GRID
  static GRID_GRID_POSTER_STR grid;
#endif

#ifdef US2MAT
  static US2MAT_GRID_POSTER_STR gridTernar;
  static US2MAT_PROBAGRID_POSTER_STR gridProba;
#endif

  int nbSegs;
  SICK_SCANPOLAR_POSTER_STR sickScanPolar;
#ifdef SICK2
  SICK_SCANPOLAR_POSTER_STR sickScanPolar2;
#endif

#ifdef ASPECT
  ASPECT_CARTESIANASPECT_POSTER_STR aspectAspectMap;
  SEGLOC_POS_VAR aspectMapSegsRobotPos;
#endif

  static SEGLOC_MAP mapSegs;
  SEGLOC_LOC_REPORT locReport; 
  SEGLOC_POS pos0 = {0.0,0.0,0.0};

  XAFF_POS posl;
#ifdef ZONE
  XAFF_POS posZ;
#endif
/*   XAFF_POS posUnc = {6.0, -10.0, 0.0}; */
  SEGLOC_VARIANCES poslV;

#ifdef US
  XR4000_MEASURES_IN_GLOBAL_CART_TAB usGlobal;
  XR4000_MEASURES_IN_ROBOT_CART_TAB usRobot;
#endif
#ifdef USRFLEX
  RFLEX_SONARSCART_POSTER_STR usRflex;
#endif

  int drawMapFrequency = 3;
  int drawMapDecount = 0;
 
  int drawTrajFrequency = 5;
  int drawTrajDecount = 0;

  int coulSick;
  int coulSegs;

  /**
   ** LECTURE DES DONNEES
   **/

#ifdef M2D
  /* On lit les obstacle m2d */
  if (drawFlags.drawM2dObstFlag)
    if (affGetM2dObst(&m2dObstSegs, drawFlags.drawM2dObstFlag)==ERROR)
      drawFlags.drawM2dObstFlag = FALSE;
#endif

  /* Les points laser avec position loco lors de l'acquisition */
  if (drawFlags.drawLaserPointsFlag) {
    
    /* Lecture par fichier */
    if (laserBaseFileName[0] != '\0') {
      if (!readLaserPoints(laserBaseFileName, laserFileNum,
			   &sickScanPolar.Header, sickScanPolar.Rho, &posLocoSick)) {
	drawFlags.drawLaserPointsFlag = FALSE;
      }
      else {
	/* Fichier suivant */
	laserFileNum++;
	printf ("laserFileNum = %d\n", laserFileNum);
      }
    }

    /* Lecture par poster */
    else {

      /* aspect points */
      if (drawFlags.drawAspectSegsFlag) {
#ifdef ASPECT
	if (affGetLaserPointsAspect(&sickScanPolar, &posLocoSick) ==ERROR) {
	  printf ("affGetLaserPointsAspect failed\n");
	  drawFlags.drawLaserPointsFlag = FALSE;
	}
#endif /* ASPECT */
      }

      /* sick points */
      else {
	if (drawFlags.drawLaserPointsFlag & LASER_1_ON) {
	  if (affGetLaserPoints(&sickScanPolar.Header, sickScanPolar.Rho,
				&posLocoSick) ==ERROR) {
	    drawFlags.drawLaserPointsFlag &= ~LASER_1_ON;
	    printf ("drawLaserPointsFlag 1 = %d\n",
		    drawFlags.drawLaserPointsFlag);
	  }
	}
#ifdef SICK2  
	if (drawFlags.drawLaserPointsFlag & LASER_2_ON) {
	  if (affGetLaser2Points(&sickScanPolar2.Header, sickScanPolar2.Rho
				 , &posLocoSick2) ==ERROR){
	    drawFlags.drawLaserPointsFlag &= ~LASER_2_ON;
	    printf ("drawLaserPointsFlag 2 = %d\n",
		    drawFlags.drawLaserPointsFlag);
	  }
	}
#endif /* SICK2 */
      }
    }
    /* Position courante du robot */
    posRecent = posLocoSick;
  }   /* draw laser points */
  
#ifdef US
  /* Les données US dans le repere odo */
  if (drawFlags.drawUsOdoFlag) 
    if (affGetLocalUS(&usGlobal)==ERROR)
      drawFlags.drawUsOdoFlag = FALSE;
  
  /* Les données US dans le repere robot (pour transfo dans segloc) */
  if (drawFlags.drawUsSeglocFlag) {
    if (affGetLastPos(&posUs, pomMEposPosterName) == ERROR) {
      drawFlags.drawUsSeglocFlag = FALSE;
    }
    else {
#ifdef USRFLEX
      if (affGetUSRflex (&usRflex) ==ERROR) 
	drawFlags.drawUsSeglocFlag = FALSE;
#else
      if (affGetLocalUS(&usRobot)==ERROR) 
	drawFlags.drawUsSeglocFlag = FALSE;
#endif
    }
  }
#endif
  
  /* La transfo de correction et sa variance */
  if (drawFlags.drawSeglocFlag) {
    affGetLoc(&tr, &locReport);
#ifdef POM
    affGetPomPos(&odoPomPos, &pomPos);
#endif
  }
  
  /* Les segments acquis et la position associée */
  if (drawFlags.drawSeglocFlag)
    affGetSeglocAcqSegs(&acqSegsSegloc, 
			drawFlags.drawAcqSegsFlag, 
			&posLocoSegloc);
  
  if (drawFlags.drawAcqSegsFlag)
    affGetSickAcqSegs(&acqSegsSick.header.nbSegs, 
		      (SICK_SEG *)&acqSegsSick.segs, 
		      &posLocoSick,
		      &posPomGlobalSick);

#ifdef ASPECT
  /* Les segments de la carte d'aspect du module aspect */
  if (drawFlags.drawAspectMapFlag) {
    affGetAspectMap(&aspectAspectMap, &aspectMapSegsRobotPos);
  }
#else

  /* Les segments de la carte d'aspect de SEGLOC, avec la position corrigée par segloc */
  if (drawFlags.drawAspectSegsFlag || drawFlags.drawMatchedAcqSegFlag || 
      drawFlags.drawMatchedModelSegFlag || drawFlags.drawMatchedPosFlag) {
    affGetAspectSegs(&aspectSegs, 
		     drawFlags.drawMatchedModelSegFlag, 
		     drawFlags.drawMatchedAcqSegFlag,
		     drawFlags.drawAspectSegsFlag);
    posMatchedAcqSeg.pos = aspectSegs.header.locoPos;
    posMatchedAcqSeg.var = aspectSegs.header.variances;
  }
#endif

  /**
   ** CONVERSIONS DES DONNEES
   **/


  /* Calcul de la position corrigée avec variance */

  if (drawFlags.drawSeglocFlag) {
    CalcAbsPos(&posLocoSegloc, &tr, &posLocoSeglocCorr);
  }

  /* Calcul position corrigée par segloc */
  if (drawFlags.drawLaserPointsFlag & LASER_1_ON)
    CalcAbsPos(&posLocoSick, &tr, &posLocoSickCorr);
#ifdef SICK2
  if (drawFlags.drawLaserPointsFlag & LASER_2_ON)
    CalcAbsPos(&posLocoSick2, &tr, &posLocoSickCorr2);
#endif
  

#ifdef US2MAT
  /**
   **  LA GRILLE
   **/
  if (drawFlags.drawGridFlag) {

    /* Grille ternarisee : lecture */
    if (drawFlags.drawGridType == GRID_TERNAR) {
      printf ("get ternar grid ...");
      fflush(stdout);
      if (!affGetUs2matTernarGrid(&gridTernar)) {
	printf ("Cannot read ternar grid\n");
	drawFlags.drawGridFlag = 0;
      }
      /* Grille ternarisee : affichage */
      if (drawFlags.drawGridFlag) {
	printf ("draw grid\n");
	affUs2MatTernarGrid(&gridTernar, 
			    drawFlags.drawGridFollowFlag, 
			    colorTab[COUL_GRIDOCCUPE].pixel, 
			    colorTab[COUL_GRIDLIBRE].pixel, 
			    colorTab[COUL_GRIDINCONNU].pixel, gc);
      }
    }
    /* Grille probabiliste */
    else {
      /* Grille instantannee : lecture */
      if (drawFlags.drawGridType == GRID_INSTANT) {
	printf ("get instant grid...");
	fflush(stdout);
	if (!affGetUs2matInstantGrid(&gridProba)) {
	  printf ("Cannot read instant grid\n");
	  drawFlags.drawGridFlag = 0;
	}
      }
      /* Grille cumulee : lecture */
      else {
	printf ("get proba grid...");
	fflush(stdout);
	if (!affGetUs2matProbaGrid(&gridProba)) {
	  printf ("Cannot read proba grid\n");
	  drawFlags.drawGridFlag = 0;
	}
      }

      /* AFFICHAGE */
      if (drawFlags.drawGridFlag) {
	printf ("draw proba grid ....");
	fflush(stdout);
	affUs2MatProbaGrid(&gridProba, drawFlags.drawGridFollowFlag, 
			   colorGridTab, gc);
	printf ("OK grid\n");
      }
    }

#if 0 
    printf ("get grid\n");
    if (!affGetGrid(&grid)) {
      printf ("Cannot read grid\n");
      drawFlags.drawGridFlag = 0;
    }
    if (drawFlags.drawGridFlag) {
      printf ("draw grid\n");
      affiche_matrice_resultat(&grid, 
			       colorTab[COUL_GRIDOCCUPE].pixel, 
			       colorTab[COUL_GRIDLIBRE].pixel, 
			       colorTab[COUL_GRIDLIBRE].pixel, 
			       colorTab[COUL_GRIDINCONNU].pixel, gc);
      printf ("OK grid\n");
    }
#endif
  }
#endif  /* US2MAT */

  /**
   **   DESSINS SEGMENTS
   **/
  
  /* La carte */
  if (!drawFlags.drawClearMapFlag) {

    if (!readMap || drawFlags.drawUpdateMapFlag) {
      /* Lecture par fichier */
      if (mapFileName[0] != '\0') {
	if (!(readMap = seglocReadSegkit2SeglocMap(mapFileName, &mapSegs))) {
	  drawFlags.drawUpdateMapFlag = FALSE;
	  drawFlags.drawClearMapFlag = TRUE;
	}
	else {
	  readMap = enlargeSegs(mapSegs.header.nbSegs, mapSegs.segs, xMapObst, yMapObst);
	}
      }

      /* Lecture par poster */
      else {
	if (drawFlags.drawZonesFlag) {
	  readMap = readMapZoneSegs(&mapSegs, xMapObst, yMapObst);
	}
	else {
	  readMap = readMapSegs(&mapSegs, xMapObst, yMapObst);
	}
      }
    }

    if (readMap && drawMapDecount == 0) {
      drawMapDecount = drawMapFrequency;
      
      /* Dessine les segments de la carte */
      if (!drawFlags.drawRectFlag) {
	XSetForeground(dpy, gc, colorTab[COUL_MAP].pixel);
	dessin_segloc_segments(&pos0, mapSegs.header.nbSegs, mapSegs.segs, 0, gc);
      }
      /* Idem mais les segemnts sont representes par des rectangles 
	 (grossit cote matiere). Permet de visualiser quel est le cote
	 de la  matiere. Necessite pre-calcul des coord des rectangles 
	 (xMapObst, yMapObst) */
      else {
	XSetForeground(dpy, gc, colorTab[COUL_MAP].pixel);
	XSetForeground(dpy, gcgris, colorTab[COLOR_GREY80].pixel);
	
	if (readMap)
	  dessin_rects(xMapObst, yMapObst, readMap, gc); 
	
	/* Idem que dessin_rects mais moins efficace
	   (mais pas besoin de xMapObst, yMapObst) */
	/*     dessin_rectangles(&mapSegs.segs, gc); */
      }
      
      /* Zones */
      if (drawFlags.drawZonesFlag) {
#ifdef ZONE
	/* Lecture */
	if (!readZones) {
	  /* Lecture par fichier */
	  if (zonesFileName[0] != '\0') {
	    readZones = zoneReadFile(zonesFileName, &zonesTargetDesc, &zonesSegsObst);
	  }
	  /* Lecture par poster */
	  else {
	    readZones = affGetZones(&zonesSegsObst, &zonesSegsTarget, 
				    &zonesSegsSpecial, 
				    &zonesTargetDesc);
	  }
	}
	/* Dessin */
	if (readZones) {
	  int i;
	  XSetForeground(dpy, gc, colorTab[COUL_ZONES_TOPO].pixel);
	  dessin_zones(&zonesSegsTarget, gc);
	  for (i=0; i<zonesTargetDesc.nbZones; i++) {
	    posZ.x = zonesTargetDesc.zones[i].finalPos.x;
	    posZ.y = zonesTargetDesc.zones[i].finalPos.y;
	    posZ.theta = zonesTargetDesc.zones[i].finalPos.theta;
	    dessin_robot_rond(&posZ, 0.4, gc);
	    posZ.x = zonesTargetDesc.zones[i].target.x;
	    posZ.y = zonesTargetDesc.zones[i].target.y;
	    posZ.theta = zonesTargetDesc.zones[i].target.theta;
	    dessin_robot_rond(&posZ, 0.1, gc);
	  }
	  dessin_zones(&zonesSegsSpecial, gc);
	  XSetForeground(dpy, gc, colorTab[COUL_ZONES_STRESS].pixel);
	  dessin_zones(&zonesSegsObst, gc);
	}
#endif  /* ZONE */
      }
      else 
	readZones = FALSE;
    }
    else
      drawMapDecount--;
  }
  else 
    drawMapDecount = 0;

  /* couleur */
  if (locReport.nbSuccesses > 0) {
    coulSick = COUL_SICKOK;
    coulSegs = COUL_SEGSOK;
  }
  else if (locReport.nbSuccesses == 0 && locReport.nbPossibilities > 0) {
    coulSick = COUL_SICKBOF;
    coulSegs = COUL_SEGSBOF;
  }
  else {
    coulSick = COUL_SICKPASOK;
    coulSegs = COUL_SEGSPASOK;
  }
  
  /* Les segments acquis avec sick + position pom */
  if (drawFlags.drawAcqSegsFlag) {
    XSetForeground(dpy, gc, colorTab[COLOR_PURPLE].pixel);
    dessin_segloc_segments(&posPomGlobalSick.pos, 
			   acqSegsSick.header.nbSegs, 
			   acqSegsSick.segs, 0, gc);
  }
  /* Les segments acquis avec segloc */
  if (drawFlags.drawSeglocFlag) {
    XSetForeground(dpy, gc, colorTab[coulSegs].pixel);
    dessin_segloc_segments(&posLocoSeglocCorr.pos, 
			   acqSegsSegloc.header.nbSegs, 
			   acqSegsSegloc.segs, 0, gc);
  }
  
#ifdef ASPECT
  /* Les segments de la carte d'aspect du module ASPECT */
  if (drawFlags.drawAspectMapFlag && !drawFlags.drawLaserPointsFlag) {
    XSetForeground(dpy, gc, colorTab[COUL_ASPECT].pixel);
    dessin_segloc_segments(&aspectMapSegsRobotPos.pos,
			   aspectAspectMap.Sick_seg_header.nbSegs, 
			   (SEGLOC_SEG *)aspectAspectMap.Sick_segs.segs, 1, gc);
  }
#endif

  /* Les segments de la carte d'aspect de SEGLOC */
  if (drawFlags.drawAspectSegsFlag) {
    XSetForeground(dpy, gc, colorTab[COUL_ASPECT].pixel);
    nbSegs = min(aspectSegs.header.nbSegs,
		 SEGLOC_MAX_ASPECT_SEGS - 2*aspectSegs.header.nbMatchSegs);
    dessin_segloc_segments(&posOrigine, nbSegs, 
			   &aspectSegs.segs[2*aspectSegs.header.nbMatchSegs],
			   1, gc);
  }
  
  /* Les segments appariés */
  if (1 || locReport.nbSuccesses > 0) {

    /* Du model */
    if (drawFlags.drawMatchedModelSegFlag) {
      XSetForeground(dpy, gc, colorTab[COUL_MATCHEDMODEL].pixel);
      dessin_segloc_segments(&posOrigine, aspectSegs.header.nbMatchSegs,
			     &aspectSegs.segs[aspectSegs.header.nbMatchSegs], 
			     1, gc);
    }

    /* Percus */
    if (drawFlags.drawMatchedAcqSegFlag) {
      XSetForeground(dpy, gc, colorTab[COUL_MATCHEDACQ].pixel);
      dessin_segloc_segments(&posMatchedAcqSeg.pos, 
			     aspectSegs.header.nbMatchSegs, 
			     aspectSegs.segs, 1, gc);
    }
  }

  /**
   **   DESSINS POINTS LASER
   **/
  if (drawFlags.drawLaserPointsFlag & LASER_1_ON) {
    XSetForeground(dpy, gc, colorTab[coulSick].pixel);
    if (drawFlags.drawSeglocFlag)
      dessin_point_laser(&posLocoSickCorr.pos, &sickScanPolar.Header, sickScanPolar.Rho, gc);
    else
      dessin_point_laser(&posLocoSick.pos, &sickScanPolar.Header, sickScanPolar.Rho, gc);
  }
#ifdef SICK2
  if (drawFlags.drawLaserPointsFlag & LASER_2_ON) {
    XSetForeground(dpy, gc, colorTab[coulSick].pixel);
    dessin_point_laser(&posLocoSickCorr2.pos, &sickScanPolar2.Header, sickScanPolar2.Rho, gc);
  }
#endif

  /**
   **   DESSINS POINTS US
   **/
#ifdef USRFLEX
  if (drawFlags.drawUsSeglocFlag) {
    dessin_points_us_rflex(&pomPos.pos, &usRflex, 
			    drawFlags.drawUsTopFlag ?
			    colorTab[COUL_US_TOP].pixel : 0, 
			    drawFlags.drawUsDownFlag ?
			    colorTab[COUL_US_DOWN].pixel : 0,
			    gc);
  }
#endif
#ifdef US
  if (drawFlags.drawUsOdoFlag) {
    dessin_points_us_odo(&usGlobal, 
			 drawFlags.drawUsTopFlag ?
			 colorTab[COUL_US_TOP].pixel : 0, 
			 drawFlags.drawUsDownFlag ?
			 colorTab[COUL_US_DOWN].pixel : 0, 
			 gc);
  }
  if (drawFlags.drawUsSeglocFlag && drawFlags.drawSeglocFlag) {
    CalcAbsPos(&posUs, &tr, &posUsCorr);
    dessin_points_us_segloc(&posUsCorr.pos, &usRobot, 
			    drawFlags.drawUsTopFlag ?
			    colorTab[COUL_US_TOP].pixel : 0, 
			    drawFlags.drawUsDownFlag ?
			    colorTab[COUL_US_DOWN].pixel : 0,
			    gc);
  }
#endif 

  /**
   **   DESSINS ROBOTS
   **/


  /* La position qui a été calculée par la loc */
  if (drawFlags.drawMatchedPosFlag) {
    XSetForeground(dpy, gc, colorTab[coulSick].pixel);

    posl.x = posMatchedAcqSeg.pos.x;
    posl.y = posMatchedAcqSeg.pos.y;
    posl.theta = posMatchedAcqSeg.pos.theta;
    poslV = posMatchedAcqSeg.var;
    
    if (drawFlags.drawRobotRondFlag)
      dessin_robot_rond(&posl, ROBOT_RADIUS, gc);
    else
      dessin_robot_h2(&posl, 1.0, gc);
    
    if (drawFlags.drawUncFlag) {
      draw_uncertainties(&posl, &poslV, gc);
      /*       draw_uncertainties(&posUnc, &poslV, gc); */
    }
  }


  /* Lecture de la donnée la plus fraiche possible  */
  if (drawFlags.drawOdoFlag || drawFlags.drawSeglocFlag) {
    
    /* odo pas dispo */
    if (affGetLastPos(&posRecent, pomMEposPosterName) == ERROR) {
      
      /* via les segments de segloc ? */
      if (drawFlags.drawSeglocFlag) {
	posRecent = posLocoSegloc;
      }
      
      /* via segloc ? */
      else if (affGetPosSegs(&posRecent) == ERROR) {
	drawFlags.drawSeglocFlag = FALSE;
	drawFlags.drawOdoFlag = FALSE;
      }
    }
  }
  
  /* La position courante recalée par la dernière transfo trouvée */
  if (drawFlags.drawSeglocFlag) {
    
    XSetForeground(dpy, gc, colorTab[coulSegs].pixel);
    
    affGetSeglocPos(&seglocPos);
    posRecentCorr = seglocPos;
  
#if 0
#ifdef ODO
    /* Odo corrige */
    CalcAbsPos(&posRecent, &tr, &posRecentCorr);
#else
    /* Sick estime ou corrige (par segloc) */
    affGetSickSeglocPos(&posRecentCorr);
#endif
#endif

    posl.x = posLocoSeglocCorr.pos.x;
    posl.y = posLocoSeglocCorr.pos.y;
    posl.theta = posLocoSeglocCorr.pos.theta;
    poslV = posLocoSeglocCorr.var;
  
    if (drawFlags.drawRobotRondFlag) {
      /* pos segloc */
      dessin_robot_rond(&posl, ROBOT_RADIUS-0.04, gc);
#ifdef POM
      /* pos pom */
      XSetForeground(dpy, gc, colorTab[COLOR_PURPLE].pixel);
      dessin_robot_rond_segloc(&pomPos, ROBOT_RADIUS, FALSE, gc);
#endif
    }
    else
      dessin_robot_h2(&posl, 1.0, gc);

    if (drawFlags.drawUncFlag) {
      draw_uncertainties(&posl, &poslV, gc);
      /*       draw_uncertainties(&posUnc, &poslV, gc); */
    }
  }

  /* Le robot selon loco */
  if (drawFlags.drawOdoFlag || drawFlags.drawLaserPointsFlag) {
    XSetForeground(dpy, gc, colorTab[COUL_LOCO].pixel);
#if 0
    posl.x = posLocoSegloc.pos.x;
    posl.y = posLocoSegloc.pos.y;
    posl.theta = posLocoSegloc.pos.theta;
    poslV = posLocoSegloc.var;
#endif
    posl.x = posRecent.pos.x;
    posl.y = posRecent.pos.y;
    posl.theta = posRecent.pos.theta;
    poslV = posRecent.var;

    if (drawFlags.drawRobotRondFlag) {
#ifdef POM
      /* odo */
      dessin_robot_rond_segloc(&odoPomPos, ROBOT_RADIUS+0.06, FALSE, gc);
#else
      dessin_robot_rond(&posl, ROBOT_RADIUS, gc);
#endif
    }
    else
      dessin_robot_h2(&posl, 1.0, gc);
    if (drawFlags.drawUncFlag) {
      draw_uncertainties(&posl, &poslV, gc);
    }
  }

  /**
   **  DESSIN TRAJECTOIRE
   **/
  if (drawFlags.drawTrajFlag && drawTrajDecount==0) {
    drawTrajDecount = drawTrajFrequency;
    XSetForeground(dpy, gc, colorTab[COUL_TRAJ].pixel);
#if defined M2D || defined VSTP
    if (!draw_m2d_trajectory(gc)) {
      printf ("aff : draw traj desactivated (t to switch on again)\n");
      drawFlags.drawTrajFlag = 0;
    }
#else
    printf ("aff: function \"draw_m2d_trajectory\" desactivated\n");
#endif
  }
  else
    drawTrajDecount--;
  
  return OK;
}


/**
 **  lecture points laser
 **
 **
 **
 **/
static BOOL readLaserPoints(char *baseFileName, int fileNum,
			    SICK_MEASURES_HEADER_STR *header, 
			    double rho[SICK_N_PT_MAX], 
			    SEGLOC_POS_VAR *pos)
{
  char fileName[256];
  double angleRange = 3.14159;

  /* Construit le nom du fichier */
  sprintf(fileName, "%s.%03d", baseFileName, fileNum); 
  
  /* Lecture fichier */
  if (!sickReadFile(fileName, NULL, header, rho)) {
    printf ("Cannot read %s. Turn off drawLaserPointsFlag\n",
	    fileName);
    return FALSE;
  }
  
  /* Enregistrement position robot (odometrie) */
  pos->pos.x = header->sickPomPos.mainToBase.euler.x;
  pos->pos.y = header->sickPomPos.mainToBase.euler.y;
  pos->pos.theta = header->sickPomPos.mainToBase.euler.yaw;
    
  /* Attention : Angle de départ de tir et non pas angle moyen !!! */
  /* XXXX il y avait  "+=" bizarre !? */
  header->sickPomPos.sensorToMain.euler.yaw = header->sickPomPos.sensorToMain.euler.yaw  - angleRange/2.;
  
  pos->var.vxx = header->sickPomPos.mainToBase.var.vxx;
  pos->var.vxy = header->sickPomPos.mainToBase.var.vxy;
  pos->var.vyy = header->sickPomPos.mainToBase.var.vyy;
  pos->var.vxt = header->sickPomPos.mainToBase.var.vox;
  pos->var.vyt = header->sickPomPos.mainToBase.var.voy;
  pos->var.vtt = header->sickPomPos.mainToBase.var.voo;
    
  return TRUE;
}

/**
 **  dessin_segloc_segments
 **
 **
 **
 **/

static void dessin_segloc_segments(SEGLOC_POS *segpos, int nbsegs, 
				   SEGLOC_SEG *segs, BOOL extremity, GC gc)
{
  int i;
  XAFF_POS pos;

  pos.x = segpos->x;
  pos.y = segpos->y;
  pos.theta = segpos->theta;

  if (extremity)
    for (i=0; i<nbsegs; i++) {
      dessin_seg_extremities(&pos, segs->x1, segs->y1, segs->x2, segs->y2, gc);
      segs++;
    }
  else
    for (i=0; i<nbsegs; i++) {
      dessin_seg(&pos, segs->x1, segs->y1, segs->x2, segs->y2, gc);
      segs++;
    }

}

/**
 **  dessin_zones
 **
 **
 **
 **/

#ifdef ZONE

void 
dessin_zones(ZONES_MAP *zones, GC gc)
{
  int i,j;
  int firstPt=0;
  XAFF_POS pos={0};

  for (i=0; i<zones->header.nbZones; i++) {
    for (j=firstPt; j<firstPt + zones->header.nbPtsPerZone[i]-1; j++) {
      if (!drawFlags.drawRectFlag) 
	dessin_seg(&pos, zones->points[j].x, zones->points[j].y,
		   zones->points[j+1].x, zones->points[j+1].y, gc);
      else
	dessin_seg_extremities(&pos, zones->points[j].x, zones->points[j].y,
			       zones->points[j+1].x, zones->points[j+1].y, gc);
    }
    if (!drawFlags.drawRectFlag) 
      dessin_seg(&pos, zones->points[j].x, zones->points[j].y,
		 zones->points[firstPt].x, zones->points[firstPt].y, gc);
    else
      dessin_seg_extremities(&pos, zones->points[j].x, zones->points[j].y,
			     zones->points[firstPt].x, zones->points[firstPt].y, gc);
    firstPt += zones->header.nbPtsPerZone[i];
  }
}

#endif /* ZONE */

#if 0
/**
 ** dessin_rectangles - dessine des rectangles par grossissement 
 **                     des segments de SEGLOC_MAP
 **
 ** Meme fonctionnalite que dessin_rects mais moins
 ** efficace car a chaque re-affichage le rectangle est re-calcule
 ** (par contre ne necessite pas des structures intermediaires
 ** xMapObst et yMapObst).
 **/ 
static void dessin_rectangles(SEGLOC_MAP *map, GC gc)
{
  int i;
  for(i=0; i<map->header.nbSegs; i++) {
    dessin_rect_from_seg(map->segs[i].x1, map->segs[i].y1, 
			  map->segs[i].x2, map->segs[i].y2, gc);
  }
}
#endif

/**
 **   readMapSegs  - lit la carte (dans mapSegs.segs) 
 **                  et si xMapObst != NULL, calcule les coord
 **                  des rectangles correspondants aux segments
 **                  grossis coté matière.
 **
 **    retourne le nombre de segments lu ou 0 si pb.
 **/

#define SEG_WIDTH 0.2

static int readMapSegs(SEGLOC_MAP *mapSegs,  
		       double *xMapObst, double *yMapObst)
{
  /**
   ** LECTURE DE LA CARTE
   **/
  if (affGetMap(mapSegs) == ERROR) {
    printf ("Cannot read map\n");
    return 0;
  }

  if (mapSegs->header.nbSegs == 0) {
    printf ("No segs in map\n");
    return 0;
  }

  enlargeSegs(mapSegs->header.nbSegs, mapSegs->segs, xMapObst, yMapObst);

  return mapSegs->header.nbSegs;
}

static int readMapZoneSegs(SEGLOC_MAP *mapSegs,  
			   double *xMapObst, double *yMapObst)
{
  /**
   ** LECTURE DE LA CARTE
   **/
#ifdef ZONE
  if (!affGetMapZone(mapSegs)) {
    printf ("Cannot read mapZone\n");
    return 0;
  }
#endif

  if (mapSegs->header.nbSegs == 0) {
    printf ("No segs in mapZone\n");
    return 0;
  }

  enlargeSegs(mapSegs->header.nbSegs, mapSegs->segs, xMapObst, yMapObst);

  return mapSegs->header.nbSegs;
}

/**
 ** CALCUL DES RECTANGLES GROSSIS
 **/

static int enlargeSegs(int nbSegs, SEGLOC_SEG *mapSegs,  
		       double *xMapObst, double *yMapObst)
{
  int seg;
  double angle;
  double cs, sn;
  double x1, y1, x2, y2;
  double dx, dy;
  double seglength = SEG_WIDTH;

  /* Pour utiliser dessin_rects il faut calculer  les coordonnees
     des rectangles correspondant aux segements grossis cote matiere 
     (xMapObst et yMapObst) */
  for (seg=0; seg<nbSegs; seg++) {
    
    x1 = mapSegs[seg].x1;
    y1 = mapSegs[seg].y1;
    x2 = mapSegs[seg].x2;
    y2 = mapSegs[seg].y2;

    angle = atan2(y2-y1, x2-x1) - M_PI/2.;
    cs = cos(angle);
    sn = sin(angle);

    dx = seglength * cos(angle);
    dy = seglength * sin(angle);

    xMapObst[seg*4]   = x1 + dx;
    yMapObst[seg*4]   = y1 + dy;
    
    xMapObst[seg*4+1]   = x1;
    yMapObst[seg*4+1]   = y1;
    
    xMapObst[seg*4+2]   = x2;
    yMapObst[seg*4+2]   = y2;
    
    xMapObst[seg*4+3]   = x2+dx;
    yMapObst[seg*4+3]   = y2+dy;
  }
  return nbSegs;
}


/**
 ** draw_m2d_trajectory
 **
 **
 **/

#if defined M2D || defined VSTP
static BOOL draw_m2d_trajectory (GC gc)
{
  PILO_TRAJECTORY traj;
  SEGLOC_POS initPos = {0.0, 0.0, 0.0};
  int nbSegs, i;
  SEGLOC_SEG segs[PILO_MAX_TRAJ_SEGMENTS];
  double px,py, pt;
  
#ifdef M2D
  /* Lecture de la trajectoire */
  if (affGetM2dTraj(&traj) == ERROR) {
    printf ("aff : Cannot read traj\n");
    return FALSE;
  }
#endif

#ifdef VSTP
  /* Lecture de la trajectoire */
  if (affGetVstpTraj(&traj) == ERROR) {
    printf ("aff : Cannot read traj\n");
    return FALSE;
  }
#endif

  /* Position initiale */
  if (traj.header.absOrRel != PILO_ABSOLUTE_TRAJ)
    printf ("Aff considers only absolute traj for now\n");

  nbSegs = traj.header.nbSegs;

  px = traj.header.x0;
  py = traj.header.y0;
  pt = traj.header.theta0;

#ifdef DEBUG
  if (numTraj != traj.header.numTraj) {
    numTraj = traj.header.numTraj;
    printf ("traj %d : x0 y0 th0  (%f %f %f) dth0 dl0  (%f %f)\n",
	    traj.header.numTraj,
	    traj.header.x0,
	    traj.header.y0,
	    traj.header.theta0,
	    traj.seg[0].dTheta,
	    traj.seg[0].dL);
  }
#endif

  /* Les segments */
  for (i=0; i<nbSegs; i++) {
    pt += traj.seg[i].dTheta;
    segs[i].x1 = px;
    segs[i].y1 = py;
    segs[i].x2 = px + traj.seg[i].dL * cos(pt);
    segs[i].y2 = py + traj.seg[i].dL * sin(pt);
    px = segs[i].x2;
    py = segs[i].y2;
  }
  dessin_segloc_segments(&initPos, nbSegs, segs, 0, gc);

  return TRUE;
}
#endif



/**
 ** dessin_point_laser
 **
 **
 **
 **/

static void dessin_point_laser(SEGLOC_POS *posRobot, 
			       SICK_MEASURES_HEADER_STR *header, 
			       double rho[SICK_N_PT_MAX],
			       GC gc)
{
  int i;
  double xl, yl;
  double xg, yg;
  double sa, ca;
  double sr, cr;
  double inc, alpha;

  inc = header->dtheta;

  /* Attention : header->pos.theta modifié dans affGetLaserPoints pour 
     contenir l'angle de départ de tir (dans le repere dur robot) 
     et non pas l'angle moyen !!! */
  alpha =  header->sickPomPos.sensorToMain.euler.yaw;

  cr = cos(posRobot->theta);
  sr = sin(posRobot->theta);

  for (i = 0;  i < header->np; i++){

    /* Angle de tir */
    sa = sin(alpha);
    ca = cos(alpha);

    /* Dans le repere robot */
    xl = rho[i] * ca + header->sickPomPos.sensorToMain.euler.x;
    yl = rho[i] * sa + header->sickPomPos.sensorToMain.euler.y;

    /* Dans le repere global */
    xg = posRobot->x + xl * cr - yl * sr;
    yg = posRobot->y + xl * sr + yl * cr;

    /* Dessin */
    dessin_gros_point(xg, yg, 0.0, gc);

    alpha += inc;
  }

}


#ifdef US
/**
 ** dessin_point_us
 **
 **
 **
 **/

static void dessin_points_us_odo(XR4000_MEASURES_IN_GLOBAL_CART_TAB *us,
				 unsigned long colTop,
				 unsigned long colDown,
				 GC gc)
{
#if defined (H2) || defined(XR4000) 
  int i, j;
  double xl, yl;
#endif

#ifdef H2
  XSetForeground(dpy, gc, colTop);
  for (i=0; i<US_NMAX_US_SENSORS/2; i++) {
    xl = us->tab[i].x;
    yl = us->tab[i].y;
    
    /* Dessin */
    dessin_croix(xl, yl, 1, gc);
  }
#elif defined(ATRV) || defined (B21R) || defined (JIDO)
  printf ("atrv US : to be done\n");
  return;
#elif defined(XR4000) /* Pour XR4000 */
  /*  Les US d'en haut */
  if (colTop != 0) {
    XSetForeground(dpy, gc, colTop);
    for (i=0; i<XR4000_NMAX_SENSORS_SETS; i=i+2) {
      for (j=0; j<XR4000_NMAX_SENSORS; j++) {
	
	/* Dans le repere odo */
	xl = us->tab[i][j].x;
	yl = us->tab[i][j].y;
	
	/* Dessin */
	dessin_gros_point(xl, yl, 1, gc);
      }
    }
  }
  
  /*  Les US d'en bas */
  if (colDown != 0) {
    XSetForeground(dpy, gc, colDown);
    for (i=1; i<XR4000_NMAX_SENSORS_SETS; i=i+2) {
      for (j=0; j<XR4000_NMAX_SENSORS; j++) {
	
	/* Dans le repere odo */
	xl = us->tab[i][j].x;
	yl = us->tab[i][j].y;
	
	/* Dessin */
	dessin_croix(xl, yl, 2, gc);
      }
    }
  }
#endif
}
#endif


#ifdef USRFLEX
static void dessin_points_us_rflex(SEGLOC_POS *posRobot, 
				   RFLEX_SONARSCART_POSTER_STR *us,
				   unsigned long colTop,
				   unsigned long colDown,
				   GC gc)
{
#if defined (H2) || defined(XR4000) || defined(B21R)
  int i;
  double xl, yl;
  double xg, yg;
#endif
  double sr, cr;

  cr = cos(posRobot->theta);
  sr = sin(posRobot->theta);
#if defined(B21R)
  XSetForeground(dpy, gc, colTop);
  for (i=0; i<RFLEX_NMAX_SENSORS; i++) {

    /* Dans le repere robot */
    xl = us->MeasureInRobotCart.tab[i].x;
    yl = us->MeasureInRobotCart.tab[i].y;
    
	/* Dans le repere global */
    xg = posRobot->x + xl * cr - yl * sr;
    yg = posRobot->y + xl * sr + yl * cr;
    
    /* Dessin */
    dessin_croix(xg, yg, 1, gc);
  }
#elif defined (ATRV)
  printf ("aff ATRV US : to be done\n");
#elif defined (XR4000) /* Pour Dili */
  /*  Les US d'en haut */
  if (colTop != 0) {
    XSetForeground(dpy, gc, colTop);
    for (i=0; i<XR4000_NMAX_SENSORS_SETS; i=i+2) {
      int j;
      for (j=0; j<XR4000_NMAX_SENSORS; j++) {
	
	/* Dans le repere robot */
	xl = us->tab[i][j].x;
	yl = us->tab[i][j].y;
	
	/* Dans le repere global */
	xg = posRobot->x + xl * cr - yl * sr;
	yg = posRobot->y + xl * sr + yl * cr;
	
	/* Dessin */
	dessin_gros_point(xg, yg, 1, gc);
      }
    }
  }

  /*  Les US d'en bas */
  if (colDown != 0) {
    XSetForeground(dpy, gc, colDown);
    for (i=1; i<XR4000_NMAX_SENSORS_SETS; i=i+2) {
      for (j=0; j<XR4000_NMAX_SENSORS; j++) {
	
	/* Dans le repere robot */
	xl = us->tab[i][j].x;
	yl = us->tab[i][j].y;
	
	/* Dans le repere global */
	xg = posRobot->x + xl * cr - yl * sr;
	yg = posRobot->y + xl * sr + yl * cr;
	
	/* Dessin */
	dessin_croix(xg, yg, 2, gc);
      }
    }
  }
#endif /* XR4000 */
}
#endif /* USRFLEX */

#ifdef US
static void dessin_points_us_segloc(SEGLOC_POS *posRobot, 
				    XR4000_MEASURES_IN_ROBOT_CART_TAB *us,
				    unsigned long colTop,
				    unsigned long colDown,
				    GC gc)
{
#if defined (H2) || defined(XR4000) 
  int i, j;
  double xl, yl;
  double xg, yg;
#endif
  double sr, cr;

  cr = cos(posRobot->theta);
  sr = sin(posRobot->theta);
#ifdef H2
  XSetForeground(dpy, gc, colTop);
  for (i=0; i<US_NMAX_US_SENSORS; i++) {

    /* Dans le repere robot */
    xl = us->tab[i].x;
    yl = us->tab[i].y;
    
	/* Dans le repere global */
    xg = posRobot->x + xl * cr - yl * sr;
    yg = posRobot->y + xl * sr + yl * cr;
    
    /* Dessin */
    dessin_croix(xg, yg, 1, gc);
  }
#elif defined (ATRV) || defined(B21R)
  printf ("aff ATRV US : to be done\n");
#elif defined (XR4000) /* Pour Dili */
  /*  Les US d'en haut */
  if (colTop != 0) {
    XSetForeground(dpy, gc, colTop);
    for (i=0; i<XR4000_NMAX_SENSORS_SETS; i=i+2) {
      for (j=0; j<XR4000_NMAX_SENSORS; j++) {
	
	/* Dans le repere robot */
	xl = us->tab[i][j].x;
	yl = us->tab[i][j].y;
	
	/* Dans le repere global */
	xg = posRobot->x + xl * cr - yl * sr;
	yg = posRobot->y + xl * sr + yl * cr;
	
	/* Dessin */
	dessin_gros_point(xg, yg, 1, gc);
      }
    }
  }

  /*  Les US d'en bas */
  if (colDown != 0) {
    XSetForeground(dpy, gc, colDown);
    for (i=1; i<XR4000_NMAX_SENSORS_SETS; i=i+2) {
      for (j=0; j<XR4000_NMAX_SENSORS; j++) {
	
	/* Dans le repere robot */
	xl = us->tab[i][j].x;
	yl = us->tab[i][j].y;
	
	/* Dans le repere global */
	xg = posRobot->x + xl * cr - yl * sr;
	yg = posRobot->y + xl * sr + yl * cr;
	
	/* Dessin */
	dessin_croix(xg, yg, 2, gc);
      }
    }
  }
#endif /* XR4000 */
}
#endif /* US */

#ifdef UNIX
static unsigned long get_mtime (void)
{
     unsigned long msec;
     struct timeval real_time;
     struct timezone tzp;

     if (gettimeofday(&real_time,&tzp) != 0)
         perror("gettimeofday");

     msec = (real_time.tv_usec / 1000) + (real_time.tv_sec * 1000);
     
     return(msec);
   }
#else
unsigned long get_mtime (void)
{
     unsigned long msec;
     H2TIME time_str;

     if (h2timeGet(&time_str) != OK)
          perror("h2timeget");

     msec = time_str.msec + (60 * ((60 * (long)time_str.hour) + (long)time_str.m
inute) + time_str.sec) * 1000;
     return(msec);
}
#endif



void calc_abs_pos2(double xrob[3], double vrob[6],
		   double tr[3], double vtr[6],
		   double abs[3], double vabs[6]);

#if 0
static void CalcAbsPos2(SEGLOC_POS_VAR *odo, 
		       SEGLOC_POS_VAR *transf, 
		       SEGLOC_POS_VAR *pos)
{
  double rob_enc[3];
  double rob_encV[6];
  double tr[3];
  double trV[6];
  double posCorr[3];
  double posCorrV[6];

  rob_enc[0] = odo->pos.x;
  rob_enc[1] = odo->pos.y;
  rob_enc[2] = odo->pos.theta;

  rob_encV[0] = odo->var.vxx;
  rob_encV[1] = odo->var.vxy;
  rob_encV[2] = odo->var.vyy;
  rob_encV[3] = odo->var.vxt;
  rob_encV[4] = odo->var.vyt;
  rob_encV[5] = odo->var.vtt;

  tr[0] = transf->pos.x;
  tr[1] = transf->pos.y;
  tr[2] = transf->pos.theta;

  trV[0] = transf->var.vxx;
  trV[1] = transf->var.vxy;
  trV[2] = transf->var.vyy;
  trV[3] = transf->var.vxt;
  trV[4] = transf->var.vyt;
  trV[5] = transf->var.vtt;

  calc_abs_pos2(rob_enc, rob_encV, tr, trV, posCorr, posCorrV);

  pos->pos.x = posCorr[0];
  pos->pos.y = posCorr[1];
  pos->pos.theta = posCorr[2];

  pos->var.vxx = posCorrV[0];
  pos->var.vxy = posCorrV[1];
  pos->var.vyy = posCorrV[2];
  pos->var.vxt = posCorrV[3];
  pos->var.vyt = posCorrV[4];
  pos->var.vtt = posCorrV[5];
}
#endif

static double norm_ang(double a,double p)
   
{
  double b;

/*   b = fmod(a,2*M_PI); */
  b = a;
  while(b <= p)
    b += 2*M_PI;

  while(b > p + 2*M_PI)
    b -= 2*M_PI;

  return(b);
}

void calc_abs_pos2(double xrob[3], double vrob[6],
		   double tr[3], double vtr[6],
		   double abs[3], double vabs[6])
{
  double Jo2a[9] = {1.,0.,0.,
		    0.,1.,0.,
		    0.,0.,1.};

  double Jr2a[9] = {1.,0.,0.,
		    0.,1.,0.,
		    0.,0.,1.};

  double aux[9];

  double  cloc2abs,sloc2abs;

  cloc2abs = cos(tr[2]);
  sloc2abs = sin(tr[2]);

  abs[0] =  cloc2abs * xrob[0] - sloc2abs * xrob[1] + tr[0]; 
  abs[1] =  sloc2abs * xrob[0] + cloc2abs * xrob[1] + tr[1];
  abs[2] =  norm_ang(xrob[2] + tr[2],-M_PI);
  if(vtr){
    Jr2a[0] = cloc2abs; Jr2a[1] = -sloc2abs;
    Jr2a[3] = sloc2abs; Jr2a[4] = cloc2abs;
    
    Jo2a[2] = tr[1] - abs[1];
    Jo2a[5] = abs[0] - tr[0];

    mxfs_mmulttr(Jr2a,vrob,aux,3,3,3,3);
    mxfs_mmulttr(Jo2a,vtr,vabs,3,3,3,3);
    
    mxs_plus(aux,vtr,vabs,3,3);
  }
}

void 
dessin_robot_rond_segloc(SEGLOC_POS_VAR *pos, double radius, BOOL drawUnc, GC gc)
{
  XAFF_POS posl;

  posl.x = pos->pos.x;
  posl.y = pos->pos.y;
  posl.theta = pos->pos.theta;
    
  dessin_robot_rond(&posl, radius, gc);

  if (drawUnc)
    draw_uncertainties(&posl, &pos->var, gc);
}

/**
 ** draw_uncertainties
 **
 **
 **
 **/

void draw_uncertainties(XAFF_POS *pos, 
			SEGLOC_VARIANCES *svar,
			GC gc)
{
  double var[6];

  var[0] = svar->vxx;
  var[1] = svar->vxy;
  var[2] = svar->vyy;
  var[3] = svar->vxt;
  var[4] = svar->vyt;
  var[5] = svar->vtt;
  dessin_incertitude(pos->x, pos->y, pos->theta, var, gc);
}


