#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "h2mathLib.h"
#include <portLib.h>
#include <h2timerLib.h>

#ifdef H2
#include "auto/locoPosterLib.h"
#include "auto/usPosterLib.h"
#include "usXR4000Compat.h"
#endif

#ifdef XR4000
#include "xr4000Struct.h"
#include "auto/xr4000PosterLib.h"
#include "locoCompat.h"
#endif

#include "genPosStruct.h"

#if defined(ATRV) || defined(B21R)
#include "rflexStruct.h"
#include "server/rflexPosterLib.h"
#include "locoCompat.h"
#include "usXR4000Compat.h"
#endif

#include "../server/seglocPosterLib.h"
#include "../server/seglocPosterReadLibProto.h"
#include "server/sickPosterLib.h"

#include "piloStruct.h"

#ifdef M2D
#include "server/m2dPosterLib.h"
#endif

#ifdef VSTP
#include "server/vstpPosterLib.h"
#endif

#ifdef SICK2
#include "server/sick2PosterLib.h"
#endif

#ifdef GRID
#include "server/gridPosterLib.h"
#include "gridConst.h"
#endif

#ifdef ZONE
#include "server/zonePosterLib.h"
#endif

#ifdef POM
#include "server/pomPosterLib.h"
#endif

#ifdef ASPECT
#include "server/aspectPosterLib.h"
#endif

#ifdef US2MAT
#include "us2matStruct.h"
#include "server/us2matPosterLib.h"
#endif

#include "libutils/seglocUtils.h"

#include "affSeglocCommProto.h"

static POSTER_ID seglocSegsPosterId=NULL;
static POSTER_ID seglocAspectPosterId=NULL;
static BOOL seglocPosterInitialized=FALSE;

typedef struct LASER_CONFIG {
  double pos [3];        /* Position (x,y,theta) of the laser on the robot */
  double range;          /* Length range of the laser (meters) */
  double accuracy;       /* Accuracy of a mesure (meter) */
  double angle_min;      /* Angle min of the laser (radians) */
  double angle_max;      /* Angle max of the laser (radians) */
  double resolution;     /* Angular resolution (radian) */
} LASER_CONFIG;

static BOOL affInitSick(LASER_CONFIG *sickConfig, int sickNum);

/**
 ** affInitComm
 **
 **/
BOOL affInitComm(void)
{
  if (!affInitSegloc())
    return FALSE;
  if (!affInitLoco())
    return FALSE;
  return TRUE;
}

BOOL affInitSegloc(void)
{

  fprintf (stderr, "init segloc poster...");
  fflush(stderr);
  if (seglocPosterInit () != OK) {
    fprintf (stderr, "Cannot Init segloc poster!");
#ifdef XR4000
    fprintf (stderr, "Is POSTER_HOST definied ?\n");
#endif
    return FALSE;
  }
  
  /**
   ** RECHERCHE DES POSTERS SEGLOC ET XR4000
   **/

  /* Poster adresse */
  if (seglocSegsPosterId==NULL) {
    if (posterFind(SEGLOC_SEGS_POSTER_NAME, &seglocSegsPosterId) == ERROR) {
      seglocSegsPosterId = NULL;
      printf ("poster %s not found\n", SEGLOC_SEGS_POSTER_NAME);
      return FALSE;
    }
  }

  /* Poster adresse */
  if (seglocAspectPosterId==NULL) {
    if (posterFind(SEGLOC_ASPECT_POSTER_NAME, &seglocAspectPosterId) 
	== ERROR) {
      seglocAspectPosterId = NULL;
      printf ("poster %s not found\n", SEGLOC_ASPECT_POSTER_NAME);
      return FALSE;
    }
  }
  seglocPosterInitialized=TRUE;
  return TRUE;
}

BOOL affInitLoco(void)
{
  return TRUE;
}

/**
 ** 
 **
 **/
STATUS affGetLoc(SEGLOC_POS_VAR *tr,
		 SEGLOC_LOC_REPORT *locReport)
{
  SEGLOC_LOC_POSTER_STR loc;
  STATUS status;

  if (!seglocPosterInitialized)
    if (!affInitSegloc())
      return ERROR;

  status = seglocLocPosterRead(&loc);
  if (status == ERROR)
    return status;
  *tr = loc.Transf;
  *locReport = loc.Report;
  return status;
}

/**
 ** 
 **
 **/
STATUS affGetSickSeglocPos(SEGLOC_POS_VAR *loc)
{
  STATUS status;

  if (!seglocPosterInitialized)
    if (!affInitSegloc())
      return ERROR;

  status = seglocMiscSickSeglocPosPosterRead(loc);
  if (status == ERROR)
    return status;
  return status;
}

#ifdef USRFLEX
BOOL affGetUSRflex (RFLEX_SONARSCART_POSTER_STR *sonars)
{
  if (rflexSonarsCartPosterRead(sonars) == ERROR) {
    printf ("affGetUS failed\n");
    return FALSE;
  }
  return TRUE;
}
#endif


#ifdef POM
/**
 ** 
 **
 **/
BOOL affGetPomPos(SEGLOC_POS_VAR *odo, SEGLOC_POS_VAR *pom)
{
  POM_POS pomPos;
  static POSTER_ID pomFrameId=NULL;

  /* Get the pom's current position */
  if (!pomFrameId) {
    if (pomPospomPosPosterRead(&pomPos) != OK) {
      fputs("rflex: could not read frame\n", stderr);
      pomFrameId = NULL;
      return FALSE;
    }
  }

  odo->pos.x = pomPos.mainToBase.euler.x;
  odo->pos.y = pomPos.mainToBase.euler.y;
  odo->pos.theta = pomPos.mainToBase.euler.yaw;

  odo->var.vxx = pomPos.mainToBase.var.vxx;
  odo->var.vxy = pomPos.mainToBase.var.vxy;
  odo->var.vyy = pomPos.mainToBase.var.vyy;
  odo->var.vxt = pomPos.mainToBase.var.vox;
  odo->var.vyt = pomPos.mainToBase.var.voy;
  odo->var.vtt = pomPos.mainToBase.var.voo;

  pom->pos.x = pomPos.mainToOrigin.euler.x;
  pom->pos.y = pomPos.mainToOrigin.euler.y;
  pom->pos.theta = pomPos.mainToOrigin.euler.yaw;

  pom->var.vxx = pomPos.mainToOrigin.var.vxx;
  pom->var.vxy = pomPos.mainToOrigin.var.vxy;
  pom->var.vyy = pomPos.mainToOrigin.var.vyy;
  pom->var.vxt = pomPos.mainToOrigin.var.vox;
  pom->var.vyt = pomPos.mainToOrigin.var.voy;
  pom->var.vtt = pomPos.mainToOrigin.var.voo;

  return TRUE;
}

#endif

STATUS affGetSeglocPos(SEGLOC_POS_VAR *pos)
{
  POM_ME_POS mepos;
  double var[6];

  if (seglocMEPosPosPosterRead(&mepos) == ERROR)
    return ERROR;

  pos->pos.x = mepos.main.euler.x;
  pos->pos.y = mepos.main.euler.y;
  pos->pos.theta = mepos.main.euler.yaw;

  var[0] = mepos.main.var.vxx;
  var[1] = mepos.main.var.vxy;
  var[2] = mepos.main.var.vyy;
  var[3] = mepos.main.var.vox;
  var[4] = mepos.main.var.voy;
  var[5] = mepos.main.var.voo;

  seglocConvertMatrix2Var(&pos->var, var);
  return OK;
}

/**
 ** Version provisoire qui essaye de récupérer la position corrigée 
 ** et l'incertitude corrigée les plus fraiches.
 ** Elles ne sont malheureusement pas en phase....
 **/

STATUS affGetLastPos(SEGLOC_POS_VAR *pos, char *pomMEposPosterName)
{
  static POM_POS	pomPos;		/* current pos */
  static POM_ME_POS	pomMEPos;	/* current pos */
  double var[6];
  static BOOL first = TRUE;      
  static BOOL posterIsPOM_POS = TRUE;
  static POSTER_ID posPosterId;

  if (!first && !posPosterId) 
    return ERROR;

  if (first) {

    first = FALSE;

    /* no poster position : keep pos in sick poster */
    if (pomMEposPosterName[0] == '\0') {
      printf("xaff: use option -p to specify the name of the poster POM_POS or POM_ME_POS where to read pos\n");
      return ERROR;
    }
    
    /* look for a poster of type POM_POS or POM_ME_POS */
    if (seglocPOM_POSPosterFind(pomMEposPosterName,
				&posPosterId) != OK) {
      /* poster found but length not seem correct : try POM_ME_POS */
      if (posPosterId) {
	if (seglocPOM_ME_POSPosterFind(pomMEposPosterName,
				       &posPosterId) != OK) {
	  /* no correct poster */
	  if (posPosterId) 
	    printf ("segloc: poster %s of bad type\n", pomMEposPosterName);
	  else 
	    printf ("segloc: can't find poster %s\n", pomMEposPosterName);
	  printf ("segloc: no robot pos poster. Use sick data.\n");
	  return ERROR;
	}
	/* poster is a POM_ME_POS poster */
	else
	  posterIsPOM_POS = FALSE;
      }
      else {
	printf ("segloc: can't find poster %s\n", pomMEposPosterName);
	printf ("segloc: no robot pos poster. Use sick data.\n");
	return ERROR;
      }
    }

    printf ("segloc: robot pos from %s poster\n", pomMEposPosterName);
  }
  
  /* real pos */
  if (posterIsPOM_POS) {
    if (seglocPOM_POSPosterRead(posPosterId, &pomPos) == OK) {
      pos->pos.x = pomPos.mainToBase.euler.x;
      pos->pos.y = pomPos.mainToBase.euler.y;
      pos->pos.theta = pomPos.mainToBase.euler.yaw;
      
      pos->var.vxx = pomPos.mainToBase.var.vxx;
      pos->var.vyy = pomPos.mainToBase.var.vyy;
      pos->var.vtt = pomPos.mainToBase.var.voo;
      pos->var.vxy = pomPos.mainToBase.var.vxy;
      pos->var.vxt = pomPos.mainToBase.var.vox;
      pos->var.vyt = pomPos.mainToBase.var.voy;
      
      var[0] = pomPos.mainToBase.var.vxx;
      var[1] = pomPos.mainToBase.var.vxy;
      var[2] = pomPos.mainToBase.var.vyy;
      var[3] = pomPos.mainToBase.var.vox;
      var[4] = pomPos.mainToBase.var.voy;
      var[5] = pomPos.mainToBase.var.voo;
    }
  }
  
  else {
    if (seglocPOM_ME_POSPosterRead(posPosterId, &pomMEPos) == OK) {
      pos->pos.x = pomMEPos.main.euler.x;
      pos->pos.y = pomMEPos.main.euler.y;
      pos->pos.theta = pomMEPos.main.euler.yaw;
      
      pos->var.vxx = pomMEPos.main.var.vxx;
      pos->var.vyy = pomMEPos.main.var.vyy;
      pos->var.vtt = pomMEPos.main.var.voo;
      pos->var.vxy = pomMEPos.main.var.vxy;
      pos->var.vxt = pomMEPos.main.var.vox;
      pos->var.vyt = pomMEPos.main.var.voy;
      
      var[0] = pomMEPos.main.var.vxx;
      var[1] = pomMEPos.main.var.vxy;
      var[2] = pomMEPos.main.var.vyy;
      var[3] = pomMEPos.main.var.vox;
      var[4] = pomMEPos.main.var.voy;
      var[5] = pomMEPos.main.var.voo;
    }
  }
  
  seglocConvertMatrix2Var(&pos->var, var);
  return OK;
}


#ifdef ASPECT
/**
 ** LECTURE DES SEGMENTS DE LA CARTE D'ASPECT
 **
 **/
STATUS affGetAspectMap(ASPECT_CARTESIANASPECT_POSTER_STR *aspectMap,
		       SEGLOC_POS_VAR *pos)
{
  int sizeheader = sizeof(SICK_SEG_HEADER);
  int status, nbSegs;
  static POSTER_ID aspectSegsPosterId;
  SICK_SEG_HEADER header;

  if (!aspectSegsPosterId)
    if (posterFind(ASPECT_CARTESIANASPECT_POSTER_NAME, &aspectSegsPosterId))
      return ERROR;

  /* Lecture header */
  status = posterRead(aspectSegsPosterId, 0, &header, sizeheader);
  if (status != sizeheader)
    return ERROR;

  /* Lecture segments */
  nbSegs = header.nbSegs;
  status = posterRead(aspectSegsPosterId, 0, aspectMap, 
		      sizeheader + nbSegs * sizeof(SICK_SEG));
  if (status != sizeheader + nbSegs * sizeof(SICK_SEG)) {
    printf ("affGetAspectSegs : posterRead error\n");
    return ERROR;
  }
  if (aspectMap->Sick_seg_header.nbSegs > nbSegs) {
    printf ("affGetAspectSegs warning : poster change and %d segs are lost\n",
	    aspectMap->Sick_seg_header.nbSegs - nbSegs);
    aspectMap->Sick_seg_header.nbSegs = nbSegs;
  }

  /* Enregistre pos loco utilisée par la segmentation */
  /* XXX FAUDRAIT sensorToOrigin XXXX */
  pos->pos.x = aspectMap->Sick_seg_header.sickPomPos.mainToOrigin.euler.x;
  pos->pos.y = aspectMap->Sick_seg_header.sickPomPos.mainToOrigin.euler.y;
  pos->pos.theta = aspectMap->Sick_seg_header.sickPomPos.mainToOrigin.euler.yaw;

/*   pos->pos = aspectMap->Sick_seg_header.locoPos; */
/*   pos->var = aspectMap->Sick_seg_header.variances; */

  return OK;
}
#endif /* ASPECT */


/**
 ** LECTURE DES SEGMENTS ACQUIS PAR SEGLOC (repère robot)
 **
 **/
STATUS affGetSeglocAcqSegs(SEGLOC_ACQ_SEGS *acqSegs, 
			   int drawAcqSegsFlag,
			   SEGLOC_POS_VAR *posLocoSegs) 
{
  int sizeheader = sizeof(acqSegs->header);
  int nbSegs;
  int status;

  if (!seglocPosterInitialized)
    if (!affInitSegloc())
      return ERROR;

  /* Lecture header */
  status = posterRead(seglocSegsPosterId, 0, &acqSegs->header, sizeheader);
  if (status != sizeheader)
    return ERROR;

  /* Lecture segments */
  if (drawAcqSegsFlag) {
    nbSegs = acqSegs->header.nbSegs;
    status = posterRead(seglocSegsPosterId, 0, acqSegs, 
			sizeheader + nbSegs * sizeof(SEGLOC_SEG));
    if (status != sizeheader + nbSegs * sizeof(SEGLOC_SEG)) {
      printf ("affGetAcqSegs : posterRead error\n");
      return ERROR;
    }
    if (acqSegs->header.nbSegs > nbSegs) {
      printf ("affGetAcqSegs warning : poster change and %d segs are lost\n",
	      acqSegs->header.nbSegs - nbSegs);
      acqSegs->header.nbSegs = nbSegs;
    }
  }  
  
  /* Enregistre pos loco utilisée par la segmentation */
  posLocoSegs->pos = acqSegs->header.locoPos;
  posLocoSegs->var = acqSegs->header.variances;

  return OK;
}

STATUS affGetPosSegs(SEGLOC_POS_VAR *posLocoSegs)
{
  int sizeheader = sizeof(SEGLOC_SEG_HEADER);
  int status;
  SEGLOC_SEG_HEADER header;

  if (!seglocPosterInitialized)
    if (!affInitSegloc())
      return ERROR;

  /* Lecture header */
  status = posterRead(seglocSegsPosterId, 0, &header, sizeheader);
  if (status != sizeheader)
    return ERROR;

  /* Enregistre pos loco utilisée par la segmentation */
  posLocoSegs->pos = header.locoPos;
  posLocoSegs->var = header.variances;

  return OK;
}

/**
 ** LECTURE DES SEGMENTS ACQUIS PAR SICK (repère robot)
 **
 **/
STATUS affGetSickAcqSegs(int *nbSegs,
			 SICK_SEG *sickSegs,
			 SEGLOC_POS_VAR *posLocoSick,
			 SEGLOC_POS_VAR *posMainToOrigin)
{
  static POSTER_ID sickSegsPosterId;
  int sizeheader = sizeof(SICK_SEG_HEADER);
  SICK_SEG_HEADER header;

  if (!sickSegsPosterId) 
    if (posterFind("sickSegs", &sickSegsPosterId) == ERROR)
      return ERROR;

  /* Lecture header */
  if (posterRead(sickSegsPosterId, 0, &header, sizeheader) != sizeheader)
    return ERROR;

  /* Lecture segments */
  *nbSegs = header.nbSegs;
  if (posterRead(sickSegsPosterId, sizeheader, sickSegs, 
		 *nbSegs * sizeof(SICK_SEG))
      != *nbSegs * sizeof(SICK_SEG)) {
    printf ("affGetAcqSegs : posterRead error\n");
    return ERROR;
  }
  
  /* Enregistre pos loco utilisée par la segmentation */
  posLocoSick->pos.x = header.sickPomPos.mainToBase.euler.x;
  posLocoSick->pos.y = header.sickPomPos.mainToBase.euler.y;
  posLocoSick->pos.theta = header.sickPomPos.mainToBase.euler.yaw;
  posLocoSick->var.vxx = header.sickPomPos.mainToBase.var.vxx;
  posLocoSick->var.vxy = header.sickPomPos.mainToBase.var.vxy;
  posLocoSick->var.vyy = header.sickPomPos.mainToBase.var.vyy;
  posLocoSick->var.vxt = header.sickPomPos.mainToBase.var.vox;
  posLocoSick->var.vyt = header.sickPomPos.mainToBase.var.voy;
  posLocoSick->var.vtt = header.sickPomPos.mainToBase.var.voo;

  posMainToOrigin->pos.x = header.sickPomPos.mainToOrigin.euler.x;
  posMainToOrigin->pos.y = header.sickPomPos.mainToOrigin.euler.y;
  posMainToOrigin->pos.theta = header.sickPomPos.mainToOrigin.euler.yaw;
  posMainToOrigin->var.vxx = header.sickPomPos.mainToOrigin.var.vxx;
  posMainToOrigin->var.vxy = header.sickPomPos.mainToOrigin.var.vxy;
  posMainToOrigin->var.vyy = header.sickPomPos.mainToOrigin.var.vyy;
  posMainToOrigin->var.vxt = header.sickPomPos.mainToOrigin.var.vox;
  posMainToOrigin->var.vyt = header.sickPomPos.mainToOrigin.var.voy;
  posMainToOrigin->var.vtt = header.sickPomPos.mainToOrigin.var.voo;

  return OK;
}

#ifdef M2D
STATUS affGetM2dObst(SEGLOC_ACQ_SEGS *acqSegs, int drawAcqSegsFlag)
{
  int sizeheader = sizeof(acqSegs->header);
  int nbSegs;
  int status;

  /* Poster adresse */
  if (m2dObstsPosterId==NULL) {
    if (posterFind(M2D_OBST_POSTER_NAME, &m2dObstsPosterId) == ERROR) {
      m2dObstsPosterId = NULL;
      printf ("poster %s not found\n", M2D_OBST_POSTER_NAME);
      return ERROR;
    }
  }

  status = posterRead(m2dObstsPosterId, 0, &acqSegs->header, sizeheader);
  if (drawAcqSegsFlag) {
    nbSegs = acqSegs->header.nbSegs;
    status = posterRead(m2dObstsPosterId, 0, acqSegs, 
			sizeheader + nbSegs * sizeof(SEGLOC_SEG));
    if (acqSegs->header.nbSegs > nbSegs) {
      printf ("affGetAcqSegs warning : poster change and %d segs are lost\n",
	      acqSegs->header.nbSegs > nbSegs);
      acqSegs->header.nbSegs = nbSegs;
    }
  }
  return status;
}
#endif

STATUS affGetAspectSegs(SEGLOC_ASPECT_SEGS *aspectSegs,
			SEGLOC_BOOL drawMatchedModelSegFlags, 
			SEGLOC_BOOL drawMatchedAcqSegFlags,
			SEGLOC_BOOL aspectSegsFlags)
{
  int sizeheader = sizeof(aspectSegs->header);
  int nbSegs=0;
  int status;

  /* Lecture en-tete */
  status = posterRead(seglocAspectPosterId, 0, &aspectSegs->header, 
		      sizeheader);
  if (status != sizeheader)
    return ERROR;

  /* Calcul du nombre de segments a lire */
  if (drawMatchedAcqSegFlags || drawMatchedModelSegFlags || aspectSegsFlags) {

    /* Les segments acquis appariés */
    nbSegs = aspectSegs->header.nbMatchSegs;

    /* Les segments modele appariés */
    if (drawMatchedModelSegFlags || aspectSegsFlags)
      nbSegs += aspectSegs->header.nbMatchSegs;

    /* Les segments de la carte d'aspect */
    if (aspectSegsFlags)
      nbSegs += aspectSegs->header.nbSegs;

    /* Test debordement */
    if (nbSegs > SEGLOC_MAX_ASPECT_SEGS) {
      printf ("affGetAspectSegs warning : too many segs (%d>%d)\n", 
	      nbSegs, SEGLOC_MAX_ASPECT_SEGS);
      nbSegs = SEGLOC_MAX_ASPECT_SEGS;
    }
  }

  /* Pas de segment */
  else 
    nbSegs=0;

  /* Lecture du poster */
  if (nbSegs != 0) {
    status = posterRead(seglocAspectPosterId, 0, aspectSegs, 
			sizeheader + nbSegs * sizeof(SEGLOC_SEG));
    if (status != sizeheader + nbSegs * sizeof(SEGLOC_SEG))
      return ERROR;
  }

/*   if (aspectSegs->header.nbSegs + 2 * aspectSegs->header.nbMatchSegs */
/*       > nbSegs) { */
/*     printf ("affGetAspectSegs warning : poster change and %d segs are lost\n", */
/* 	    aspectSegs->header.nbSegs  + 2 * aspectSegs->header.nbMatchSegs */
/* 	    - nbSegs); */
/*     aspectSegs->header.nbSegs = nbSegs; */
/*   } */

  return OK;
}


/**
 ** 
 **
 **/

STATUS affGetMap(SEGLOC_MAP *mapSegs)
{
  STATUS status;

  if (!seglocPosterInitialized)
    if (!affInitSegloc())
      return ERROR;

  status = seglocMapMapPosterRead(mapSegs);
  return status;
}


/**
 ** 
 ** TRAJECTORY
 **/
#ifdef M2D
STATUS affGetM2dTraj(PILO_TRAJECTORY *traj)
{
  static int m2dTrajInit = 0;

  if (m2dTrajInit == 0) {
    if (m2dPosterInit() != OK) {
      printf ("aff : cannot init m2dTrajPoster\n");
      return ERROR;
    }
    m2dTrajInit = 1;
  }

  if (m2dTrajtrajPosterRead(traj) == ERROR) {
    printf ("aff : Cannot read poster traj\n");
    m2dTrajInit = 0;
    return ERROR;
  }
  return OK;
}
#endif

#ifdef VSTP
STATUS affGetVstpTraj(PILO_TRAJECTORY *traj)
{
  static int vstpTrajInit = 0;

  if (vstpTrajInit == 0) {
    if (vstpPosterInit() != OK) {
      printf ("aff : cannot init vstpTrajPoster\n");
      return ERROR;
    }
    vstpTrajInit = 1;
  }

  if (vstpTrajtrajPosterRead(traj) == ERROR) {
    printf ("aff : Cannot read poster traj\n");
    vstpTrajInit = 0;
    return ERROR;
  }
  return OK;
}
#endif

#ifdef US
/**
 ** 
 **
 **/
STATUS affGetLocalUS(XR4000_MEASURES_IN_ROBOT_CART_TAB *us)
{
  if (xr4000UsMeasureInRobotCartPosterRead(us) == ERROR) {
    printf ("aff : Cannot read poster US\n");
    return ERROR;
  }
  return OK;
}
#endif

#ifdef ASPECT

/**
 ** 
 **
 **/
STATUS affGetLaserPointsAspect(SICK_SCANPOLAR_POSTER_STR *sickScan,
			       SEGLOC_POS_VAR *posLocoSick)
{
/*   static LASER_CONFIG aspectConfig; */
  SICK_MEASURES_HEADER_STR *header = &sickScan->Header;


  /* aspect poster */
  if(aspectLaserEchosSick_scanpolar_poster_strPosterRead ( sickScan ) != OK) {
    printf ("affGetLaserPoints: poster read failed\n");
    return ERROR;
  }

  /* Attention : Angle de départ de tir et non pas angle moyen !!! */
  /* XXXXXX ??? header->aspectPomPos.sensorToMain.yaw = aspectConfig.angle_min; */

  header->sickPomPos.sensorToMain.yaw += DEG_TO_RAD(-90);

  posLocoSick->pos.x = header->sickPomPos.mainToBase.euler.x;
  posLocoSick->pos.y = header->sickPomPos.mainToBase.euler.y;
  posLocoSick->pos.theta = header->sickPomPos.mainToBase.euler.yaw;
  posLocoSick->var.vxx = header->sickPomPos.mainToBase.var.vxx;
  posLocoSick->var.vxy = header->sickPomPos.mainToBase.var.vxy;
  posLocoSick->var.vyy = header->sickPomPos.mainToBase.var.vyy;
  posLocoSick->var.vxt = header->sickPomPos.mainToBase.var.vox;
  posLocoSick->var.vyt = header->sickPomPos.mainToBase.var.voy;
  posLocoSick->var.vtt = header->sickPomPos.mainToBase.var.voo;
    

  return  OK;
}

#endif /* ASPECT */

/**
 ** 
 **
 **/
STATUS affGetLaserPoints(SICK_MEASURES_HEADER_STR *header, 
			 double rho[SICK_N_PT_MAX],
			 SEGLOC_POS_VAR *posLocoSick)
{
  static int sickScanInit = 0;
  static LASER_CONFIG sickConfig;

  /* Init */
  if (sickScanInit == 0) {
    if (!affInitSick(&sickConfig, 1)) {
      printf ("aff : cannot init sickScanPoster\n");
      return ERROR;
    }
    sickScanInit = 1;
  }

  if(sickScanPolarHeaderPosterRead(header) != OK) {
    printf ("affGetLaserPoints: sickScanPolarHeaderPosterRead failed\n");
    sickScanInit = 0;
    return ERROR;
  }

  /* Attention : Angle de départ de tir et non pas angle moyen !!! */
  /* XXXXXX ??? header->sickPomPos.sensorToMain.yaw = sickConfig.angle_min; */
  header->sickPomPos.sensorToMain.euler.yaw += sickConfig.angle_min; 

  if (sickScanPolarRhoPosterRead(rho) != OK) {
    printf ("affGetLaserPoints: sickScanPolarRhoPosterRead failed\n");
    sickScanInit = 0;
    return ERROR;
  }
 
  posLocoSick->pos.x = header->sickPomPos.mainToBase.euler.x;
  posLocoSick->pos.y = header->sickPomPos.mainToBase.euler.y;
  posLocoSick->pos.theta = header->sickPomPos.mainToBase.euler.yaw;
  posLocoSick->var.vxx = header->sickPomPos.mainToBase.var.vxx;
  posLocoSick->var.vxy = header->sickPomPos.mainToBase.var.vxy;
  posLocoSick->var.vyy = header->sickPomPos.mainToBase.var.vyy;
  posLocoSick->var.vxt = header->sickPomPos.mainToBase.var.vox;
  posLocoSick->var.vyt = header->sickPomPos.mainToBase.var.voy;
  posLocoSick->var.vtt = header->sickPomPos.mainToBase.var.voo;
    
  return  OK;
}

#ifdef SICK2
STATUS affGetLaser2Points(SICK_MEASURES_HEADER_STR *header, 
			 double rho[SICK_N_PT_MAX],
			 SEGLOC_POS_VAR *posLocoSick)
{
  static int sickScanInit = 0;
  STATUS res;
  static LASER_CONFIG sickConfig;

  /* Init */
  if (sickScanInit == 0) {
    if (!affInitSick(&sickConfig, 2)) {
      printf ("aff : cannot init sick2ScanPoster\n");
      return ERROR;
    }
    sickScanInit = 1;
  }
    
  /* Get header */
  if(sick2ScanPolarHeaderPosterRead(header) != OK) {
    printf ("affGetLaserPoints: sick2ScanHeaderPosterRead failed\n");
    sickScanInit = 0;
    return ERROR;
  }

  /* Attention : Angle de départ de tir et non pas angle moyen !!! */
  header->sickPomPos.sensorToMain.yaw += header->theta_min;
  
  /* Get data */
  if (sick2ScanPolarRhoPosterRead(rho) != OK) {
    printf ("affGetLaserPoints: sick2ScanPolarDistancePosterRead failed\n");
    sickScanInit = 0;
    return ERROR;
  }
 
  posLocoSick->pos.x = header->sickPomPos.mainToBase.euler.x;
  posLocoSick->pos.y = header->sickPomPos.mainToBase.euler.y;
  posLocoSick->pos.theta = header->sickPomPos.mainToBase.euler.yaw;
  posLocoSick->var.vxx = header->sickPomPos.mainToBase.var.vxx;
  posLocoSick->var.vxy = header->sickPomPos.mainToBase.var.vxy;
  posLocoSick->var.vyy = header->sickPomPos.mainToBase.var.vyy;
  posLocoSick->var.vxt = header->sickPomPos.mainToBase.var.vox;
  posLocoSick->var.vyt = header->sickPomPos.mainToBase.var.voy;
  posLocoSick->var.vtt = header->sickPomPos.mainToBase.var.voo;
    
  return  OK;
}
#endif /* SICK2 */


/**
 ** affInitSick
 **
 **/

static BOOL affInitSick(LASER_CONFIG *sickConfig, int sickNum)
{   

  SICK_CONFIG_STR config;
  SICK_MEASURES_HEADER_STR Header;
  double angleRange;
  
  /**
   ** INIT SICK 1
   **/
  if (sickNum == 1) {
    /* Init posters */
    if (sickPosterInit() == ERROR) {
      printf ("affInitSick: sickPosterInit failed\n");
      return FALSE;
    }
    
    /* Lecture du poster pour init position sick */
    if(sickScanPolarHeaderPosterRead(&Header) != OK) {
      printf ("affInitSick: sickScanHeaderPosterRead failed\n");
      return FALSE;
    }
    if(sickConfigConfigPosterRead(&config) != OK) {
      printf ("affInitSick: sickCurrentConfigPosterRead failed\n");
      return FALSE;
    }
  }
#ifdef SICK2
  /**
   ** INIT SICK 2
   **/
  else {
    /* Init posters */
    if (sick2PosterInit() == ERROR) {
      printf ("affInitSick2: sick2PosterInit failed\n");
      return FALSE;
    }
    
    /* Lecture du poster pour init position sick2 */
    if(sick2ScanPolarHeaderPosterRead(&Header) != OK) {
      printf ("affInitSick2: sick2ScanHeaderPosterRead failed\n");
      return FALSE;
    }
    if(sick2ConfigConfigPosterRead(&config) != OK) {
      printf ("affInitSick2: sick2CurrentConfigPosterRead failed\n");
      return FALSE;
    }

  }
#endif /*SICK2*/

  /**
   ** RECORD CONFIG
   **/
  sickConfig->pos[0] = Header.sickPomPos.sensorToMain.euler.x;
  sickConfig->pos[1] = Header.sickPomPos.sensorToMain.euler.y;
  sickConfig->pos[2] = Header.sickPomPos.sensorToMain.euler.yaw;
  switch(config.angle_range) {
  case SICK_SCAN_100_DEG:
    angleRange = DEG_TO_RAD(100.0);
    break;
  case SICK_SCAN_180_DEG:
    angleRange = DEG_TO_RAD(180.0);
    break;
  default:
    printf ("affInitSick: unknown angle sick range (default: 180 deg)\n");
    angleRange = DEG_TO_RAD(180.0);
    break;
  }
  
  /* Angle min-max in robot frame */
/* XXXXXX
  sickConfig->angle_min = Header.sickPomPos.sensorToMain.yaw + Header.theta_min;
  sickConfig->angle_max = sickConfig->angle_min + angleRange;
  sickConfig->range = config.length_range_max;
*/
  sickConfig->angle_min = Header.sickPomPos.sensorToMain.euler.yaw  - angleRange/2.; 
  sickConfig->angle_max = Header.sickPomPos.sensorToMain.euler.yaw  + angleRange/2.;
  sickConfig->range = config.length_range_max;

  
  /* Angular resolution */
  switch(config.angle_resolution) {
  case SICK_SCAN_ONE_DEG_RESOLUTION:
    sickConfig->resolution = DEG_TO_RAD(1);
    break;
  case SICK_SCAN_HALF_DEG_RESOLUTION:
    sickConfig->resolution = DEG_TO_RAD(0.5);
    break;
  default:
    printf ("affInitSick: unknown angle sick resolution (default: 1 deg)\n");
    sickConfig->resolution = DEG_TO_RAD(1);
    break;
  }
  
  /* Length accuracy */
  switch(config.length_resolution) {
  case SICK_MEASURE_IN_MM:
    sickConfig->accuracy = 0.001;
    break;
  case SICK_MEASURE_IN_CM:
    sickConfig->accuracy = 0.01;
    break;
  case SICK_MEASURE_IN_10CM:
    sickConfig->accuracy = 0.1;
    break;
  default:
    printf ("affInitSick: unknown measure sick precision (default: 1 mm)\n");
    sickConfig->accuracy = 0.001;
    break;
  }

  /* Init OK */
  
  printf ("affInitSick : pos (%g %g %g)\n"
	  "           range %g meters (+/- %g)\n"
	  "           angle %g to %g radian (step of %g rd)\n",
	  sickConfig->pos[0], sickConfig->pos[1], sickConfig->pos[2], 
	  sickConfig->range, sickConfig->accuracy,
	  sickConfig->angle_min, sickConfig->angle_max, 
	  sickConfig->resolution);
  
  return TRUE;
}

#ifdef US2MAT

/**
 ** 
 **
 **/
BOOL 
affGetUs2matProbaGrid(US2MAT_PROBAGRID_POSTER_STR *grid)
{
  static int affGridPosterInit=FALSE;

  if (!affGridPosterInit) {
    if (us2matPosterInit() == ERROR) {
       h2perror("posterInit affGetUs2matProbaGrid");
      return FALSE;
    }     
    affGridPosterInit=TRUE;
  }

  if (us2matProbaGridPosterRead(grid) == ERROR) {
    printf ("aff : Cannot read poster Us2matProbaGrid\n");
    affGridPosterInit = FALSE;
    return FALSE;
  }
  return TRUE;
}

BOOL 
affGetUs2matInstantGrid(US2MAT_PROBAGRID_POSTER_STR *grid)
{
  static int affGridPosterInit=FALSE;

  if (!affGridPosterInit) {
    if (us2matPosterInit() == ERROR) {
       h2perror("posterInit affGetUs2matInstantGrid");
      return FALSE;
    }     
    affGridPosterInit=TRUE;
  }

  if (us2matInstantGridPosterRead((US2MAT_INSTANTGRID_POSTER_STR *)grid) == ERROR) {
    printf ("aff : Cannot read poster Us2matInstantGrid\n");
    affGridPosterInit = FALSE;
    return FALSE;
  }
  return TRUE;
}

BOOL 
affGetUs2matTernarGrid(US2MAT_GRID_POSTER_STR *grid)
{
  static int affGridPosterInit=FALSE;

  if (!affGridPosterInit) {
    if (us2matPosterInit() == ERROR) {
       h2perror("posterInit affGetUs2matGrid");
      return FALSE;
    }     
    affGridPosterInit=TRUE;
  }

  if (us2matGridPosterRead(grid) == ERROR) {
    printf ("aff : Cannot read poster Us2matGrid\n");
    affGridPosterInit = FALSE;
    return FALSE;
  }
  return TRUE;
}

/**
 ** 
 **
 **/
BOOL 
affGetGrid(GRID_GRID_POSTER_STR *grid)
{
  static POSTER_ID affGridPosterId=0;
  GRID_GRID_POSTER_STR *x = NULL;
  int offset = (int)&(x->grid) - (int)(x);
  int size;

  if (affGridPosterId == 0) {
    if (gridPosterInit() == ERROR) {
       h2perror("posterInit gridGridPoster");
      return FALSE;
    }     
    if (posterFind(GRID_GRID_POSTER_NAME, &affGridPosterId) == ERROR) {
      h2perror("posterInit gridGridPoster");
      return FALSE;
    }
    if (affGridPosterId == 0) {
      printf ("aff : affGridPosterId == 0!\n");
      return FALSE;
    }
  }

  gridGridrobotPosPosterRead(&grid->robotPos);
  gridGridgridPosPosterRead(&grid->gridPos);
  gridGridparamsPosterRead(&grid->params);
  size = grid->params.nbXCells * grid->params.nbYCells * sizeof(GRID_CELL);
  if (posterRead(affGridPosterId, offset, (char *)(&grid->grid), size) != size) {
    printf ("aff : Cannot read poster grid\n");
    affGridPosterId = 0;
    return FALSE;
  }
  return TRUE;
}

#endif /* ATRV */

#ifdef ZONE
/**
 ** 
 **
 **/
BOOL 
affGetZones(ZONES_MAP *zonesObst,  
	    ZONES_MAP *zonesTarget, 
	    ZONES_MAP *zonesSpecial,
	    ZONES_DESC *zonesTargetDesc)
{
  if (zoneVIRTUAL_OBSTMapZONES_MAPPosterRead(zonesObst) == ERROR) {
    printf ("aff : Cannot read poster zones obst\n");
    return FALSE;
  }
  if (zoneTARGETMapZONES_MAPPosterRead(zonesTarget) == ERROR) {
    printf ("aff : Cannot read poster zones obst\n");
    return FALSE;
  }
  if (zoneSPECIALMapZONES_MAPPosterRead(zonesSpecial) == ERROR) {
    printf ("aff : Cannot read poster zones obst\n");
    return FALSE;
  }
  if (zoneTARGETDescZONES_DESCPosterRead(zonesTargetDesc) == ERROR) {
    printf ("aff : Cannot read poster zones obst\n");
    return FALSE;
  }
  return TRUE;
}

BOOL 
affGetMapZone(SEGLOC_MAP *map)
{
  if (zoneNavMapMapPosterRead (map)== ERROR) {
    printf ("aff : Cannot read poster zones map\n");
    return FALSE;
  }
  return TRUE;
}
#endif
