#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>

#include <portLib.h>

#include "affSeglocOptions.h"

static const char *colorName[] = COLOR_NAME;

void affSeglocDrawOptionsInit(DRAW_FLAGS *drawFlags)
{
  /* Ellipses d'incertitude */
  drawFlags->drawUncFlag = DRAW_UNC_FLAG; 
  /* Traj m2d */
  drawFlags->drawTrajFlag = DRAW_TRAJ_FLAG;    
  /* Obst m2d */
  drawFlags->drawM2dObstFlag = DRAW_M2DOBST_FLAG;    
  /* Segments-obstacles grossis cot matire */
  drawFlags->drawRectFlag = DRAW_RECT_FLAG;    
  /* pas de carte */
  drawFlags->drawClearMapFlag = DRAW_CLEARMAP_FLAG;   
  drawFlags->drawUpdateMapFlag = DRAW_UPDATEMAP_FLAG;
  /* Position odo non-corrige */
  drawFlags->drawOdoFlag = DRAW_ODO_FLAG;     
  /* Position corrige par segloc */
  drawFlags->drawSeglocFlag = DRAW_SEGLOC_FLAG;  
#ifdef H2
  /* Robot  la hilare */
  drawFlags->drawRobotRondFlag = DRAW_ROBOTROND_FLAG;  
#else
  /* Robot  la diligent */
  drawFlags->drawRobotRondFlag = DRAW_ROBOTROND_FLAG;  
#endif
  /* Rafraichissement a chaque affichage */
  drawFlags->drawRedrawFlag = DRAW_REDRAW_FLAG;      
  /* Points laser */
  drawFlags->drawLaserPointsFlag = DRAW_LASERPOINTS_FLAG;  
  /* Segments laser */
  drawFlags->drawAcqSegsFlag = DRAW_ACQSEGS_FLAG;      
  /* Segments de la carte d'aspect (ie, utiliss pour la loc) */
  drawFlags->drawAspectSegsFlag = DRAW_ASPECTSEGS_FLAG;   
  /* la carte d'aspect calculee par le module aspect */
  drawFlags->drawAspectMapFlag = DRAW_ASPECTMAP_FLAG;   
#ifdef CLASSIFICATION
  drawFlags->drawMatchedPosFlag = DRAW_MATCHEDPOS_FLAG;
  /* Segments acquis utiliss pour la loc */
  drawFlags->drawMatchedAcqSegFlag = DRAW_MATCHEDACQSEG_FLAG;   
  /* Segments du modle utiliss pour la loc */
  drawFlags->drawMatchedModelSegFlag = DRAW_MATCHEDMODELSEG_FLAG; 
#else
  drawFlags->drawMatchedPosFlag = DRAW_MATCHEDPOS_FLAG;
  /* Segments acquis utiliss pour la loc */
  drawFlags->drawMatchedAcqSegFlag = DRAW_MATCHEDACQSEG_FLAG;   
  /* Segments du modle utiliss pour la loc */
  drawFlags->drawMatchedModelSegFlag = DRAW_MATCHEDMODELSEG_FLAG; 
#endif
  /* Suspend l'affichage */
  drawFlags->drawStopFlag = DRAW_STOP_FLAG;   

  drawFlags->drawZonesFlag = DRAW_ZONES_FLAG;
  drawFlags->drawGridFlag = DRAW_GRID_FLAG;
  drawFlags->drawGridType = DRAW_GRIDTYPE;
  drawFlags->drawGridFollowFlag = DRAW_GRIDFOLLOW_FLAG;

  /* US */
  drawFlags->drawUsOdoFlag = DRAW_US_ODO_FLAG;
  drawFlags->drawUsSeglocFlag = DRAW_US_SEGLOC_FLAG;
  drawFlags->drawUsTopFlag = DRAW_US_TOP_FLAG;
  drawFlags->drawUsDownFlag = DRAW_US_DOWN_FLAG;
}


/**
 ** draw_option
 **
 **
 **
 **/

#define STRINGKEYSIZE 8

BOOL draw_option(DRAW_FLAGS *drawFlags, XKeyEvent *event_key,
		 double *echelle, int *offset_x, int *offset_y) 
{
  static int mult_move = 1;
  static int mult_scale = 1;
  int offset_move = 20;
  double scale_move = 1.1;
  char keystring[STRINGKEYSIZE];
  char keychar;
  int i;
  char *gridTypeName[] = GRID_TYPE_NAME;

  /* On commence par debloquer un eventuel stop */
  drawFlags->drawStopFlag = 0;

  i = XLookupString(event_key, keystring, STRINGKEYSIZE, NULL, NULL);

  /* Caractre ascii identifi par XLookupString */
  if (i == 1) {

    keychar = keystring[0];

    /* q comme quit (61) */
    if (keychar == 'q')  
      return FALSE;
    
    /* Redraw : blanc 128 */
    else if (keychar == ' ') 
    {}

    /* 109:  c  comme clear (map) */
    else if (keychar == '?') {   
      printf (HELP);
    }
    
    /* 109:  c  comme clear (map) */
    else if (keychar == 'c')   
      if (drawFlags->drawClearMapFlag) {
	drawFlags->drawClearMapFlag = 0;
	printf("map on\n");
      }
      else {
	drawFlags->drawClearMapFlag = 1;
	printf("map off\n");
      }

    /* Update map */
    else if (keychar == 'u') {
      if (drawFlags->drawUpdateMapFlag) {
	drawFlags->drawUpdateMapFlag = 0;
	printf("update map off\n");
      }
      else {
	drawFlags->drawUpdateMapFlag = 1;
	printf("update map on\n");
      }
    }
    
    /* 64: r comme redraw (a chaque cycle) */
    else if (keychar == 'r') 
      if (drawFlags->drawRedrawFlag) {
	drawFlags->drawRedrawFlag = 0;
	printf("redraw off\n");
      }
      else {
	drawFlags->drawRedrawFlag = 1;
	printf("redraw on\n");
      }
    
    /* z : stop */
/*     else if (keychar == 'z')  */
/*       drawFlags->drawStopFlag = 1; */
    
   /* . comme point */
    else if (keychar == '.') 
      if (drawFlags->drawLaserPointsFlag & LASER_1_ON) {
	drawFlags->drawLaserPointsFlag &= ~LASER_1_ON;
	printf("laser 1 points off\n");
      }
      else {
	drawFlags->drawLaserPointsFlag |= LASER_1_ON;
	printf("laser 1 points on\n");
      }

    /* : comme point */
    else if (keychar == ':') 
      if (drawFlags->drawLaserPointsFlag & LASER_2_ON) {
	drawFlags->drawLaserPointsFlag &= ~LASER_2_ON;
	printf("laser 2 points off\n");
      }
      else {
	drawFlags->drawLaserPointsFlag |= LASER_2_ON;
	printf("laser 2 points on\n");
      }

    /* 63:  e  comme ellipse */
    else if (keychar == 'e') 
      if (drawFlags->drawUncFlag) {
	drawFlags->drawUncFlag = 0;
	printf("ellipse uncertainty off\n");
      }
      else {
	drawFlags->drawUncFlag = 1;
	printf("ellipse uncertainty on\n");
      }
    
    /* 65: t   comme traj */
    else if (keychar == 't') 
      if (drawFlags->drawTrajFlag) {
	drawFlags->drawTrajFlag = 0;
	printf("traj off\n");
      }
      else {
	drawFlags->drawTrajFlag = 1;
	printf("traj on\n");
      }

    else if (keychar == 'v') 
      if (drawFlags->drawM2dObstFlag) {
	drawFlags->drawM2dObstFlag = 0;
	printf("virtual obst off\n");
      }
      else {
	drawFlags->drawM2dObstFlag = 1;
	printf("virtual obst on\n");
      }
    
    /* 62: w  comme width (segments grossis) */
    else if (keychar == 'w')
      if (drawFlags->drawRectFlag) {
	drawFlags->drawRectFlag = 0;
	printf("width segs off\n");
      }
      else {
	drawFlags->drawRectFlag = 1;
	printf("width segs on\n");
      }
    
    /* 70: p  comme pos odo */
    else if (keychar == 'p')  
      if (drawFlags->drawOdoFlag) {
	drawFlags->drawOdoFlag = 0;
	printf("odo pos off\n");
      }
      else {
	drawFlags->drawOdoFlag = 1;
	printf("odo pos on\n");
      }
    
    /* 88: g   comme grid */ 
    else if (keychar == 'g') 
      if (drawFlags->drawGridFlag) {
	drawFlags->drawGridFlag = 0;
	printf("grid off\n");
      }
      else {
	drawFlags->drawGridFlag = 1;
	printf("grid on\n");
      }

   /* z :   comme zones */ 
    else if (keychar == 'z') 
#ifdef ZONE
      if (drawFlags->drawZonesFlag) {
	drawFlags->drawZonesFlag = 0;
	printf("zones off\n");
      }
      else {
	drawFlags->drawZonesFlag = 1;
	printf("zones on\n");
      }
#else
	printf ("WARNING: To display ZONE, edit Makefile.in and recompile...\n");
#endif
    /*  y :  comme les 3 types de grille */ 
    else if (keychar == 'y') {
      drawFlags->drawGridType++;
      if (drawFlags->drawGridType >= GRID_NB_TYPE)
	drawFlags->drawGridType=0;
	printf("grid type : %s\n", gridTypeName[drawFlags->drawGridType]);
      }
    
    /* x comme xr4000 (switch h2 <-> xr4000) */
    else if (keychar == 'x') 
      if (drawFlags->drawRobotRondFlag) {
	drawFlags->drawRobotRondFlag = 0;
	printf("h2 model\n");
      }
      else {
	drawFlags->drawRobotRondFlag = 1;
	printf("xr4000 model\n");
      }
    
    /* 85: s  comme segloc */
    else if (keychar == 's') 
      if (drawFlags->drawSeglocFlag) {
	drawFlags->drawSeglocFlag = 0;
	printf ("segloc pos off\n");
      }
      else {
	drawFlags->drawSeglocFlag = 1;
	printf ("segloc pos on\n");
      }
    
    /* 116 / comme segs */
    else if (keychar == '/') 
      if (drawFlags->drawAcqSegsFlag) {
	drawFlags->drawAcqSegsFlag = 0;
	printf ("acq segs off\n");
      }
      else {
	drawFlags->drawAcqSegsFlag = 1;
	printf ("acq segs on\n");
      }
    
    /* xxx \ comme autres segs */ 
    else if (keychar == '\\')  
      if (drawFlags->drawAspectMapFlag) {
	drawFlags->drawAspectMapFlag = 0;
	printf ("aspect map off\n");
      }
      else {
	drawFlags->drawAspectMapFlag = 1;
	printf ("aspect map on\n");
      }
    
    /* 84:   comme aspect */
    else if (keychar == 'a') 
      if (drawFlags->drawAspectSegsFlag) {
	drawFlags->drawAspectSegsFlag = 0;
	printf ("aspect map off\n");
      }
      else {
	drawFlags->drawAspectSegsFlag = 1;
	printf ("aspect map on\n");
      }
    
    /* 112: n  comme matched */
    else if (keychar == 'n') 
      if (drawFlags->drawMatchedAcqSegFlag) {
	drawFlags->drawMatchedAcqSegFlag = 0;
	printf ("matched acq segs off\n");
      }
      else {
	drawFlags->drawMatchedAcqSegFlag = 1;
	printf ("matched acq segs on\n");
      }
    
    /* 113: m  comme matched */
    else if (keychar == 'm')   
      if (drawFlags->drawMatchedModelSegFlag) {
	drawFlags->drawMatchedModelSegFlag = 0;
	printf ("matched model segs off\n");
      }
      else {
	drawFlags->drawMatchedModelSegFlag = 1;
	printf ("matched model segs on\n");
      }
    
    /* 111:   comme matched */
    else if (keychar == 'b') 
      if (drawFlags->drawMatchedPosFlag) {
	drawFlags->drawMatchedPosFlag = 0;
	printf ("matched pos off\n");
      }
      else {
	drawFlags->drawMatchedPosFlag = 1;
	printf ("matched pos on\n");
      }

    /* +   comme US top */
    else if (keychar == '+') 
      if (drawFlags->drawUsTopFlag) {
	drawFlags->drawUsTopFlag = 0;
	printf ("us top off\n");
      }
      else {
	drawFlags->drawUsTopFlag = 1;
	printf ("us top on\n");
      }
    
     /*  -  comme US down */
    else if (keychar == '-') 
      if (drawFlags->drawUsDownFlag) {
	drawFlags->drawUsDownFlag = 0;
	printf ("us down off\n");
      }
      else {
	drawFlags->drawUsDownFlag = 1;
	printf ("us down on\n");
      }
    
    /* =   comme US odo */
    else if (keychar == '=') 
      if (drawFlags->drawUsOdoFlag) {
	drawFlags->drawUsOdoFlag = 0;
	printf ("us odo off\n");
      }
      else {
	drawFlags->drawUsOdoFlag = 1;
	printf ("us odo on\n");
      }
    
     /*  _  comme US segloc */
    else if (keychar == '_') 
      if (drawFlags->drawUsSeglocFlag) {
	drawFlags->drawUsSeglocFlag = 0;
	printf ("us segloc off\n");
      }
      else {
	drawFlags->drawUsSeglocFlag = 1;
	printf ("us segloc on\n");
      }
    
    /*  68  i : Zoom in */
    else if (keychar == 'i') {
      double prevEch;

      (*offset_x) -= *echelle * (*offset_x);
      (*offset_y) -= *echelle * (*offset_y);
      prevEch = 1;
      (*echelle) /= (0.1*mult_scale+1)  * scale_move;
      (*offset_x) += *echelle/prevEch * (*offset_x);
      (*offset_y) += *echelle/prevEch * (*offset_y);
      printf ("scale %g\n", *echelle);
    }
    /* 69: o : Zoom out */
    else if (keychar == 'o') {
      double prevEch;
      prevEch = 1;
      (*offset_x) -= *echelle * (*offset_x);
      (*offset_y) -= *echelle * (*offset_y);
      (*echelle) *= (0.1*mult_scale+1) * scale_move;
      (*offset_x) += *echelle/prevEch * (*offset_x);
      (*offset_y) += *echelle/prevEch * (*offset_y);
      printf ("scale %g\n", *echelle);
    }
    
    /*  h : left  31: 89: */
    else if (keychar == 'h') {
      (*offset_x) += mult_move * offset_move;
      printf ("x offset %d\n", (*offset_x));
    }
    /* j : up   27 90 */
    else if (keychar == 'j') {
      (*offset_y) -= mult_move * offset_move;
      printf ("y offset %d\n", (*offset_y));
    }
    /* k : down 34 91 */
    else if (keychar == 'k') {
      (*offset_y) += mult_move * offset_move;
      printf ("y offset %d\n", (*offset_y));
    }
    /* l : right  35 92*/
    else if (keychar == 'l') {
      (*offset_x) -= mult_move * offset_move;
      printf ("x offset %d\n", (*offset_x));
    }
    /* Multiplicateur de mouvement */
    else if (keychar == '1' ||
	     keychar == '2' ||
	     keychar == '3' ||
	     keychar == '4' ||
	     keychar == '5' ||
	     keychar == '6' ||
	     keychar == '7' ||
	     keychar == '8' ||
	     keychar == '9') {
      mult_scale = atoi(keystring);
      mult_move = 2*atoi(keystring);
      printf("motions X %d\n", mult_move);
      printf("scale X %d\n", mult_scale);
    }

    else 
      printf ("Unkown commande : %c\n", keychar);
  }

  /* Pas decode' par XLookupString (fleches) */
  else if (i==0) {
    switch(event_key->keycode) {
      /* fleche left */
    case 31:
    case 100:
      (*offset_x) += mult_move * offset_move;
      printf ("x offset %d\n", (*offset_x));
      break;
      /* fleche up */
    case 27:
    case 98:
      (*offset_y) -= mult_move * offset_move;
      printf ("y offset %d\n", (*offset_y));
      break;
      /* fleche down */
    case 34:
    case 104:
      (*offset_y) += mult_move * offset_move;
      printf ("y offset %d\n", (*offset_y));
      break;
    /* fleche right */
    case 35:
    case 102:
      (*offset_x) -= mult_move * offset_move;
      printf ("x offset %d\n", *offset_x);
      break;

    default:
      printf("Unknown event.xkey.keycode = %d\n", event_key->keycode);
    } /* switch */
  }
  
  else {
    printf ("i=%d  keystring = %s\n", i, keystring);
  }

  return TRUE;
}


void initColorTab(Display *dpy, int screen, XColor colorTab[],
		  int pixnum[NB_GRID_COUL][NB_GRID_COUL])
{
  int i, j;
  XColor  rgb;
  int prevPixnum=0;

  for (i=0;i<MAX_COLOR;i++) {
    if (XAllocNamedColor(dpy, DefaultColormap(dpy, screen), 
			 colorName[i], &colorTab[i], &rgb) == 0) {
      printf ("Cannot alloc named color :  %s\n", colorName[i]);
/*       perror ("initColorTab"); */
      /* On re-utilise la couleur precedente */
      colorTab[i] = rgb;
    }  
  }

  for(i=0;i<NB_GRID_COUL;i++) {
    for(j=0;j<NB_GRID_COUL;j++) {
      rgb.red = 65535/NB_GRID_COUL*i;
      rgb.green = 65535/NB_GRID_COUL*j;
      rgb.blue = 0;
      rgb.pixel = 0;
      if (XAllocColor(dpy, DefaultColormap(dpy, screen), &rgb) == 0) {
	printf ("Cannot alloc color :  red %d green %d\n", 
		65535/NB_GRID_COUL*i, 65535/NB_GRID_COUL*j);
	/* On re-utilise la couleur precedente */
	pixnum[i][j] = prevPixnum;
      }
      else {
	pixnum[i][j] = rgb.pixel;
	prevPixnum = rgb.pixel;
      }
    }
  }
}

