#ifndef AFF_SEGLOC_OPTION_H
#define AFF_SEGLOC_OPTION_H


/**
 **  LES OPTIONS D'AFFICHAGE
 **
 **/

#define HELP "\n"\
	       "Controle de l'affichage\n"\
	       "   bouton milieu         : help\n"\
	       "   bouton droit          : help\n"\
	       "   bouton gauche         : affiche position souris\n"\
	       "\n"\
	       "   q                     : quit\n"\
	       "   c                     : carte on/off\n"\
	       "   r                     : redraw automatique on/off\n"\
	       "   barre d'espace        : redraw (une fois)\n"\
	       "\n"\
	       "   i/o                   : zoom in/out\n"\
	       "   h/j/k/l   ou fleches  : left/up/down/right\n"\
	       "   un entier de 1 a 9    : demultiplie les mouvements d'autant\n"\
	       "Elements dessines (switchs) :\n"\
	       "   p                     : affiche position odometrique\n"\
	       "   s                     : affiche position segloc\n"\
	       "   b                     : position calcule par SEGLOC (et non loco corrige)\n"\
	       "   e                     : ellipses d'incertitudes\n"\
	       "   t                     : trajectoire M2D\n"\
	       "   v                     : virtual obst M2D\n"\
	       "   z                     : ZONES\n"\
	       "   w                     : segments grossis\n"\
	       "   u                     : update map\n"\
	       "   g                     : grid (proba->instant->tern->non)\n"\
               "   f                     : follow (grid follow robot)\n"\
	       "   .                     : points laser sick 1\n"\
	       "   :                     : points laser sick 2\n"\
	       "   /                     : segments laser\n"\
	       "   \\                     : segments carte d'aspect du module aspect\n"\
	       "   n                     : segments matchs acquis\n"\
	       "   m                     : segments matchs modele\n"\
	       "   a                     : segments aspect\n"\
	       "   x                     : hilare/xr4000\n"\
	       "   =                     : us in odo frame (square: top - cross down)\n"\
	       "   _                     : us in segloc frame\n"\
	       "   +                     : us TOP\n"\
	       "   -                     : us DOWN\n"
  
  

typedef struct DRAW_FLAGS {
  int drawUncFlag;     /* Ellipses d'incertitude */
  int drawTrajFlag;    /* Traj m2d */
  int drawM2dObstFlag;    /* Obst m2d */
  int drawRectFlag;    /* Segments-obstacles grossis cot matire */
  int drawClearMapFlag;   /* pas de carte */
  int drawUpdateMapFlag;
  int drawOdoFlag;     /* Position odo non-corrige */
  int drawSeglocFlag;  /* Position corrige par segloc */
  int drawRobotRondFlag;  /* Robot  la diligent */
  int drawRedrawFlag;      /* Rafraichissement a chaque affichage */
  int drawLaserPointsFlag;  /* Points laser */
  int drawAcqSegsFlag;      /* Segments laser */
  int drawAspectMapFlag;      /* Segments du module aspect */
  int drawAspectSegsFlag;   /* Segments de la carte d'aspect */
  int drawMatchedPosFlag;
  int drawMatchedAcqSegFlag;   /* Segments acquis utiliss pour la loc */
  int drawMatchedModelSegFlag; /* Segments du modle utiliss pour la loc */
  int drawStopFlag;   /* Suspend l'affichage */
  int drawGridFlag;
  int drawGridType;  /* GRID_TYPE */
  int drawGridFollowFlag; 
  int drawZonesFlag;

  int drawUsOdoFlag;  /* US */
  int drawUsSeglocFlag;
  int drawUsTopFlag;
  int drawUsDownFlag;
} DRAW_FLAGS;

typedef enum GRID_TYPE {
  GRID_PROBA,
  GRID_INSTANT,
  GRID_TERNAR,
  GRID_NB_TYPE
} GRID_TYPE;
#define GRID_TYPE_NAME {"proba", "instant", "ternaire"}

typedef enum LASER_FLAGS {
  LASERS_OFF = 0x0,
  LASER_1_ON = 0x1,
  LASER_2_ON = 0x2
} LASER_FLAGS;
#define  LASERS_1_AND_2_ON  (LASER_1_ON | LASER_2_ON)


#define DRAW_UNC_FLAG         FALSE  /* Ellipses d'incertitude */
#define DRAW_TRAJ_FLAG        FALSE  /* Traj m2d */
#define DRAW_M2DOBST_FLAG     FALSE  /* Obst m2d */
#define DRAW_RECT_FLAG        FALSE  /* Segments-obstacles grossi cot matire */
#define DRAW_CLEARMAP_FLAG    FALSE  /* pas de carte */
#define DRAW_UPDATEMAP_FLAG   FALSE
#define DRAW_ODO_FLAG         TRUE   /* Position odo non-corrige */
#define DRAW_SEGLOC_FLAG      TRUE   /* Position corrige par segloc */
#ifdef H2
#define DRAW_ROBOTROND_FLAG   FALSE  /* Robot  la hilare */
#else
#define DRAW_ROBOTROND_FLAG   TRUE   /* Robot  la diligent */
#endif
#define DRAW_REDRAW_FLAG      FALSE  /* Rafraichissement a chaque affichage */
#define DRAW_LASERPOINTS_FLAG LASERS_OFF  /* Points laser */
#define DRAW_ACQSEGS_FLAG     TRUE   /* Segments laser */
#define DRAW_ASPECTMAP_FLAG      FALSE   /* Segments laser */
#define DRAW_ASPECTSEGS_FLAG  FALSE  /* Segments de la carte d'aspect
					  (ie, utiliss pour la loc) */
#ifdef CLASSIFICATION
#define DRAW_MATCHEDPOS_FLAG    TRUE
#define DRAW_MATCHEDACQSEG_FLAG TRUE   /* Segments acquis apparis */
#define DRAW_MATCHEDMODELSEG_FLAG TRUE /* Segments du modle apparis */
#else
#define DRAW_MATCHEDPOS_FLAG    FALSE
#define DRAW_MATCHEDACQSEG_FLAG  FALSE   /* Segments acquis apparis */
#define DRAW_MATCHEDMODELSEG_FLAG  FALSE /* Segments du modle apparis */
#endif
#define DRAW_STOP_FLAG  FALSE   /* Suspend l'affichage */

#define DRAW_GRID_FLAG  FALSE
#define DRAW_GRIDTYPE   GRID_TERNAR
#define DRAW_GRIDFOLLOW_FLAG  TRUE

#define DRAW_ZONES_FLAG  FALSE

#define DRAW_US_ODO_FLAG FALSE
#define DRAW_US_SEGLOC_FLAG FALSE
#define DRAW_US_TOP_FLAG FALSE
#define DRAW_US_DOWN_FLAG FALSE

/**
 **  LES COULEURS
 **
 **  (utiliser xcolorsel pour trouver des noms de couleur)
 **/

/* ATTENTION: COLOR_ENUM et COLOR_NAME doivent tre cohrents */
typedef enum COLOR_ENUM {
  COLOR_WHITE,
  COLOR_BLACK,
  COLOR_BLUE,
  COLOR_GREEN,
  COLOR_DARK_GREEN,
  COLOR_ORANGE,
  COLOR_DARK_ORANGE,
  COLOR_RED,
  COLOR_DARK_RED,
  COLOR_PURPLE,
  COLOR_YELLOW,
  COLOR_GREY80,
  MAX_COLOR
} COLOR_ENUM; 

/* ATTENTION: COLOR_ENUM et COLOR_NAME doivent tre cohrents */
#define COLOR_NAME {\
  "white",\
  "black",\
  "blue",\
  "green",\
  "YellowGreen",\
  "orange",\
  "DarkOrange",\
  "red",\
  "maroon",\
  "purple",\
  "yellow",\
  "grey80"\
}

#define COUL_LOCO COLOR_BLUE
#define COUL_ASPECT COLOR_BLUE
#define COUL_MAP COLOR_BLACK
#define COUL_TRAJ COLOR_PURPLE

#define COUL_SICKOK COLOR_DARK_GREEN
#define COUL_SICKBOF COLOR_DARK_ORANGE
#define COUL_SICKPASOK COLOR_DARK_RED

#define COUL_SEGSOK COLOR_GREEN
#define COUL_SEGSBOF COLOR_ORANGE
#define COUL_SEGSPASOK COLOR_RED

#define COUL_MATCHEDACQ COLOR_PURPLE
#define COUL_MATCHEDMODEL COLOR_ORANGE

#define COUL_GRIDOCCUPE COLOR_RED
#define COUL_GRIDLIBRE COLOR_YELLOW
#define COUL_GRIDINCONNU COLOR_WHITE

#define COUL_M2DSEGS COLOR_PURPLE

#define COUL_ZONES_TOPO COLOR_GREY80
#define COUL_ZONES_STRESS  COLOR_DARK_GREEN

#define COUL_US_TOP COLOR_BLUE
#define COUL_US_DOWN COLOR_GREEN

#define NB_GRID_COUL 10

BOOL draw_option(DRAW_FLAGS *drawFlags, XKeyEvent *event_key,
		 double *echelle, int *offset_x, int *offset_y);
void affSeglocDrawOptionsInit(DRAW_FLAGS *drawFlags);
void initColorTab(Display *dpy, int screen, XColor colorTab[],
		  int pixnum[NB_GRID_COUL][NB_GRID_COUL]);

#endif /* AFF_SEGLOC_OPTION_H */

