#include <stdio.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <sys/time.h>

#include <portLib.h>
#include <h2timerLib.h>

#ifdef H2
#include "locoStruct.h"
#endif

#ifdef XR4000
#include "xr4000Struct.h"
#include "locoCompat.h"
#endif

#if defined(ATRV) || defined(B21R)
#include "rflexStruct.h"
#include "locoCompat.h"
#endif

#include "dessin_seg.h"
#include "affSegloc.h"

#ifdef PS_OUT
#include "sam_list.h"

static SAM_LIST *xpl = NULL;
static GC last_gc = NULL;
#endif /* PS_OUT */

double        echelle=SCALE;			/* scale */
int offset_x = OFFSET_X;  /* XSIZE/2.; */
int offset_y = OFFSET_Y;  /* YSIZE/1.5 +200; */
int dim_y =  DIM_Y;

double robot_width  = ROBOT_WIDTH;
double robot_length =  ROBOT_LENGTH;
double robot_edge = ROBOT_EDGE;

static XAFF_POS s_pos = {0., 0., 0.};


int 
r_m_to_pixel(double r)
{
  return (int)(rint(r/echelle + 0.5));
}

int 
x_metre_to_pix(double x)
{
    return (int)(rint(x/echelle)+offset_x);
}

double
x_pix_to_metre(short x)
{
  return (double)(x-offset_x)*echelle;
}
int 
y_metre_to_pix(double y)
{
    return (dim_y - (int)(rint(y/echelle))-offset_y);
}
double
y_pix_to_metre(short y)
{
  return (double)( -y - offset_y + dim_y)*echelle;
}

void metrePix(double x, double y, int *px, int *py)
{
    *px = x_metre_to_pix(x);
    *py = y_metre_to_pix(y);
}
void metrePixShort(double x, double y, short *px, short *py)
{
    *px = x_metre_to_pix(x);
    *py = y_metre_to_pix(y);
}
void 
local_to_global(XAFF_POS *pos, double xl, double yl, double *xg, double *yg)
{
    double c, s;

    c = cos(pos->theta);
    s = sin(pos->theta);
    
    *xg = pos->x + xl * c - yl *s;
    *yg = pos->y + xl * s + yl *c;
}

void 
dessin_seg (XAFF_POS *pos, double x0, double y0, double x1, double y1, GC gc)
{
  int xe0,ye0,xe1,ye1;
  double xl1,xl2,yl1,yl2;        
  
#ifdef PS_OUT
  if (ps_out)
    ps_dessin_seg (pos, x0, y0, x1, y1, gc);
#endif

  /* Null position */
  if(!pos)
    pos = &s_pos;
  
  local_to_global(pos, x0, y0, &xl1, &yl1);
  local_to_global(pos, x1, y1, &xl2, &yl2);
  
  metrePix(xl1,yl1,&xe0,&ye0);
  metrePix(xl2,yl2,&xe1,&ye1);
  XDrawLine(dpy, win, gc,xe0,ye0,xe1,ye1);
}

/**
 ** dessin_rects - dessine nbRects rectangles dont les coord des 4 points
 **                sont successivement stockes dans les tableaux (xtab, ytab)
 **
 **/

void 
dessin_rects (double *xtab, double *ytab, int nbRects, GC gc)
{
  int rec, i;
  XPoint xpoint[4];
  short sx, sy;

  for (rec=0;rec<nbRects;rec++) {
    for (i=0; i<4; i++) {

      metrePixShort(xtab[rec*4+i],ytab[rec*4+i],&sx,&sy);

      xpoint[i].x = sx;
      xpoint[i].y = sy;
    }
    XFillPolygon(dpy, win, gcgris, xpoint, 4, Convex, CoordModeOrigin);
    XDrawLines(dpy, win, gc, xpoint, 4, CoordModeOrigin);
  }
}

/**
 ** dessin_rect_from_seg - dessine un rectangle aprs grossissement 
 **                du segment (x1,x2,y1,y2)
 **
 **    Suppose que la matire est  droite quand on marche du point 1
 **    vers le point 2.
 **/

void 
dessin_rect_from_seg (double x1, double y1, double x2, double y2, GC gc)
{
  double xl1,yl1;
    
  double angle;
  double cs, sn;
  double dx, dy;

  double seglength = .1;  /* Grossissement du segment */

  XPoint xpoint[4];
  int x[4];
  int y[4];
  int i;

  angle = atan2(y2-y1, x2-x1) - M_PI/2.;
  cs = cos(angle);
  sn = sin(angle);
  
  dx = seglength * cos(angle);
  dy = seglength * sin(angle);

  xl1   = x1 + dx;
  yl1   = y1 + dy;
  metrePix(xl1,yl1,&x[0],&y[0]);

  xl1   = x1;
  yl1   = y1;
  metrePix(xl1,yl1,&x[1],&y[1]);

  xl1   = x2;
  yl1   = y2;
  metrePix(xl1,yl1,&x[2],&y[2]);

  xl1   = x2+dx;
  yl1   = y2+dy;
  metrePix(xl1,yl1,&x[3],&y[3]);

  for (i=0; i<4; i++) {
    xpoint[i].x = x[i];
    xpoint[i].y = y[i];
  }

  XFillPolygon(dpy, win, gcgris, xpoint, 4, Convex, CoordModeOrigin);
  XDrawLines(dpy, win, gc, xpoint, 4, CoordModeOrigin);

}

/**
 ** dessin_seg_extremities - dessine un seg avec extremitees
 **                du segment (x1,x2,y1,y2)
 **
 **    Suppose que la matire est  droite quand on marche du point 1
 **    vers le point 2.
 **/

void 
dessin_seg_extremities (XAFF_POS *pos, double x1in, double y1in, double x2in, double y2in, GC gc)
{
  double x1, y1, x2, y2;

  double xl1,yl1;
    
  double angle;
  double cs, sn;
  double dx, dy;

  double seglength = .1;  /* Grossissement du segment */

  XPoint xpoint[4];
  int x[4];
  int y[4];
  int i;

  /* Null position */
  if(!pos)
    pos = &s_pos;

  local_to_global(pos, x1in, y1in, &x1, &y1);
  local_to_global(pos, x2in, y2in, &x2, &y2);

  angle = atan2(y2-y1, x2-x1) - M_PI/2.;
  cs = cos(angle);
  sn = sin(angle);
  
  dx = seglength * cos(angle);
  dy = seglength * sin(angle);

  xl1   = x1 + dx;
  yl1   = y1 + dy;
  metrePix(xl1,yl1,&x[0],&y[0]);

  xl1   = x1;
  yl1   = y1;
  metrePix(xl1,yl1,&x[1],&y[1]);

  xl1   = x2;
  yl1   = y2;
  metrePix(xl1,yl1,&x[2],&y[2]);

  xl1   = x2+dx;
  yl1   = y2+dy;
  metrePix(xl1,yl1,&x[3],&y[3]);

  for (i=0; i<4; i++) {
    xpoint[i].x = x[i];
    xpoint[i].y = y[i];
  }

  XDrawLines(dpy, win, gc, xpoint, 4, CoordModeOrigin);

}



void 
dessin_my_robot(XAFF_POS *pos, double scale, GC gc)
{
    double xl1,xl2,yl1,yl2;	/* coord. segments en metres */
    int xp1,xp2,yp1,yp2;		/* coord. segments en pixels */
    double robot_widthl = robot_width * scale;
    double robot_lengthl = robot_length * scale;
    double robot_edgel = robot_edge *scale;

    /* Null position */
    if(!pos)
	pos = &s_pos;

    /* Axis */
    local_to_global(pos, -.1 * scale, 0, &xl1, &yl1);
    local_to_global(pos,  .2 * scale, 0, &xl2, &yl2);
    
    metrePix(xl1,yl1,&xp1,&yp1); 
    metrePix(xl2,yl2,&xp2,&yp2);
    XDrawLine(dpy, win, gc,xp1,yp1,xp2,yp2);
/*     dessin_arrow(xl2, yl2, pos->theta, NULL, gc); */
    
    local_to_global(pos, 0., -.1 * scale, &xl1, &yl1);
    local_to_global(pos, 0., 0.2 * scale, &xl2, &yl2);
    
    metrePix(xl1,yl1,&xp1,&yp1); 
    metrePix(xl2,yl2,&xp2,&yp2);
    XDrawLine(dpy, win, gc,xp1,yp1,xp2,yp2);
/*     dessin_arrow(xl2, yl2, (pos->theta + M_PI/2), NULL, gc); */


    local_to_global(pos, -(robot_lengthl/2.),-(robot_widthl/2.), /* TOUR ROBOT */
		    &xl1,&yl1);
    local_to_global(pos, (robot_lengthl/2.)-robot_edgel,-(robot_widthl/2.), &xl2, &yl2);
    metrePix(xl1,yl1,&xp1,&yp1); 
    metrePix(xl2,yl2,&xp2,&yp2);
    
    XDrawLine(dpy, win, gc,xp1,yp1,xp2,yp2);
    
    local_to_global(pos, (robot_lengthl/2.), -(robot_widthl/2.)+robot_edgel, &xl1, &yl1);
    metrePix(xl1,yl1,&xp1,&yp1); 
    XDrawLine(dpy, win, gc,xp1,yp1,xp2,yp2);
    
    local_to_global(pos, (robot_lengthl/2.),(robot_widthl/2.)-robot_edgel, &xl2, &yl2);
    metrePix(xl2,yl2,&xp2,&yp2);
    XDrawLine(dpy, win, gc,xp1,yp1,xp2,yp2);
    
    local_to_global(pos, (robot_lengthl/2.)-robot_edgel,(robot_widthl/2.), &xl1, &yl1);
    metrePix(xl1,yl1,&xp1,&yp1); 
    XDrawLine(dpy, win, gc,xp1,yp1,xp2,yp2);
    
    local_to_global(pos, -(robot_lengthl/2.),(robot_widthl/2.), &xl2, &yl2);
    metrePix(xl2,yl2,&xp2,&yp2);
    XDrawLine(dpy, win, gc,xp1,yp1,xp2,yp2);
    
    local_to_global(pos, -(robot_lengthl/2.),-(robot_widthl/2.), &xl1, &yl1);
    metrePix(xl1,yl1,&xp1,&yp1); 
    XDrawLine(dpy, win, gc,xp1,yp1,xp2,yp2);
    
}


void 
dessin_robot_rond(XAFF_POS *pos, double radiusl, GC gc)
{
    double xl,yl;	/* coord. segments en metres */
    int xp,yp;		/* coord. segments en pixels */
    int xc, yc;
    int radiusp;
    
    /* Null position */
    if(!pos)
      return;

    xl = pos->x;
    yl = pos->y;

    metrePix(xl,yl,&xp,&yp); 
    radiusp = r_m_to_pixel(radiusl);
    
    xc = xp - radiusp;
    yc = yp - radiusp;
    XDrawArc(dpy, win, gc, xc, yc, 2*radiusp, 2*radiusp, 0, 360*64);


    XDrawLine(dpy, win, gc,xp,yp,
	      xp + (int)(radiusp*cos(pos->theta)),
	      yp - (int)(radiusp*sin(pos->theta)));
}


void 
dessin_robot_h2(XAFF_POS *pos, double scale, GC gc)
{
   
#ifdef PS_OUT
    if (ps_out)
      ps_dessin_robot (pos, gc);
#endif

    dessin_my_robot(pos, scale, gc);
}


void 
dessin_gros_point(double xl, double yl, double size, GC gc)
{
    int xp,yp;		/* coord. point en pixels */
    int sizei;

    metrePix(xl,yl,&xp,&yp); 
    
    sizei = (int)size;

    /* Faudrait-il faire les carres zoomable ou pas ? 
     (sizei=r_m_to_pixel(sizei);*/

    if (sizei != 0) {
      XDrawRectangle(dpy, win, gc, 
		     xp-sizei, yp-sizei, 
		     2*sizei, 2*sizei); 
    }

    XDrawPoint(dpy, win, gc, xp, yp);
}

void 
dessin_croix(double xl, double yl, double size, GC gc)
{
    int xp,yp;		/* coord. point en pixels */
    int sizei;

    metrePix(xl,yl,&xp,&yp); 
    
    sizei = (int)size;

    /* Faudrait-il faire les croix zoomables ou pas ? 
     (sizei=r_m_to_pixel(sizei);*/


    if (sizei != 0) {
      XDrawLine(dpy, win, gc, xp-sizei, yp-sizei, xp+sizei, yp+sizei);
      XDrawLine(dpy, win, gc, xp-sizei, yp+sizei, xp+sizei, yp-sizei);
    }
    else
      XDrawPoint(dpy, win, gc, xp, yp);
}




/* Peut etre utilie un jour si on veut produire du postscript ... */
#ifdef PS_OUT

typedef struct {
  unsigned long pixel;
  XColor xc;
} X_PS_COLOR;


void
ps_draw_line (FILE *out, int x1, int y1, int x2, int y2)
{
  fprintf(out, "\t%d %d moveto\n\t%d %d lineto\n", x1, y1, x2, y2);
}

void
ps_find_rgb (GC gc, int *r, int *g, int *b)
{
  X_PS_COLOR *s;

  *r = *g = *b = 0;

  if (!xpl) 
    return;
  sl_goto_first (xpl);
  while (sl_more(xpl)) {
    s = sl_get_data(xpl);
    if (s->pixel == gc->values.foreground) {
      *r = s->xc.red;
      *g = s->xc.green;
      *b = s->xc.blue;
      return;
    }
    sl_next(xpl);
  }
}

void
ps_set_color (FILE *out, GC gc)
{
  int r, g, b;
  
  if (gc == last_gc) 
    return;

  ps_find_rgb (gc, &r, &g, &b);
  if (control->ps_color)
    fprintf(out, "%.3f %.3f %.3f setcolor\n", 
	    (double)r / (double)0xffff, 
	    (double)g / (double)0xffff, 
	    (double)b / (double)0xffff);
  else
    fprintf (out, "%.1f setgray\n", .333 * (double) (r+g+b) / (double)0xffff);
  
  last_gc = gc;
}

void
ps_comment (FILE *out, char *c)
{
  fprintf(out, "%% %s\n", c);
}

void 
ps_dessin_seg (XAFF_POS *pos, double x0, double y0, double x1, double y1, GC gc)
{
  int xe0,ye0,xe1,ye1;
  double xl1,xl2,yl1,yl2;        
  
  ps_comment(ps_out, "Segment");

  ps_set_color (ps_out, gc);

  /* Null position */
  if(!pos)
    pos = &s_pos;
  

  local_to_global(pos, x0, y0, &xl1, &yl1);
  local_to_global(pos, x1, y1, &xl2, &yl2);
  
  metrePix(xl1,yl1,&xe0,&ye0);
  metrePix(xl2,yl2,&xe1,&ye1);
  ps_draw_line (ps_out, xe0, ye0, xe1, ye1);

}

#endif /* PS_OUT */
