/***
 *** Interface graphique pour Hilare 2
 ***
 *** Trace' des ellipses d'incertitude du robot
 ***
 *** D'apres calife (Pierrick Grandjean et Philippe Moutarlier)
 *** Matthieu Herrb - Juin 1992
 ***
 *** Copyright (c) 1991-92 LAAS/CNRS 
 ***
 *** $Source: /cvs/openrobots/modules/segloc/xaff/ellipses.c,v $
 *** $Revision: 1.3 $
 *** $Date: 2006/07/13 18:08:13 $
 *** 
 ***/
#include <stdio.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <sys/time.h>

#include <portLib.h>
#include <h2timerLib.h>

#ifdef H2
#include "locoStruct.h"
#endif

#ifdef XR4000
#include "xr4000Struct.h"
#include "locoCompat.h"
#endif

#if defined(ATRV) || defined(B21R) || defined(JIDO)
#include "rflexStruct.h"
#include "locoCompat.h"
#endif

#include "affSegloc.h"
#include "dessin_seg.h"
#include "ellipses_p.h"

#if 0
#include <stdio.h>
#include <math.h>
#include <X11/Xlib.h>
#include <xview/xview.h>
#include <h2mathLib.h>
#include "gr_types.h"
#include "graphique_ui.h"
#include "coord.h"
#include "global.h"

#include "libc.h"
#endif /* 0 */

static void dess_ell_gen (int x0, int y0, double *var, double ech, GC gc);
static void dess_ell0 (int x0, int y0, double u, double v,
		       double dx, double dy, GC gc);
static void diag2 (double m11, double m12, double m22,
		   double *vp1, double *vp2, double *u, double *v);


/*----------------------------------------------------------------------*/

void dessin_incertitude(x, y, theta, var, gc)
     double x, y, theta;
     double *var;
     GC gc;
{
    double chi2_95 = 2.447;	/* valeur du test du Chi2 a 2 degres 
				   de liberte a 95% */
    double gauss_95 = 1.96;	/* idem pour 1 dL (gauss) */
    double dtheta;
    int r = r_m_to_pixel(0.6);
    double x0, y0, x1, y1, x2, y2, x3, y3;

    

#ifdef PS_OUT
    if (ps_out) {
	fprintf(ps_out, "%% incertitude:\ngsave 0 setlinewidth\n");
    }
#endif
    /* dessin ellipse */
    dess_ell_gen(x_metre_to_pix(x), y_metre_to_pix(y), var,
		 chi2_95*r_m_to_pixel(1), gc);

    /* dessin incertitude direction */
    x0 = (double)x_metre_to_pix(x);
    y0 = (double)y_metre_to_pix(y);
    dtheta =  sqrt(var[5])*gauss_95;	/* ecart type * prob a 95 %*/
    
    x1 = x0 + r * cos(theta + dtheta/2);
    y1 = y0 - r * sin(theta + dtheta/2);
    XDrawLine(dpy, win, gc, (int)(x0 + 0.5), (int)(y0 + 0.5),
	      (int)(x1 + 0.5), (int)(y1 + 0.5));
    
    x2 = x0 + r * cos(theta - dtheta/2);
    y2 = y0 - r * sin(theta - dtheta/2);
    XDrawLine(dpy, win, gc, (int)(x0 + 0.5), (int)(y0 + 0.5),
	      (int)(x2 + 0.5), (int)(y2 + 0.5));
	      
    x3 = x0 + r * cos(theta);
    y3 = y0 - r * sin(theta);
    XDrawLine(dpy, win, gc, (int)(x0 + 0.5), (int)(y0 + 0.5),
	      (int)(x3 + 0.5), (int)(y3 + 0.5));

#ifdef PS_OUT
    if (ps_out) {
	fprintf(ps_out, "newpath %.1f %.1f moveto \
%.1f %.1f %d %.1f %.1f arc closepath  stroke\n",
		x0, y0, x0, y0, r, RAD_TO_DEG(-theta - dtheta/2),
		RAD_TO_DEG(-theta + dtheta/2));
	fprintf(ps_out, "grestore\n\n");
    }
#endif
    
}

/************************************************************
  
  dr_ellipse1
  
  dessin d'une ellipse definie par son centre et sa matrice d'inertie
  
  - Entree : fen    = numero fenetre (initialisee)
  x0, y0 = centre de l'ellipse
  sx2, sxy, sy2 = coefficients d'inertie
  ech = facteur de grossissement a appliquer a l'ellipse
  -----------------------------------------------------------*/

#define NPAS_ELLIPSE 10

static void dess_ell_gen (x0, y0, var, ech, gc)
     int x0, y0;
     double *var, ech;
     GC gc;
{
    double vp1, vp2, u, v;
    
    /* diagonalisation de la matrice d'inertie */
    
    diag2 (var[0], var[1], var[2], &vp1, &vp2, &u, &v);
    
    if (vp1 <= 0.) /* protection contre erreurs d'arrondi */
	vp1 = 0.;
    if (vp2 <= 0.)
	vp2 = 0.;
    
    vp1 = sqrt (vp1);
    vp2 = sqrt (vp2);
    
    /* dessin de l'ellipse */
    
    dess_ell0 (x0, y0, u, -v, vp1 * ech, vp2 * ech, gc);
}

/************************************************************
  
  dr_ellipse0
  
  dessin d'une ellipse definie par son centre, ses axes,
  et ses longueurs.
  
  - Entree : fen    = numero fenetre (initialisee)
  x0, y0 = centre de l'ellipse
  u,v    = vect. directeur (norme) d'un des 2 axes
  dx, dy = 1/2 longueurs des grand et petit axe
  -----------------------------------------------------------*/

static void dess_ell0 (x0, y0, u, v, dx, dy, gc)
     int x0, y0;
     double u, v, dx, dy;
     GC gc;
{
    double x, y, xr, yr, dydx, dx2, temp, deltax;
    int i1, j1, i2, j2, i3, j3, i4, j4;
    
    /* si necessaire on inverse les axes pour echantillonner
       le long du plus grand axe */
    
    if (fabs (dy) > fabs (dx)) {
	temp = dx;
	dx = dy;
	dy = temp;
	temp = u;
	u = - v;
	v = temp;
    }
    
    /* dessin de l'ellipse */
    /*---------------------*/
    
    dydx = dy / dx;
    dx2 = dx * dx;
    
    xr = x0 + dx * u;
    yr = y0 + dx * v;
    i1 = xr;
    j1 = yr;
    i2 = i1;
    j2 = j1;
    
    deltax = dx / NPAS_ELLIPSE;
    if (deltax < 1.)
	deltax = 1.;
    
    for (x = dx - 1; x > - dx; x -= deltax) {
	y = dydx * sqrt (dx2 - x * x);
	
	xr = x0 + x * u - y * v;
	yr = y0 + x * v + y * u;
	i3 = (int)(xr + 0.5);
	j3 = (int)(yr + 0.5);
	XDrawLine(dpy, win, gc, i1, j1, i3, j3);
	
	xr = x0 + x * u + y * v;
	yr = y0 + x * v - y * u;
	i4 = (int)(xr + 0.5);
	j4 = (int)(yr + 0.5);
	XDrawLine(dpy, win, gc, i2, j2, i4, j4);
	
	i1 = i3;
	j1 = j3;
	i2 = i4;
	j2 = j4;
    }
    
    xr = x0 - dx * u;
    yr = y0 - dx * v;
    i3 = (int)(xr + 0.5);
    j3 = (int)(yr + 0.5);
    i4 = i3;
    j4 = j3;
    
    XDrawLine(dpy, win, gc, i1, j1, i3, j3);
    XDrawLine(dpy, win, gc, i2, j2, i4, j4);

#ifdef PS_OUT
    if (ps_out) {
	fprintf(ps_out, "gsave %d %d translate %.1f rotate\n",
		x0, y0, RAD_TO_DEG(atan2(v, u)));
	fprintf(ps_out, "%.1f %.1f scale\n", dx, dy);
	fprintf(ps_out, "newpath 0 0 1 0 360 arc stroke grestore\n");
    }
#endif
}


/**********************************************************
  
  diagms2
  
  diagonalisation d'une matrice symetrique 2*2
  
  Entree : m11, m12, m22 = coeffs de la matrice (m21 = m12)
  Sortie : *vp1, *vp2    = valeurs propres
  *u, *v        = vecteur propre associe a *vp1
  (l'autre est (- *v, *u))
  
  Principe : on ecrit directement
  M. (cos(teta), sin(teta)) = vp (cos(teta), sin(teta))
  et on en tire teta puis vp.
  ---------------------------------------------------------*/

static void diag2 (m11, m12, m22, vp1, vp2, u, v)
     double m11, m12, m22, *vp1, *vp2, *u, *v;
{
    double teta, co, si, co2, si2, cosi;
    
    teta = atan2 ((double)(2. * m12), (double)(m11 - m22)) / 2.;
    
    co = cos (teta);
    si = sin (teta);
    co2 = co * co;
    si2 = si * si;
    cosi = co * si;
    
    *vp1 = m11 * co2 + 2 * m12 * cosi + m22 * si2;
    *vp2 = m11 * si2 - 2 * m12 * cosi + m22 * co2;
    *u = co;
    *v = si;
}

